# Copyright (c) 2012  Gerald Knizia
# 
# This file is part of the IR/WMME program
# (See https://sites.psu.edu/knizia/)
# 
# IR/WMME is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
# 
# IR/WMME is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/

# This file provides support for reading libmol files. May be used to convert
# them to other formats or for patching up the basis functions.
from __future__ import print_function
import wmme  # for FBasisShell objects.
import numpy as np

class FNamedElementBasis(object):
   """Represents a complete basis of a given name for a given element.
   E.g., objects like "cc-pVTZ for Ne atom". This normally includes multiple
   FBasisShell objects."""
   def __init__(self, Name, Element, BasisFns, Type, Reference, Ecp=None):
      self.BasisFns = BasisFns  # should be list of FBasisShell objects.
      self.Name = Name
      if ( Type != 'ORBITAL' ):
         self.Name += "-%s" % Type
      self.Element = Element
      self.Reference = Reference
      #if ( not self.Reference.startswith(" ") ):
         #self.Reference = " " + self.Reference
      self.Ecp = Ecp
   def __str__(self):
      return "%s[%s]\n%s" % (self.Element, self.Name, "\n".join([str(o) for o in self.BasisFns]))

class FBasisLibraryError(Exception):
   def __init__(self, Msg):
      Exception.__init__(self, Msg)

class FBasisLibrary(object):
   def __init__(self, NamedBases):
      """Construct a basis set library from a list of FNamedElementBasis objects.
      FNamedElementBasis can be made, for example, via ReadMolproBasisLibrary."""
      self.NamedBases = NamedBases
      # make indices: for each element, find all FNamedElementBasis objects
      # assigned to it.

      # mape Element name -> basis name -> FNamedElementBasis object.
      self.ElementSets = {}
      self.ElementBasisNames = {}
      self.AllBasisNames = set([])
      for NamedBasis in self.NamedBases:
         if NamedBasis.Element not in self.ElementSets:
            self.ElementSets[NamedBasis.Element] = {}
         if NamedBasis.Name in self.ElementSets[NamedBasis.Element]:
            raise FBasisLibraryError("Multiple declarations of %s/%s." % (NamedBasis.Element, NamedBasis.Name))
         self.ElementSets[NamedBasis.Element][NamedBasis.Name] = NamedBasis

         L = self.ElementBasisNames.get(NamedBasis.Element, set())
         L.add(NamedBasis.Name)
         self.ElementBasisNames[NamedBasis.Element] = L
         self.AllBasisNames.add(NamedBasis.Name)

   def FindElementBasisNames(self, Element):
      return list(sorted(self.ElementBasisNames.get(Element, [])))

   def Find(self, Element, BasisName):
      """find a registered FNamedElementBasis object for given element name and basis name."""
      ElementSet = self.ElementSets.get(Element, None)
      if ElementSet is None:
         raise FBasisLibraryError("No basis for element '%s' was found." % Element)
      NamedBasis = ElementSet.get(BasisName, None)
      if NamedBasis is None:
         raise FBasisLibraryError("Basis '%s' for element '%s' not found. Available bases for '%s':"%(BasisName, Element, Element) + "".join("   \n"+o for o in ElementSet.keys()))
      assert(NamedBasis.Name == BasisName)
      return NamedBasis


def MakeBasisForAtoms(Atoms, BasisName, BasisLib):
   # construct a wmme.FBasisSet object for the given Atoms object.
   Shells = []
   for At in Atoms:
      if callable(BasisName):
         BasisName1 = BasisName(At)
      else:
         BasisName1 = BasisName
      NamedBasis = BasisLib.Find(At.Element, BasisName1)
      for Bf in NamedBasis.BasisFns:
         Shells.append(wmme.FBasisShell1(At, Bf))
   return wmme.FBasisSet(Shells, Atoms)



AngMomList = "spdfghikl"

def ReadMolproBasisLibrary(FileName):
   """returns list of FNamedElementBasis objects."""
   sets = []
   with open(FileName, "r") as File:
      FileContent = File.read()
   Lines = FileContent.splitlines()
   iEnd = 0
   while iEnd < len(Lines):
      iLine = iEnd
      iEnd += 1
      if ( Lines[iLine].startswith("*") or\
           Lines[iLine].startswith("!") or\
           Lines[iLine].strip() == "" ):
         # comment
         continue
      # next lines (contractions may be emitted):
      #[iLine+0]      BR s cc-pV5Z-JKFIT : 16 1 1.3
      #[iLine+1]      F. Weigend
      #[iLine+2]        40743.58269600   8836.99929430   2867.84280150   1125.87025770    492.27094145
      #[iLine+n]          211.31569981     97.79626271     51.17559433     26.50773254     14.47971897
      #[iLine+n]            6.69875789      3.37772865      1.82726486      0.85235747      0.40687569
      #[iLine+n]            0.18575841      0.08281921      0.33024477      0.80999791
      #
      # We keep on joining lines from iLine+2 until we find a line containing something
      # different than numbers.
      Numbers = []
      while 1:
         iEnd += 1
         try:
            L = Lines[iEnd].split()
            Ln = [ReadFortranFloat(o) for o in L]
         except:
            break
         Numbers += Ln

      sets += ReadMolproDecl(Lines[iLine], Lines[iLine+1], Numbers)

   SetsDirect = sets

   # join sets with multiple independent declaration parts for the
   # basis/element combination (this can also include multiple independent
   # declaration parts for the same same angular momenta in rare cases)
   AllBasisNames = list(set([o.Name for o in SetsDirect]))
   AllBasisNames.sort()
   AllSets = {}
   JoinedSets = []
   for Name in AllBasisNames:
      Sets = [o for o in SetsDirect if o.Name == Name]
      Elements = list(set([o.Element for o in Sets]))
      for Element in Elements:
         AngMomSets = [o for o in Sets if o.Element == Element]

         bfns = []
         Reference = AngMomSets[0].Reference.strip()
         for Set in AngMomSets:
            bfns += Set.BasisFns
            if ( Set.Reference.strip() != Reference ):
               print("Set.Reference: {0}".format(Set.Reference))
               print("   .Reference: {0}".format(Reference))
            assert( Set.Reference.strip() == Reference )

         bfns.sort(key=CmpBfnKey)
         JoinedSets.append(FNamedElementBasis(Name, Element, bfns, "ORBITAL", Reference))
   return JoinedSets

def ReadFortranFloat(s):
   try:
      return float(s)
   except ValueError as e:
      pass
   # may have FORTRAN style float exponent decorations (123d+94 instead of 123e+94).
   #print("!!ReadFloat: '%s'" % s)
   return float(s.replace("D-","e-").replace("d-","e-").replace("D+","e+").replace("d+","e+"))


def ReadMolproDecl(DeclLine,Reference,Numbers):
   # BR s cc-pVTZ-JKFIT cc-pVDZ-JKFIT : 16 1 1.3
   Decl = DeclLine.split()
   Element = Decl[0]
   Element = Element[:1].upper() + Element[1:].lower()
   AngMom = AngMomList.index(Decl[1].lower().strip())

   iSep = Decl.index(":")
   SetNames = Decl[2:iSep]
   sPattern = Decl[iSep+1:]

   nPrimitives = int(sPattern[0])
   AllExponents = Numbers[:nPrimitives]
   AllCoeffs = Numbers[nPrimitives:]

   nContractions = int(sPattern[1])
   assert(nContractions == len(sPattern)-2)
   Contractions = [[int(o) for o in s.split(".")] for s in sPattern[2:]]
   # this number of exponents occurs in contractions and is not added
   # as additional primitive without a contraction signature.
   if ( len(Contractions) != 0 ):
      MaxContractedExp = max([ p[1] for p in Contractions])
   else:
      MaxContractedExp = 0
   nNonContractedBfn = nPrimitives - MaxContractedExp

   # form the final contraction matrix.
   Co = np.zeros((nPrimitives, nContractions + nNonContractedBfn))

   # first for the functions with explicit contraction declaration
   iCo = 0
   iCoeff = 0
   for (cs,ce) in Contractions:
      iCoeffEnd = iCoeff + ce-cs+1
      Co[cs-1:ce, iCo] = AllCoeffs[iCoeff:iCoeffEnd]
      iCoeff = iCoeffEnd
      iCo += 1
   # add additional functions for primitives with implicit contraction declaration.
   for i in range(MaxContractedExp,nPrimitives):
      Co[i,iCo] = 1.
      iCo += 1

   # generate FNamedElementBasis objects for all basis names this set of
   # functions is associated with. Note that all of them refer to THE SAME OBJECT.
   Output = []
   Sh = wmme.FBasisShell(AngMom, AllExponents, Co)
   for BasisName in SetNames:
      Output.append(FNamedElementBasis(BasisName, Element, [Sh], "ORBITAL", Reference))
      # ^- note: object is still incomplete. multiple named basis objects with the same name
      #    still need to be joined.
   return Output

def CmpAtomBasis(x,y):
   i = cmp(ElementNumber(x.Element),ElementNumber(y.Element))
   if ( i != 0 ): return i
   return CmpBasisName(x.Name, y.Name)

#def CmpBfn(x,y):
   #i = cmp(x.AngMom, y.AngMom)
   #if ( i != 0 ): return i
   #i = cmp(len(x.Exp),len(y.Exp))
   #if ( i != 0 ): return -i
   #i = cmp(max(x.Exp),max(y.Exp))
   #if ( i != 0 ): return -i
   ##i = cmp(x.Coeffs[argmax(x.Exps)],y.Coeffs[argmax(y.Exps)])
   #i = cmp(x.Co.size, y.Co.size)
   #return i
def CmpBfnKey(x):
   return (x.AngMom, len(x.Exp), max(x.Exp), x.Co.size, tuple(x.Exp), tuple(x.Co))

def MangleBasisName(s):
   # cuts up basis names for Molpro libmol format.
   s = s.replace("/","-").replace("-OPTRI","_OPT")
   return s[:MaxLen]

def _TestBasisLibrary():
   BasisLibFiles = ["def2-nzvpp-jkfit.libmol", "minao.libmol", "emsl_cc-pVDZ.libmol"]
   BasisLib = FBasisLibrary(sum((ReadMolproBasisLibrary(o) for o in BasisLibFiles), []))

   TestElement = "He"
   print("Registered basis sets for element '%s':" % TestElement)
   BasisNames = BasisLib.FindElementBasisNames(TestElement)
   print("   " + "  ".join(BasisNames))
   print("These consist of the following functions:")
   for BasisName in BasisNames:
      print("  %s/%s" % (TestElement, BasisName))
      NamedBasis = BasisLib.Find(TestElement, BasisName)
      for bf in NamedBasis.BasisFns:
         print(bf)
      print()


if __name__ == "__main__":
   _TestBasisLibrary()
