# Copyright (c) 2012  Gerald Knizia
# 
# This file is part of the IR/WMME program
# (See https://sites.psu.edu/knizia/)
# 
# IR/WMME is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
# 
# IR/WMME is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/

from __future__ import print_function
import numpy as np

if 0:
   pResultFmt = " {:<32}{:18.12f}\n"
   pResultFmtAnnotated = " {:<32}{:18.12f}  ({})\n"
   pCountFmt = " {:<32}{:5}\n"
   pCountFmtAnnotated =" {:<32}{:5}  ({})\n"
   pInfoFmt = " {:<32}{:>18}\n"
   pInfoFmtAnnotated = " {:<32}{:>18}  ({})\n"
   pTimingFmt = "{:<40}{:10.2f} sec\n"
   pTimingFmtAnnotated = "{:<40}{:10.2f} sec  (n={})\n"
else:
   pResultFmt = " {:<42}{:18.12f}\n"
   pResultFmtAnnotated = " {:<42}{:18.12f}  ({})\n"
   pCountFmt = " {:<42}{:5}\n"
   pCountFmtAnnotated =" {:<42}{:5}  ({})\n"
   pInfoFmt = " {:<42}{:>18}\n"
   pInfoFmtAnnotated = " {:<42}{:>18}  ({})\n"
   pTimingFmt = "{:<50}{:10.2f} sec\n"
   pTimingFmtAnnotated = "{:<50}{:10.2f} sec  (n={})\n"

def WriteOutput(s):
   print(s, end='')


def WriteResult(Caption, Value, Annotation = None):
   if Annotation:
      WriteOutput(pResultFmtAnnotated.format(Caption, Value, Annotation))
   else:
      WriteOutput(pResultFmt.format(Caption, Value))


def WriteCount(Caption, Value, Annotation = None):
   if Annotation:
      WriteOutput(pCountFmtAnnotated.format(Caption, Value, Annotation))
   else:
      WriteOutput(pCountFmt.format(Caption, Value))


def WriteInfo(Caption, Value, Annotation = None):
   if Annotation:
      WriteOutput(pInfoFmtAnnotated.format(Caption, Value, Annotation))
   else:
      WriteOutput(pInfoFmt.format(Caption, Value))


def WriteInfoExpf(Caption, Value, Annotation = None):
   sValue = "{:8.2e}".format(Value)
   return WriteInfo(Caption, sValue, Annotation)


def WriteTiming(Caption, Value, N = None):
   if Annotation:
      WriteOutput(pTimingFmtAnnotated.format(Caption, Value/min(1,N), N))
   else:
      WriteOutput(pTimingFmt.format(Caption, Value))

def WriteArray(Caption, Array, FormatEntry = None):
   WriteOutput(" {}:\n {}\n".format(Caption, FormatArray(Array, FormatEntry=FormatEntry)))

def WriteMatrix(Caption, Array, FormatEntry = None, OpenSepClos = None):
   WriteOutput(" {}:\n{}\n".format(Caption, FormatMatrix(Array, FormatEntry=FormatEntry, OpenSepClos=OpenSepClos)))


def _FormatEntry(a, FormatEntry):
   if FormatEntry is None:
      return str(a)
   elif isinstance(FormatEntry, str):
      return FormatEntry.format(a)
   elif callable(FormatEntry):
      return FormatEntry(a)
   else:
      raise Exception("FormatArray encountered an FormatEntry object ('{}') it does not know what to do with.".format(repr(FormatEntry)))


def FormatArray(a, FormatEntry=None):
   if hasattr(a, '__iter__'):
      # it's still a sequence (list or array likely). Iterate over first rank and recurse.
      return "[{}]".format(", ".join(FormatArray(o, FormatEntry) for o in a))
   else:
      return _FormatEntry(a, FormatEntry)

def FormatMatrix(a, FormatEntry=None, OpenSepClos=None, Left=None, Right=None):
   if FormatEntry is None:
      FormatEntry = "{:14.6f}"
   if Left is None:
      Left = " "
   if Right is None:
      Right = ""
   if OpenSepClos is None:
      OpenSepClos = ("[", ",", "]")
   Open, Sep, Clos = OpenSepClos

   nRank = len(a.shape)
   if nRank == 0:
      return Left + _FormatEntry(a, FormatEntry) + Right
   elif nRank == 1:
      Elements = [_FormatEntry(o, FormatEntry) for o in a]
      return "{}{}{}{}{}".format(Left, Open, Sep.join(Elements), Clos, Right)

   Lines = []
   for iRow in range(a.shape[0]):
      Left0 = Open if iRow == 0 else (len(Open)*" ")
      Right0 = Clos if iRow + 1 == a.shape[1] else Sep
      Lines.append(FormatMatrix(a[iRow,:], FormatEntry, OpenSepClos, Left+Left0, Right0+Right))
   return "\n".join(Lines)


def FloatToStr1(f):
   if 1:
      # yes, I know...
      for c in range(-15,16):
         for d in range(1,16):
            cd = c/(1.*d)
            if abs(f - cd)<1e-10: f = cd
            if abs(f + cd)<1e-10: f = -cd
      # c/p'd from VtfAsm
      s = "%.15e" % f
      # ^- FIXME: I don't get it anymore...
      # (hm.. i think this is supposed to turn rational numbers into exact rationals)
      # unfortunately, it is very slow...
   else:
      if abs(f-1.)<1e-10: f = 1.
      if abs(f+1.)<1e-10: f = -1.
   s = "%.15e" % f

   ie = s.index('e')
   ilast = ie-1
   while s[ilast] == '0':
      ilast -= 1
   s = s[:ilast+1] + s[ie:]
   if ( s.endswith("e+00") ):
      s = s[:-4]
   #print >> stderr, "Float2Str: f=",f,"  ->  ",s
   return s


def FloatToStr(f):
   return "%r" % f
   return FloatToStr1(f)
