# Copyright (c) 2012  Gerald Knizia
# 
# This file is part of the IR/WMME program
# (See https://sites.psu.edu/knizia/)
# 
# IR/WMME is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
# 
# IR/WMME is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/

from __future__ import print_function
import numpy as np
from collections import namedtuple


FInputLine = namedtuple("FInputLine", ["iLine", "s"])



class FXyzAtom(object):
   def __init__(self, Element, Tag, Position, Index):
      self.Element = Element
      self.Tag = Tag
      self.Pos = Position
      assert(len(self.Pos) == 3)
      self.Index = Index
   @property
   def Label(self):
      # return element and center index combined.
      return "{:2}{:3}".format(self.Element + self.Tag, 1 + self.Index)
   @property
   def iElement(self):
      return ElementNumbers[self.Element]
   @property
   def ElementAndTag(self):
      return self.Element + self.Tag
   def __str__(self):
      return "{} ({:6.3f},{:6.3f},{:6.3f})".format(self.Label, self.Pos[0], self.Pos[1], self.Pos[2])


class FXyzFrame(object):
   def __init__(self, Caption, Data, InputScale, SourceFile = None, SourceFrame = None):
      self.Caption = Caption

      AtomNamesRaw = Data[0,:]
      ElementsRaw = []
      AtomTagsRaw = []
      for AtomName in AtomNamesRaw:
         e,t = SplitAtomElementTag(AtomName)
         ElementsRaw.append(NormalizeElementName(e))
         AtomTagsRaw.append(t)
      
      self.Elements = np.array(ElementsRaw)
      #self.AtomTags = np.array(AtomTagsRaw + ["abcdfd"], dtype=str)
      #self.AtomTags = self.AtomTags[:-1]
      #print("!type(self.AtomTags) {}".format(self.AtomTags.dtype))
      self.AtomTags = AtomTagsRaw  # for arrays we need to know beforehand what is the maximum length of contained strs.
      self.Pos = Data[1:4,:].astype(float) * InputScale
      
      # if given, specifies source .xyz file and the frame id (0-based)
      # if from a multi-frame .xyz
      self.FileName = SourceFile
      self.iFrame = SourceFrame
      assert(self.iFrame is None or type(self.iFrame) is int)
   def __str__(self):
      return "{}:{} ({} atoms)".format(self.SourceFile, self.SourceFrame, len(self))
   def MakeXyz(self,NumFmt = "%15.8f",Scale=1.):
      Lines = []
      LineFmt = " %5s {0} {0} {0}".format(NumFmt)
      for i in range(len(self.Elements)):
         Lines.append(LineFmt % (\
            self.Elements[i] + self.AtomTags[i], Scale*self.Pos[0,i], Scale*self.Pos[1,i], Scale*self.Pos[2,i]))
      return "\n".join(Lines)

   def __len__(self):
      return len(self.Elements)
   def __getitem__(self,key):
      if isinstance(key, slice):
         return [self[o] for o in range(*key.indices(len(self)))]
      elif isinstance(key, int):
         if key < 0:
            key += len(self)
         if key < 0 or key >= len(self):
            raise IndexError("FXyzFrame: index {} is out of range. Have indices [0, {}).".format(key, len(self)))
         return FXyzAtom(self.Elements[key], self.AtomTags[key], self.Pos[:,key], key)
      else:
         raise TypeError("FXyzFrame[..]: Index type must be slice or integer, but got '{}' (type: {})".format(key, type(key)))
   def __iter__(self):
      for (iAt,(Type,Tag,Xyz)) in enumerate(zip(self.Elements, self.AtomTags, self.Pos.T)):
         yield FXyzAtom(Type, Tag, Xyz, iAt)


class EXyzError(Exception):
   def __init__(self, iLine, Message, FileName = None):
      self.FileName = FileName
      self.iLine = iLine
      self.Message = Message
   def __str__(self):
      if self.FileName:
         s = "{}:{}: ".format(self.FileName, self.iLine+1)
      else:
         s = "while reading .xyz file line {}: ".format(self.iLine+1)
      return s + self.Message


def ReadXyzFrames(FileName, InputScale):
   """read all geometries from a .xyz frame. Return a list of FXyzFrame objects.
   If given, all .xyz positions will be multiplied by InputScale. E.g., if
   InputScale = 1./ToAng, coordinates will be converted to atomic units."""
   try:
      # strips comments and leaves non-empty lines.
      Lines = ReadXyzLines(FileName)

      iFrame = 0
      XyzFrames = []
      while len(Lines) >= 2:
         if 0:
            print("ReadXyzFrames: {} lines remain.".format(len(Lines)))
            print("Upcoming lines at start of stack:")
            nLinesToPrint = 5
            for (iLine, Line) in Lines[:min(nLinesToPrint, len(Lines))]:
               print('{:4} {}'.format(iLine,Line))
            if len(Lines) > nLinesToPrint:
               print('{:>4} {}'.format('+','...'))
         Caption, Data, RemainingLines = ReadXyzFrameData(Lines)
         XyzFrames.append(FXyzFrame(Caption = Caption, Data = Data,
            InputScale = InputScale,
            SourceFile = FileName, SourceFrame = iFrame))
         
         Lines = RemainingLines
         iFrame += 1
      return XyzFrames
   except EXyzError, e:
      raise EXyzError(e.iLine, e.Message, FileName)
   pass

def ReadXyzFrame(FileName, InputScale):
   """As ReadXyzFrames, but only return first frame."""
   XyzFrames = ReadXyzFrames(FileName, InputScale)
   return XyzFrames[0]



def ElementNameDummy():
   ElementNames = "X H He Li Be B C N O F Ne Na Mg Al Si P S Cl Ar K Ca Sc Ti V Cr Mn Fe Co Ni Cu Zn Ga Ge As Se Br Kr Rb Sr Y Zr Nb Mo Tc Ru Rh Pd Ag Cd In Sn Sb Te I Xe Cs Ba La Ce Pr Nd Pm Sm Eu Gd Tb Dy Ho Er Tm Yb Lu Hf Ta W Re Os Ir Pt Au Hg Tl Pb Bi Po At Rn".split()
   ElementNumbers = dict([(o,i) for (i,o) in enumerate(ElementNames)])
   return ElementNames, ElementNumbers
ElementNames, ElementNumbers = ElementNameDummy()


def NormalizeElementName(ElementName):
   assert(type(ElementName) is str)
   return ElementName[0].upper() + ElementName[1:].lower()


def SplitAtomElementTag(ElementName):
   # in some .xyz files, we may have multiple differently-tagged atoms of the
   # same element type. E.g. 'Au1' with one type of basis, and 'Au2' with
   # another type of basis. This function is supposed to split such entries
   # into the raw element name (letters only) and the rest (everything following
   # after the initial letters)
   iSplit = 0
   for c in ElementName:
      if not c.isalpha():
         break
      iSplit += 1
   return ElementName[:iSplit], ElementName[iSplit:]



def ReadXyzLines(FileName):
   with open(FileName, "r") as File:
      Lines = File.read().splitlines()
   OutputLines = []
   for (iLine, Line) in enumerate(Lines):
      # skip comments at end of lines and empty lines (neither strictly
      # following .xyz file format, but useful when computing stuff)
      if "#" in Line:
         Line = Line[:Line.find("#")]
      if "!" in Line:
         Line = Line[:Line.find("!")]
      Line = Line.strip()
      if not Line:
         continue
      OutputLines.append(FInputLine(iLine, Line))
   return OutputLines


def ReadXyzFrameData(Lines):
   try:
      nAtomsExpected = int(Lines[0].s)
   except ValueError,e:
      raise EXyzError(0, "Expected integer (number of atoms), but got '{}'".format(Lines[0].s))
      
      
   if len(Lines)+2 < nAtomsExpected:
      raise EXyzError(2, "expected >= {} atom lines, but only got {}.".format(nAtomsExpected, len(Lines)-2))

   DataLines = []
   for (iLine, Line) in Lines[2:]:
      DataLines.append(Line.split())

   # count number of columns in each row. Should be >= 4 and be consistent for all atoms.
   nColSet = set([len(dl) for dl in DataLines])
   if len(nColSet) != 1:
      raise EXyzError(2, ".xyz data misformed: need constant number of columns in every .xyz data row. But got: {}".format(str(nColSet)))
   if min(nColSet) < 4:
      raise EXyzError(2, ".xyz data misformed: need at least 4 data columns (element,x,y,z) in every .xyz data row. But got: {}".format(min(nColSet)))
   
   Caption = Lines[1].s
   Data = np.array(DataLines).T
   
   RemainingLines = Lines[2 + len(DataLines):]
   return Caption, Data, RemainingLines
