# Copyright (c) 2012  Gerald Knizia
# 
# This file is part of the IR/WMME program
# (See https://sites.psu.edu/knizia/)
# 
# IR/WMME is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
# 
# IR/WMME is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/


#
# This version of "rhf_in_python.py" uses the wmme module to write
# inputs and read outputs of the ir/wmme executable, instead of calling
# the program directly.
#
# One way of invoking wmme.py is to leave wmme.py in the wmme binary's directory
# (it searches for its own module directory for the wmme executable), and invoke
# modules using wmme via
#
#   export PYTHONPATH=<path_to_wmme.py>:$PYTHONPATH
#   python rhf_in_python_via_wmme_module.py
#
#

from __future__ import print_function
import numpy as np
import numpy.linalg as la
from diis import FDiisContext
from time import time
import wmme
from wmme import mdot

np.set_printoptions(precision=8,linewidth=10060,suppress=True,threshold=np.nan)

pResultFmt = " %-32s%18.12f"
pResultFmtAnnoted = " %-32s%18.12f  (%s)"
pResultFmtS = " %-32s%18s"
pResultFmtI = " %-32s%5i"
pResultFmtIAnnoted = " %-32s%5i  (%s)"

THRORB = 1e-6       # threshold for orbital gradient in RHF
THRDEN = 1e-8       # threshold for change in energy in RHF
MAXIT_HF = 1002     # maximum num. iterations for RHF
ORB_OCC = 2.

def dot2(A,B): return np.dot(A.flatten(),B.flatten())


def MakeFockOp(Rdm, CoreH, Int2e_Frs, Orb=None, ExchFactor=1.):
   """make closed-shell/spin ensemble Fock matrix:
         f[rs] = h[rs] + j[rs] - .5 k[rs].
   where j[rs] = (rs|tu) rdm[tu]
   and   k[rs] = (rs|tu) rdm[su]
   Return (f, j, k)."""
   if Orb is None:
      OccNum,Orb = la.eigh(-Rdm); OccNum *= -1.
   else:
      OccNum = np.diag(mdot(Orb.T, Rdm, Orb)).copy()

   # filter out occupation numbers slightly broken due to rounding.
   OccNum[OccNum < 0.] = 0.
   OccNum[OccNum > ORB_OCC] = ORB_OCC

   nFit, nOrb = Int2e_Frs.shape[0], Int2e_Frs.shape[1]

   # make coulomb.
   gamma = np.einsum("Frs,rs",Int2e_Frs, Rdm)
   j = np.einsum("Frs,F",Int2e_Frs,gamma)

   # make exchange -- first need to find the occupied orbitals.
   # (they can be non-standard, e.g., in atomic density initial guess)
   nOcc = len([o for o in OccNum if abs(o-ORB_OCC)<1e-8])
   OccOrb = Orb[:,:nOcc] * OccNum[:nOcc]**.5
   k1 = np.einsum("Frs,si->Fir", Int2e_Frs, OccOrb).reshape((nFit * nOcc, nOrb))
   k = np.dot(k1.T, k1)

   Fock = CoreH + j - (ExchFactor/ORB_OCC) * k
   return Fock, j, k


def RunHf(CoreEnergy, CoreH, Int2e_Frs, nOcc, Print):
   p = Print
   StartTime = time()
   nOrb = CoreH.shape[0]
   # coreH guess
   Fock = 1.*CoreH

   LastEnergy = 0.
   if p:
      print("*** HARTREE-FOCK\n")
      print(" %-35s%5i FUNCTIONS" % ("ORBITAL BASIS:", nOrb))
      print(" %-35s%5i FUNCTIONS" % ("FITTING BASIS:", Int2e_Frs.shape[0]))
      print()

   Converged = False
   dc = FDiisContext(12)
   if p: print("   {0:^5s} {1:^14s} {2:^14s} {3:^11s} {4:>4s}  {5:^11s}".format(
      "ITER.","ENERGY","ENERGY CHANGE", "GRAD", "DIIS","TIME"))
   LastRdm = None
   for it in range(MAXIT_HF):
      # make new orbitals and new rdm.
      Ew,Orbs = la.eigh(Fock)

      Rdm = ORB_OCC * np.dot(Orbs[:,:nOcc], Orbs[:,:nOcc].T)

      # make closed-shell Fock matrix:
      #   f[rs] = h[rs] + j[rs] - .5 k[rs]
      # and HF electronic energy:
      #   E = dot(rdm,.5*(h+f))
      Fock, j, k = MakeFockOp(Rdm, CoreH, Int2e_Frs, Orb=Orbs)
      Energy = dot2(.5*Fock+.5*CoreH, Rdm) + CoreEnergy

      # calculate orbital gradient and energy
      OrbGrad = np.dot(Fock,Rdm)
      OrbGrad = OrbGrad - OrbGrad.T

      fOrbGrad = la.norm(OrbGrad.flatten())
      dEnergy = Energy - LastEnergy
      LastEnergy = Energy

      if p: print(" {0:5d} {1:14.8f} {2:+14.8f} {3:11.2e} {4:>6s} {5:8.2f}".format(
         it+1, Energy, dEnergy, fOrbGrad, dc, time() - StartTime))

      if (fOrbGrad < THRORB and abs(dEnergy) < THRDEN):
         Converged = True
         break
      if (it == MAXIT_HF - 1):
         break # not converged.

      Fock, OrbGrad, c0 = dc.Apply(Fock, OrbGrad)

   if (not Converged and MAXIT_HF != 1):
      ErrMsg = "%s failed to converge."\
               "  NIT =%4i  GRAD=%8.2e  DEN=%8.2e" % (WF_TYPE, it+1, fOrbGrad, dEnergy)
      raise Exception(ErrMsg)
   elif p:
      print()
      print(pResultFmt % ("1e energy", dot2(Rdm, CoreH)))
      print(pResultFmt % ("2e energy", .5*dot2(Rdm, Fock - CoreH)))
      if (CoreEnergy != 0.0):
         print()
         print(pResultFmt % ("Core energy", CoreEnergy))
         print(pResultFmt % ("Electronic energy", Energy - CoreEnergy))
      print(pResultFmt % ("Hartree-Fock energy", Energy))
      print()

   return Energy, Rdm

def TestWmmeModule():
   FileName = "benzene.xyz"
   FileName = "acrylic_acid.xyz"
   Atoms = wmme.ReadAtomsFromXyzFile(FileName)

   OrbBasisName = "cc-pVDZ"
   FitBasisName = "univ-JKFIT"
   print("*** SYSTEM '%s' [%s/%s]\n" % (FileName, OrbBasisName, FitBasisName))
   BasisLibFiles = ["def2-nzvpp-jkfit.libmol", "minao.libmol", "emsl_cc-pVDZ.libmol"]
   if 0:
      # normally you would give wmme just names of basis sets, and the basis
      # set libraries to load them from.
      OrbBasis = OrbBasisName
      FitBasis = FitBasisName
      ic = wmme.FIntegralContext(Atoms, OrbBasis, FitBasis, BasisLibFiles)
      print(" Read basis libraries via wmme.\n")
   else:
      # alternatively you can specify a list of explicit basis function declarations
      # (i.e., primitive exponents, contraction coefficients, etc).
      # In the current case, the basis function declarations are read from a Molpro
      # basis set library (in python) and then explicitly passed to wmme, which
      # then does *not* read them from library files anymore.
      import MolproBasisLib as mbl
      from os import path
      # read basis library files
      NamedBases = []
      for LibFile in BasisLibFiles:
         NamedBases += mbl.ReadMolproBasisLibrary(path.join("bases",LibFile))
      BasisLib = mbl.FBasisLibrary(NamedBases)
      # construct basis sets as wmme.FBasisSet objects (effectively, list of FBasisShell1 objects)
      OrbBasis = mbl.MakeBasisForAtoms(Atoms, OrbBasisName, BasisLib)
      FitBasis = mbl.MakeBasisForAtoms(Atoms, FitBasisName, BasisLib)
      ic = wmme.FIntegralContext(Atoms, OrbBasis, FitBasis, BasisLibs=[])
      print(" Read basis libraries in python.\n")

   Smh = True
   CoreEnergy, CoreH, Int2e = ic.MakeBaseIntegrals(Smh=Smh, MakeS=False)


   HfEnergy, HfRdm = RunHf(CoreEnergy, CoreH, Int2e, nOcc=Atoms.nElecNeutral()/2, Print=True)

   if 1:
      # compute dipole moments. note: these DO NOT have the sign of the
      # electron charge absorbed. This is why there is a -1. here.
      DipN = ic.MakeDipoleIntegrals(Smh=Smh)
      ElectronCharge = -1.
      DipMom = ElectronCharge * np.einsum("MNi,MN->i", DipN, HfRdm)
      for ixyz in range(3):
         print(pResultFmt % ("Dipole moment [%s/a.u.]" % "xyz"[ixyz], DipMom[ixyz]))



if ( __name__ == "__main__" ):
   TestWmmeModule()




