/* Copyright (c) 2012  Gerald Knizia
 * 
 * This file is part of the IR/WMME program
 * (See https://sites.psu.edu/knizia/)
 * 
 * IR/WMME is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * IR/WMME is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/
 */

#include <stdexcept>
#include <cmath>
#include <cctype> // for std::isdigit
#include <set>
// #include <iostream> // FIXME: remove this.

#include "CtAtomSet.h"
#include "CtIo.h"
#include "CtMatrix.h"
#include "CtBasisSet.h"
#include "CtConstants.h"
#include "CtBasisLibrary.h"
#include "CxParse1.h"

#ifdef USE_CTINT1E_H
   #include "CtInt1e.h"
#endif

#ifdef _MSC_VER
	#define strcasecmp _stricmp    // both are non-standard functions for case-insensitive string comparisons.
#endif

namespace ct {

double
   SymmetryTolerance = 1e-10;
double const
   g_XyzDistanceScale = ToAng,
   g_XyzGradientScale = -1. * (ToEv/ToAng); // we store negative gradients, in eV/A.


FScalar
   // atomic positions will be multiplied with this value when FAtom(pos)
   // is called. This may be useful to convert A to a.u. or reverse.
   g_AtomPositionInputFactor = 1.0;


std::ostream &operator << ( std::ostream &out, FVector3 const &Pos ){
   out << fmt::format("({:7.5f} {:7.5f} {:7.5f})", Pos[0], Pos[1], Pos[2]);
//    out << format("(%16.9f %16.9f %16.9f)") %  Pos[0] % Pos[1] % Pos[2];
   return out;
}


FAtomSet::FAtomSet()
{
   m_LastEnergy = 0.;
}

FAtomSet::~FAtomSet()
{}


FScalar FAtomSet::NuclearRepulsionEnergy() const
{
   FScalar
      fEnergy = 0;
   FAtomList::const_iterator
      itAtom1, itAtom2;
   for ( itAtom1 = Atoms.begin(); itAtom1 != Atoms.end(); ++itAtom1 )
      for ( itAtom2 = itAtom1, ++itAtom2; itAtom2 != Atoms.end(); ++itAtom2 )
         fEnergy += itAtom1->Charge * itAtom2->Charge *
                  1.0/std::sqrt(DistSq(itAtom1->vPos, itAtom2->vPos));
   return fEnergy;
};

FVector3 FAtomSet::NuclearDipoleMoment( FVector3 const &ExpansionPoint ) const
{
   FVector3
      Result(0,0,0);
   FAtomList::const_iterator
      itAtom;
   for ( itAtom = Atoms.begin(); itAtom != Atoms.end(); ++itAtom ){
      Result += static_cast<FScalar>( itAtom->Charge ) *
         ( itAtom->vPos - ExpansionPoint );
   }
   return Result;
}



// ecp-reduced total nuclear charge.
int FAtomSet::NuclearCharge() const
{
   int
      Result = 0;
   FAtomList::const_iterator
      itAtom;
   for ( itAtom = Atoms.begin(); itAtom != Atoms.end(); ++itAtom )
      Result += itAtom->Charge;
   return Result;
}


void FAtomSet::AddAtom( FAtom const &Atom )
{
   Atoms.push_back( Atom );
}

// converts a map with string keys into another map of the same format but
// which's keys have been lowercased.
template<class FValue>
std::map<std::string,FValue> MakeStringKeysLowercaseInMap(
   std::map<std::string,FValue> In )
{
   std::map<std::string,FValue>
      Out;
   typename std::map<std::string,FValue>::iterator
      it;
   _for_each( it, In )
      Out[tolower(it->first)] = it->second;
   return Out;
}


FXyzLoadException::FXyzLoadException(std::string const &Reason, std::string const &FileName)
   : std::runtime_error(fmt::format("Failed to load '{}' as .xyz: {}",FileName, Reason))
{}

FXyzLoadException::FXyzLoadException(std::string const &Reason, std::string const &CurLine, std::string const &FileName)
   : std::runtime_error(fmt::format("Failed to load '{}' as .xyz: {} (offending line: '{}')",FileName, Reason, CurLine))
{}


// Adds the molecules out of the Rasmol XYZ file to the current atom set.
// Atoms will be added with basis string StandardBasis unless some other
// basis is specified in *pOtherOptions.
void FAtomSet::AddAtomsFromXyzFile( std::string const &FileName,
   FBasisDescs const &DefaultBases,
   FXyzLoadOptions const *pOtherOptions )
{
   TArray<char>
      pFileContent;
   if ( false == LoadFileIntoMemory( pFileContent, FileName ) )
      throw FXyzLoadException("Could not read file from disk", FileName);
   std::stringstream
      str( &pFileContent[0], std::stringstream::in );
   AddAtomsFromXyzStream(str, FileName, DefaultBases, pOtherOptions);
   SetName(FileName);
}

void FAtomSet::AddAtomsFromXyzStream( std::istream &str, std::string const &FileName,
   FBasisDescs const &DefaultBases,
   FXyzLoadOptions const *pOtherOptions )
{
   // the .xyz file format is quite popular because it is extremely simple.
   // format is as follows:
   // - first line: number of atoms =: N
   // - second line: some title or information
   // - N x [Element Name Xcoord Ycoord Zcoord]. Elements may be separated
   //   by whitespace. Following the lines some other data may or may not stand.
   //
   // In practice it turned out to be a bit more messy...
   FXyzLoadOptions
      LoadOptions;
   if ( 0 != pOtherOptions ){
      LoadOptions = *pOtherOptions;
      LoadOptions.ElementBases = MakeStringKeysLowercaseInMap( pOtherOptions->ElementBases );
      LoadOptions.ElementEcps = MakeStringKeysLowercaseInMap( pOtherOptions->ElementEcps );
   }

   bool
      bNumAtomsSet = false,
      bProcessCurrentLine = false;
   uint
      nAtoms = 0;
   std::string
      CurLine;
//    xout << "PROCESS FRAME: Cl0 = '" << CurLine << "'"<< std::endl;
   // read first line. Should be number of atoms, OR, in some cases, the first
   // line of the xyz file itself (in this case there is no pre-determined number
   // of atoms and no comment line).
   while (str.good()) {
      if (std::getline(str, CurLine)) {
         Trim(CurLine);
         // ignore empty lines.
         if (!CurLine.empty())
            break;
      } else {
         if (str.eof())
            // encountered only empty lines while scanning for the header...
            // that means the file is empty.
            return;
         throw FXyzLoadException("Encountered stream error while looking for header line", FileName);
      }
   }
//    xout << "PROCESS FRAME: Cl1 = '" << CurLine << "'"<< std::endl;
   {
      std::stringstream ss(CurLine);
      ss >> nAtoms;
      if (!ss.fail()) {
         // okay, it starts with an integer, apparently.
         bNumAtomsSet = true;

         // read comment line.
         std::getline(str, CurLine);
         // check if there is anything in there which looks like
         // an energy. We take the first thing which cleanly converts
         // to double (or would it be better to take the last?).
         std::stringstream ss2(CurLine);
         SetCaption(CurLine);
         std::string MaybeEnergy;
         while (ss2.good()) {
            // this is indeed the most compact C++ way I could come up with
            // for checking if something "cleanly" converts to 'double'...
            // (and even this is not really clean, because strtod uses "errno",
            // which is not threadsafe).
            ss2 >> MaybeEnergy;
            char const *pBeg = MaybeEnergy.c_str();
            char *pEnd;
            double E = std::strtod(pBeg, &pEnd);
            if (pEnd - pBeg == ptrdiff_t(MaybeEnergy.size())) {
               SetLastEnergy(E);
               break;
            }
         }
      } else {
         // doesn't start with an integer. So guess this line is part of
         // the data already. Remember to not discard the line which we
         // already read.
         bProcessCurrentLine = true;
      }
   }
   bool
      GradSet = false;

//    xout << "PROCESS FRAME: nAtSet? " << bNumAtomsSet << "  nAt = " << nAtoms << std::endl;
   for ( uint iAt = 0; !bNumAtomsSet || iAt < nAtoms; ){
      bool GetLineWorked = true;
      if (bProcessCurrentLine) {
         // first line --- process the line in CurLine which we already loaded
         // for putatively dealing with the file header (which turned out to be
         // not actually there)
         bProcessCurrentLine = false;
      } else {
         // read next data line
         GetLineWorked = bool(std::getline(str, CurLine));
         Trim(CurLine);
      }
//       if (str.bad()) {
      if (!GetLineWorked) {
         if (bNumAtomsSet)
            // if a given number of atoms was set than there'd
            // better be enough of them.
            throw FXyzLoadException("There are fewer atom data lines than were declared.", FileName);
         break;
      }

//       xout << "L: '" << CurLine << "'" << std::endl;
      // we put in some lentinence: we will ignore empty lines and lines
      // starting with typical comment characters (all of which would be invalid
      // in the data body of a regular .xyz file).
      StripLineComment(CurLine, "#");
      StripLineComment(CurLine, "//");
      if (CurLine.empty())
         continue;
//       xout << "  >'" << CurLine << "'" << std::endl;

      // So... let's interpret it as actual data.
      std::stringstream
         ss(CurLine);
      std::string
         Element;
      FScalar
         x, y, z;
      ss >> Element >> x >> y >> z;
      if ( ss.bad() ) {
         // if number of atoms was set in the file, then there'd better
         // be the right number of atoms in there.
         if (bNumAtomsSet)
            throw FXyzLoadException("Data line not understood (should be: <Element> <x> <y> <z> [...])", CurLine, FileName);
         // if the number was not set, we just reached EOL or some other
         // stuff and just quit here.
         break;
      }

      FScalar
         q, gx, gy, gz;
      ss >> q >> gx >> gy >> gz; // check if there is some other stuff (mysterious double, gradient)
      if (iAt == 0 && !ss.bad())
         GradSet = true; // so we have some gradient information.
      if (iAt != 0 && (!ss.bad() != GradSet))
         throw FXyzLoadException("Data line inconsistent: Some data lines contain gradient info, and some do not.", CurLine, FileName);

//       xout << "LINE: " << CurLine << "  GradSet: " << GradSet << "  Grad:" << gx << " " << gy << " " << gz << std::endl;

      Element = tolower(Element);
      // in some formats elements are marked with numbers in order to distinguish
      // different versions of them (e.g., there might be symbols C1 and C2 for
      // carbons using differnt basis sets). We cannot deal with this here, so we
      // simply strip off any digits which might be in there.
      while (!Element.empty() && std::isdigit(Element[Element.size()-1]))
         Element.resize(Element.size() - 1);


/*      FXyzLoadOptions::FElementBasesMap::iterator
         itBasis, itEcp;
      itBasis = LoadOptions.ElementBases.find( Element );
      if ( itBasis != LoadOptions.ElementBases.end() )
         Basis = itBasis->second;
      itEcp = LoadOptions.ElementBases.find( Element );
      if ( itEcp != LoadOptions.ElementBases.end() )
         Ecp = itEcp->second;*/

      FScalar
         // our internal data structures are actually input in a_bohrs.
         f = LoadOptions.InputCoordsToAngstromFactor/ToAng;
      FVector3
         vPos( f*x, f*y, f*z );
      try {
         FAtom
            At(vPos, Element, DefaultBases);
         if (GradSet)
            At.vGrad = (1./g_XyzGradientScale) * FVector3(gx,gy,gz);
         AddAtom(At);
      } catch (std::runtime_error &e) {
         throw FXyzLoadException(e.what(), CurLine, FileName);
      }
      ++ iAt;
//       str.ignore( 0xffff, '\n');
   }
//    xout << "PROCESS FRAME: -> Out" << std::endl;
}

void LoadMultiXyz(FAtomSetList &Frames, std::string const &FileName_, FBasisDescs const &DefaultBases, FAtomSet::FXyzLoadOptions const *pOtherOptions)
{
   bool
      LoadAllFrames = true;
   std::string
      FileName = FileName_;
   string_slice
      slFileName(FileName.begin(), FileName.end());

   // check if we were asked to only load a subset of the input frames.
   typedef std::set<ptrdiff_t>
      FFrameIdSet;
   FFrameIdSet
      iFramesToAdd;
   if (1) {
      // if given, extract the specified frame ID.
      // This would be specified as, for example,
      //
      //    statpt.xyz:123
      //
      // or
      //
      //    statpt.xyz:[0,1,2]
      //
      // where the part before ':' denotes the xyz file and
      // the part after ':' the frame index/indices (starting at 0).

      string_slice::FPair
         slp1 = slFileName.rsplit1(':');
//       std::cout << fmt::format(":slp1 '{}' '{}'", slp1.first.to_str(), slp1.second.to_str()) << std::endl;
      string_slice
         slFrameIds = slp1.second;
      slFrameIds.trim();
      if (!slFrameIds.empty() && slFrameIds.find_c(is_path_separator_c()) == slFrameIds.last) {
         // is this a list of frames?
         if (slFrameIds.startswith('[')) {
            long_split_result
               srFrameIdList;
            slFrameIds.split_list(srFrameIdList);
            for (size_t iSrFrameId = 0; iSrFrameId != srFrameIdList.size(); ++ iSrFrameId) {
               iFramesToAdd.insert(srFrameIdList[iSrFrameId].to_int());
            }
         } else {
            // nope. Assume just a single integer representing the frame ID..
            iFramesToAdd.insert(slFrameIds.to_int());
         }
         LoadAllFrames = false;
         FileName = slp1.first.to_str();
      } else {
         // do nothing. ':' is not at the end of the string.
         // Might be a directory separator or something else.
      }
   }


   TArray<char>
      pFileContent;
   if (false == LoadFileIntoMemory(pFileContent, FileName))
      throw FXyzLoadException("Could not read file from disk", FileName);
   std::stringstream
      str( &pFileContent[0], std::stringstream::in );
   size_t
      iFrame = 0;
   while (str.good()) {
      FAtomSetPtr
         pAtoms(new FAtomSet());
      pAtoms->AddAtomsFromXyzStream(str, FileName, DefaultBases, pOtherOptions);
      pAtoms->SetName(fmt::format("{}:{}", FileName, iFrame));

      if (LoadAllFrames || iFramesToAdd.find(iFrame) != iFramesToAdd.end())
      {
         iFramesToAdd.erase(iFrame);
         if (!pAtoms->empty())
            Frames.push_back(pAtoms);
      }
      iFrame += 1;
   }

   if (!iFramesToAdd.empty()) {
      fmt::MemoryWriter w;
//       w << fmt::format("LoadMultiXyz: While processing input '{}': ", FileName_);
      w << "Failed to load frame(s) ";
      for (FFrameIdSet::const_iterator itLostFrame = iFramesToAdd.begin(); itLostFrame != iFramesToAdd.end(); ++ itLostFrame) {
         if (itLostFrame != iFramesToAdd.begin())
            w << ", ";
         w << *itLostFrame;
      }
      w << " because input .xyz file contains only " << iFrame << " frames total.";
      w << " (note: frame indices are 0-based!)";
      throw FXyzLoadException(w.str(), FileName_);
   }
}

double FAtomSet::GetTotalGradient() const
{
   double g = 0;
   for (size_t iAt = 0; iAt < Atoms.size(); ++ iAt)
      g += LengthSq(Atoms[iAt].vGrad);
   return g;
}

bool FAtomSet::HaveGradients() const
{
   for (size_t iAt = 0; iAt < size(); ++ iAt)
      if (Atoms[iAt].vGrad != FVector3(0., 0., 0.))
         return true;
   return false;
}

void FAtomSet::PrintAsXyz(std::ostream &out, std::string const &Lead, double DistanceFactor, double GradientFactor, bool IncludeGradients)
{
   if (!HaveGradients())
      IncludeGradients = false;

   fmt::MemoryWriter w;
   w.write("E = {:18.12f}", GetLastEnergy());
   std::string Desc = w.str();
   out << fmt::format("{}{}\n{}{}\n", Lead, size(), Lead, Desc);
   for (size_t i = 0; i < size(); ++i) {
      FAtom const
         &At = Atoms[i];
      out << fmt::format("{}{:<4s} {:16.10f} {:16.10f} {:16.10f}",
            Lead, At.GetElementName(), (DistanceFactor * At.vPos[0]), (DistanceFactor*At.vPos[1]), (DistanceFactor*At.vPos[2]));
      if (IncludeGradients)
         out << fmt::format("  0.00 {:12.7f} {:12.7f} {:12.7f}",
            (At.vGrad[0]*GradientFactor), (At.vGrad[1]*GradientFactor), (At.vGrad[2]*GradientFactor));
      out << "\n";
   }
   out.flush();
}


// return the expected number of core electrons for an atom of the given element.
// If ECP is assigned, this number is reduced by the number of ECP electrons.
size_t FAtom::nCoreElectrons() const
{
   static size_t const
      s_nRareGasAtoms = 7,
      s_RareGasAtomicNumbers[] = { 0, 2, 10, 18, 36, 54, 86 };
   size_t
      nCoreElectrons1 = 0;
   for (size_t i = 0; i < s_nRareGasAtoms; ++ i) {
      if (this->AtomicNumber > s_RareGasAtomicNumbers[i]) {
         nCoreElectrons1 = s_RareGasAtomicNumbers[i] - this->nEcpElectrons;
      } else {
         break;
      }
   }
   return nCoreElectrons1;
}



// returns ecp-reduced number of orbitals in rare gas configurations.
size_t FAtomSet::nCoreElectrons() const
{
   size_t
      Result = 0;
   FAtomList::const_iterator
      itAtom;
   _for_each( itAtom, Atoms )
      Result += itAtom->nCoreElectrons();
   return Result;
}


size_t FAtomSet::nEcpElectrons() const
{
   size_t
      Result = 0;
   FAtomList::const_iterator
      itAtom;
   _for_each( itAtom, Atoms )
      Result += itAtom->nEcpElectrons;
   return Result;
}


FAtomSet::FXyzLoadOptions::FXyzLoadOptions()
    : InputCoordsToAngstromFactor( 1.0 )
{}


std::string FAtom::GetBasisDesc() const
{
   FBasisDescs::const_iterator
      itDesc = BasisDesc.find(BASIS_Orbital);
   assert(itDesc != BasisDesc.end());
   if ((this->EcpDesc.empty() || this->EcpDesc == "(none)") && this->nEcpElectrons == 0)
      return itDesc->second;
   else
      return fmt::format("{}, {}", itDesc->second, EcpDesc, nEcpElectrons);
//       return fmt::format("{}, {} ({} e- in ecp)", itDesc->second, EcpDesc, nEcpElectrons);
}

FAtom::FAtom( FVector3 const &vPos_, std::string const &ElementName_, std::string
   const &BasisDesc_, std::string const &Ecp_, bool AutoFix )
{
   BasisDesc[BASIS_Orbital] = BasisDesc_;
   Init(vPos_, ElementName_, Ecp_, AutoFix);
}

FAtom::FAtom( FVector3 const &vPos_, std::string const &ElementName_, FBasisDescs
         const &BasisDesc_, std::string const &Ecp_, bool AutoFix )
   : BasisDesc(BasisDesc_)
{
   Init(vPos_, ElementName_, Ecp_, AutoFix);
}

static bool _in_range(int i, int iMin, int iMax) { return iMin <= i && i <= iMax; }

static void FixDef2BasisAndEcpDecl(std::string &OrbBasis, std::string &EcpName, uint &nEcpElectrons, int iElement)
{
   // remove def2/dhf prefixes. Will be added back later.
   if (StartsWith(OrbBasis, "def2-")) OrbBasis = OrbBasis.substr(5);
   if (StartsWith(OrbBasis, "dhf-")) OrbBasis = OrbBasis.substr(4);

   if (!(StartsWith(OrbBasis, "SV") || StartsWith(OrbBasis, "TZV") || StartsWith(OrbBasis, "QZV")))
      throw std::runtime_error("logic error: Failed to recognize Weigend/Ahlrichs-style basis type.");
   if (iElement > 86) throw std::runtime_error("sorry, no default element associations beyond element 86 programmed.");

   // replace some def2 basis sets by the updated dhf basis sets, which are associated with MDF ECPs
   // which were not yet available when the original def2 sets were made (see: Weigend, Baldes: J. Chem. Phys. 133, 174102 (2010)).
   // Notes:
   //   - Xe (54) had already MDF ECP in original def2 sets
   //   - La was in def2, but has no MDF ECP as of 2015.
   // With these replacements, all elements in range 36-86, except for Lanthan (which has a def2 set) and the lanthanides (which have no def2 sets)
   // are now associated with MDF ECPs. La is still associated with a MWB ECP.
   if ( _in_range(iElement, 37, 48) ||
       (_in_range(iElement, 53, 80) && iElement != 57 && iElement != 54)) {
      OrbBasis = "dhf-" + OrbBasis;
   } else {
      OrbBasis = "def2-" + OrbBasis;
   }
   if (_in_range(iElement, 72, 86)) { EcpName = "ECP60MDF"; nEcpElectrons = 60; }
   if (_in_range(iElement, 55, 71) && iElement != 57) { EcpName = "ECP46MDF"; nEcpElectrons = 46; }
   if (_in_range(iElement, 57, 57)) { EcpName = "ECP46MWB"; nEcpElectrons = 46; } // La: no MDF ECP there, and no dhf set (as of 2015).
   if (_in_range(iElement, 37, 54)) { EcpName = "ECP28MDF"; nEcpElectrons = 28; }
   // note: There are also ECP10MDFs for elements between K and Kr, but the def2 sets do not use them (yet?).
   // They are all-electron sets for those. If dhf sets for those come up, feel free to contact me (Gerald Knizia)
   // and I will update MicroScf.
}


// "If this element had an ECP, how many electrons would the ECP likely represent?"
static size_t iLikelyEcpCharge(int iElement)
{
   // these are for ECP10MDFs used with vtz-pp. def2 sets are always all-electron for those.
   // And at the time of writing, the PP sets are not actually there for 21-28.
   if (_in_range(iElement,21,36)) return 10;

   // these are def2-nzvpp/dhf-nvzpp associations (mostly ECPxxMDF, but some MWB are left).
   if (_in_range(iElement,1,36)) return 0;
   if (_in_range(iElement,37,54)) return 28;
   if (_in_range(iElement,55,71)) return 46; // I guess? somehow I don't seem to have sets for lanthanides (except lanthan)
   if (_in_range(iElement,72,86)) return 60;

   return 0;
}


void FAtom::Init( FVector3 const &vPos_, std::string const &ElementName_, std::string const &EcpName_, bool AutoFix )
{
   vPos = vPos_;
   AtomicNumber = ElementNumberFromName(ElementName_.c_str());
   nEcpElectrons = 0;
   if ( EcpName_.empty() )
      EcpDesc = "(none)";
   else
      EcpDesc = EcpName_;
   vGrad = FVector3(0.,0.,0.);
   if (AutoFix)
      this->AutoFixBasisAssociations();
   Charge = AtomicNumber - nEcpElectrons;
}



void FAtom::AutoFixBasisAssociations()
{
   // fix up some basis declarations; E.g., replace generic terms like "TZVPP" by their explicit def2-/dhf- basis names,
   // and change MINAO to a suitable minimal basis set compatible with a given ECP.
   std::string OrbBasis = tolower(BasisDesc[BASIS_Orbital]);
   if (StartsWith(OrbBasis, "def2-") || StartsWith(OrbBasis, "dhf-") || StartsWith(OrbBasis, "sv") || StartsWith(OrbBasis, "tzv") || StartsWith(OrbBasis, "qzv"))
      FixDef2BasisAndEcpDecl(BasisDesc[BASIS_Orbital], EcpDesc, nEcpElectrons, AtomicNumber);
   else if (StartsWith(OrbBasis, "cc-p") || StartsWith(OrbBasis, "aug-cc-p") || StartsWith(OrbBasis, "minao")) {
      if (AtomicNumber > 36) {
         nEcpElectrons = iLikelyEcpCharge(AtomicNumber);
         if (!EndsWith(OrbBasis, "-pp"))
            OrbBasis += "-pp";
         EcpDesc = "CC-ECP?-FIXME";
      }
   }


   if (nEcpElectrons != 0 || AtomicNumber > 36) {
//          xout << "ELEMENT: " << ElementName_ << "  nEcpElectrons: " << nEcpElectrons << "  iLikelyEcpCharge: " << iLikelyEcpCharge(AtomicNumber) << "  iElement: " << AtomicNumber << std::endl;
      // patch up MINAO to MINAO-PP.
      if (BasisDesc[BASIS_Guess] == "MINAO") {
//             // is the number of ECP electrons we were given consistent with the VTZ-PP derived MINAO-PP basis sets?
//             if (iLikelyEcpCharge(AtomicNumber) != nEcpElectrons)
//                throw std::runtime_error("Sorry, failed to assign ECP-dependent MINAO basis set. Please specify MINAO set explicitly.");
         // ^- FIXME: put this back in.
         BasisDesc[BASIS_Guess] = "MINAO-PP";
         nEcpElectrons = iLikelyEcpCharge(AtomicNumber);
      }
   }

//       if (1) {
//          std::cout << fmt::format(" *********  {:2}  nEcpElec = {}  OrbBas = {}  MINAO = {}", ElementName_, nEcpElectrons, BasisDesc[BASIS_Orbital], BasisDesc[BASIS_Guess]) << std::endl;
//       }
   Charge = AtomicNumber - nEcpElectrons;
}


void FAtom::AssignBasis(FBasisContext Context, std::string const &BasisName, bool AutoFix)
{
   BasisDesc[Context] = BasisName;
   if (AutoFix)
      AutoFixBasisAssociations();
}



std::ostream& operator << ( std::ostream &out, FAtom const &a )
{
   double const ToAng = 1.;
//    out << boost::format( "%+2s:   (%16.10f,%16.10f,%16.10f)    %s" )
//    out << boost::format( "%+2s:   (%10.6f,%10.6f,%10.6f)    %s" )
//    out << boost::format( "%+2s:    %10.6f %10.6f %10.6f     %s" )
//       % ElementNameFromNumber( a.AtomicNumber )
//       % (ToAng * a.vPos[0])
//       % (ToAng * a.vPos[1])
//       % (ToAng * a.vPos[2])
//       % a.GetBasisDesc();
   out << fmt::format( "{:>2}:    {:10.6f} {:10.6f} {:10.6f}     {}",
      a.GetElementName(),
      (ToAng * a.vPos[0]),
      (ToAng * a.vPos[1]),
      (ToAng * a.vPos[2]),
      a.GetBasisDesc());
   return out;
}

// std::ostream& operator << ( std::ostream &out, FAtom const &a )
// {
//    out << boost::format( "%+2s: (%+.5f,%+.5f,%+.5f) basis: '%s' ecp: '%s'" )
//       % ElementNameFromNumber( a.AtomicNumber )
//       % Distance( a.vPos.x(), false )
//       % Distance( a.vPos.y(), false )
//       % Distance( a.vPos.z(), false )
//       % a.GetBasisDesc() % a.EcpDesc;
//    return out;
// }

// std::ostream& operator << ( std::ostream &out, FAtomSet const &a )
// {
//    out << "ATOMS: " << a.Atoms.size() << " atoms. (distance unit: "
//       << DistanceUnit() << ")\n";
//    for( uint i = 0; i < a.Atoms.size(); ++ i ){
//
//       out << boost::format("%+6d  ") % i << a.Atoms[i] << std::endl;
//    }
//    return out;
// }

std::ostream& operator << ( std::ostream &xout, FAtomSet const &a )
{
   xout << "   ATOM  ELM       POS/X      POS/Y      POS/Z        BASIS\n";
   for( uint i = 0; i < a.Atoms.size(); ++ i ){
      xout << (fmt::format("{:>6}   ", i)) << a.Atoms[i] << std::endl;
   }
   xout << "\n   (" << a.Atoms.size() << " atoms;"
        << " " << a.NuclearCharge() - a.nCoreElectrons() << " valence electrons;";
   {
      size_t nEcpElec = a.nEcpElectrons();
      if (nEcpElec != 0)
         xout << fmt::format(" {} electrons absorbed in ECPs;", nEcpElec);
   }
   xout << " distance unit: " << "A" << ")\n";


   size_t Ops[8], nOps,  Gen[3], nGen;
   a.FindMirrorSymmetryOps(Ops, nOps, Gen, nGen);

   xout << "\n";
   if ( 1 ) {
      // Generators    Point group
      // (null card)    $C_1$ (i.e. no point group symmetry)
      // X (or Y or Z)    $C_s$
      // XY    $C_2$
      // XYZ    $C_i$
      // X,Y    $C_{2v}$
      // XY,Z    $C_{2h}$
      // XZ,YZ    $D_2$
      // X,Y,Z    $D_{2h}$
      //
      // 1 X Y Z XY XZ YZ XYZ
      char const
         *pPointGroupName = "?!";
      if ( nOps == 1 ) pPointGroupName = "C1";
      else if ( nOps == 8 ) pPointGroupName = "D2h";
      else if ( nOps == 2 && Ops[0] == 7 ) pPointGroupName = "Ci";
      else if ( nOps == 2 && (Ops[0] == 1 || Ops[0] == 2 || Ops[0] == 4 ) ) pPointGroupName = "Cs";
      else if ( nOps == 2 && (Ops[0] == 3 || Ops[0] == 6 || Ops[0] == 5 ) ) pPointGroupName = "C2";
      else {
         uint
            n1 = 0,
            n2 = 0;
         for ( uint i = 0; i < nOps; ++ i ) {
            uint
               nBits = ((Ops[i] & 1)>>0) + ((Ops[i] & 2)>>1) + ((Ops[i] & 4)>>2);
            if (nBits == 1) n1 += 1;
            if (nBits == 2) n2 += 1;
         }
         if ( n1 == 2 && n2 == 1 ) pPointGroupName = "C2v"; // id X Y XY
         if ( n1 == 1 && n2 == 1 ) pPointGroupName = "C2h"; // id XY Z XYZ
         if ( n1 == 0 && n2 == 3 ) pPointGroupName = "D2"; // id XZ YZ XY
      }

//       if ( Verbosity >= 0 ) {
      if ( 1 ) {
         xout << (fmt::format(" {:<31}{}", "POINT GROUP:", pPointGroupName)) << std::endl;
         xout << fmt::format(" {:<31}", "SYMMETRY OPS:");
         for ( uint iOp = 0; iOp < nOps; ++ iOp ) {
            if ( iOp != 0 )
               xout << "  ";
            for ( uint i = 0; i < 3; ++ i )
               if ( (Ops[iOp] & (1 << i)) != 0 )
                  xout << "XYZ"[i];
            for ( uint i = 0; i < nGen; ++ i )
               if ( iOp == Gen[i] )
                  xout << "*";
            if ( Ops[iOp] == 0 )
               xout << "id";
         }
         xout << std::endl;
      }
   }

   return xout;
}









#ifdef USE_CTINT1E_H
void FAtomSet::Add1eIntegralMatrix( FMatrixView &Dest,
      FBasisSet const &RowBasis, FBasisSet const &ColBasis,
      FDoublettIntegralFactory &IntFactory, double Factor, FMemoryStack &Mem ) const
{
//    xout << "\n\nFAtomSet::Add1eIntegralMatrix WAS CALLED!!\n\n" << std::endl;
   bool
      MatrixSymmetric = (&RowBasis == &ColBasis);
//    MatrixSymmetric = false; // FIXME: remove this.

   assert( Dest.nRows == RowBasis.nFn() );
   assert( Dest.nCols == ColBasis.nFn() );

   FShellDoublettIntegral
      IntResult;
   size_t
      StartFnA = 0, StartFnB; // starting indices of the functions of the
                        // respective shells.
   for ( size_t nShellA = 0; nShellA < RowBasis.Shells.size(); ++ nShellA ){
      FBasisShell const
         *pShellA = &RowBasis.Shells[nShellA];
      StartFnB = (!MatrixSymmetric)? 0 : StartFnA;
      for ( size_t nShellB = (!MatrixSymmetric)? 0 : nShellA;
         nShellB < ColBasis.Shells.size(); ++ nShellB )
      {
         FBasisShell const
            *pShellB = &ColBasis.Shells[nShellB];
         IntFactory.EvalDoublett( IntResult, pShellA, pShellB, Mem );

         assert( IntResult.nSizeA == pShellA->nFn() );
         assert( IntResult.nSizeB == pShellB->nFn() );
         assert( StartFnA + IntResult.nSizeA <= RowBasis.nFn() );
         assert( StartFnB + IntResult.nSizeB <= ColBasis.nFn() );

         // fill data we gathered into the matrices
         for ( size_t i_ = 0; i_ < IntResult.nSizeA; ++ i_ )
            for ( size_t j_ = 0; j_ < IntResult.nSizeB; ++ j_ )
            {
               size_t
                  i = StartFnA + i_,
                  j = StartFnB + j_;
               FScalar const
                  &r = IntResult.Int( i_, j_ );
               Dest(i,j) += Factor * r;
               if ( MatrixSymmetric && nShellA != nShellB )
                  Dest(j,i) += Factor * r;
            }

         StartFnB += pShellB->nFn();
      }
      StartFnA += pShellA->nFn();
   }
   assert( StartFnA == RowBasis.nFn() );
}

void FAtomSet::Make1eIntegralMatrixMultiComp( double *pDest,
      FBasisSet const &RowBasis, FBasisSet const &ColBasis,
      FDoublettIntegralFactory &IntFactory, double Factor, FMemoryStack &Mem_ ) const
{
   bool
      MatrixSymmetric = (&RowBasis == &ColBasis);
//    MatrixSymmetric = false; // FIXME: remove this.

//    FShellDoublettIntegral
//       IntResult;
   size_t
      nComp = IntFactory.nComp();
   size_t
      nFnBasisA = RowBasis.nFn(), nFnBasisB = ColBasis.nFn();
   {
      FMemoryStackArray MemStacks(Mem_);
      #pragma omp parallel for schedule(dynamic)
      // ^- this doesn't help at all...
//       for ( size_t nShellA = 0; nShellA < RowBasis.Shells.size(); ++ nShellA ){
      for ( int nShellA_ = 0; nShellA_ < (int)RowBasis.Shells.size(); ++ nShellA_ ){
         size_t nShellA = size_t(nShellA_); // <- OMP 2.0 only allows "int" as loop variable, not uint or size_t or ptrdiff_t.
         FShellDoublettIntegral
            IntResult;
         FMemoryStack &Mem = MemStacks.GetStackOfThread();
         FBasisShell const
            *pShellA = &RowBasis.Shells[nShellA];
         for ( size_t nShellB = (!MatrixSymmetric)? 0 : nShellA;
            nShellB < ColBasis.Shells.size(); ++ nShellB )
         {
            FBasisShell const
               *pShellB = &ColBasis.Shells[nShellB];
            IntFactory.EvalDoublett( IntResult, pShellA, pShellB, Mem );

            size_t
               StartFnA = RowBasis.pRawBasis->iFn(nShellA),
               StartFnB = ColBasis.pRawBasis->iFn(nShellB); // starting indices of the functions of the


            // fill data we gathered into the matrices
            for ( size_t iComp = 0; iComp < nComp; ++ iComp) {
               FMatrixView
                  Dest(pDest + nFnBasisA*nFnBasisB * iComp, nFnBasisA, nFnBasisB);
               for ( size_t i_ = 0; i_ < IntResult.nSizeA; ++ i_ ) {
                  for ( size_t j_ = 0; j_ < IntResult.nSizeB; ++ j_ )
                  {
                     size_t
                        i = StartFnA + i_,
                        j = StartFnB + j_;
                     FScalar const
                        &r = IntResult.Int( i_, j_, iComp );
                     Dest(i,j) = Factor * r;
                     if ( MatrixSymmetric && nShellA != nShellB )
                        Dest(j,i) = Factor * r;
                  }
               }
            }

         }
      }
   }
}


#endif // USE_CTINT1E_H

void FAtomSet::MakeCoreHamiltonMatrix( FMatrixView &Out, FBasisSet const &RowBasis, FBasisSet const &ColBasis, FMemoryStack &Mem ) const
{
   Out.Clear();
   AddKineticMatrix(Out, RowBasis, ColBasis, Mem);
   AddNuclearAttractionMatrix(Out, RowBasis, ColBasis, Mem);

   // TODO: apply other CoreH patches from Environment object etc.

//    AddDipoleMatrix( Out, RowBasis, ColBasis, FVector3(0.0,0.0,1e-5), FVector3(0,0,0), Mem );
}


void FAtomSet::AccCoreHamiltonDeriv1( FMatrixView &GradXyz, FMatrixView const &Rdm, FBasisSet const &RowBasis, FBasisSet const &ColBasis, FMemoryStack &Mem ) const
{
   if (GradXyz.nRowSt != 1 || GradXyz.nColSt != 3 || GradXyz.nRows != 3 || GradXyz.nCols != this->size())
      throw std::runtime_error("Output gradient should be matrix of format 3 x nAtoms, without strides.");
   FKrn2iPtr
      pCorePotentialKernel = MakeNuclearPotentialKernel(),
      pKineticEnergyKernel = MakeKineticEnergyKernel();

//    dynamic_cast<FKrn2i_PointMultipoles*>(pCorePotentialKernel.get())->SetDerivativeFlags(FKrn2i_PointMultipoles::DERIVFLAG_IncludePointDerivsP);
   AccGradient2ix(GradXyz.pData, Rdm, *RowBasis.pRawBasis, *ColBasis.pRawBasis, *pCorePotentialKernel, Mem);
   AccGradient2ix(GradXyz.pData, Rdm, *RowBasis.pRawBasis, *ColBasis.pRawBasis, *pKineticEnergyKernel, Mem);
}

void FAtomSet::AccNuclearRepulsionGradient(FMatrixView &GradXyz, FMemoryStack &Mem, double Factor) const
{
   // add derivative of nuclear-nuclear repulsion (i.e., of the contribution of this->NuclearRepulsionEnergy()).
   for (size_t iAtA = 0; iAtA < size(); ++ iAtA)
      for (size_t iAtB = iAtA + 1; iAtB < size(); ++ iAtB) {
         FAtom const
            *pAtA = &Atoms[iAtA],
            *pAtB = &Atoms[iAtB];
         // grad pow(rSqAB, -.5) = -.5 * pow(rSqAB, -1.5) * grad rSqAB
         // grad rSqAB = 2 * DistAB
         FVector3
            vDist = pAtA->vPos - pAtB->vPos;
         double
            rSqAB = vDist.LengthSq();
         double
            f = -.5 * 2 * pAtA->Charge * pAtB->Charge *
                  std::pow(rSqAB, -1.5) * Factor;
         for (uint ixyz = 0; ixyz < 3; ++ ixyz) {
            GradXyz(ixyz, iAtA) += f * vDist[ixyz];
            GradXyz(ixyz, iAtB) -= f * vDist[ixyz];
         }
      }
   IR_SUPPRESS_UNUSED_WARNING(Mem);
}


void FAtomSet::AccNuclearRepulsionHessian(FMatrixView &HessXyz, FMemoryStack &Mem, double Factor) const
{
//    double Z_[] = { -0.25568, -0.20364, 0.49278, -0.45786, -0.47351, 0.18079, 0.16838, 0.16767, 0.38108};
   double Z_[] = {-0.12302, -0.14391, 0.47680, -0.44731, -0.50834, 0.12607, 0.10888, 0.11408, 0.39673};
#define Z(iAt) Z_[iAt]
// #define Z(iAt) Atoms[iAtA].Charge

   // add derivative of nuclear-nuclear repulsion (i.e., of the contribution of this->NuclearRepulsionEnergy()).
   for (size_t iAtA = 0; iAtA < size(); ++ iAtA)
      for (size_t iAtB = iAtA + 1; iAtB < size(); ++ iAtB) {
         FAtom const
            *pAtA = &Atoms[iAtA],
            *pAtB = &Atoms[iAtB];
         // f(r) = pow(rSqAB, -.5)
         // gr = -.5 * pow(rSqAB, -1.5)
         // hr = -1.5 * pow(rSqAB, -2.5)
         //    = 3*gr/rSqAB
         // -> grad grad pow(rSqAB, -.5) = 4*outer(DistAB,DistAB)*hr + 2*Identity(3,3)*gr
         FVector3
            vDist = pAtA->vPos - pAtB->vPos;
         double
            rSqAB = vDist.LengthSq();
         double
//             gr = -0.5 * pAtA->Charge * pAtB->Charge * std::pow(rSqAB, -1.5) * Factor,
            gr = -0.5 * Z(iAtA) * Z(iAtB) * std::pow(rSqAB, -1.5) * Factor,
            hr = 3*gr/rSqAB;
         for (uint ixyzA = 0; ixyzA < 3; ++ ixyzA) {
            for (uint ixyzB = 0; ixyzB < 3; ++ ixyzB) {
               double hAB = 2*gr*((ixyzA==ixyzB)? 1. : 0.) + 4*hr*vDist[ixyzA]*vDist[ixyzB];
               HessXyz(3*iAtA+ixyzA, 3*iAtB+ixyzB) += 2*hAB; // all * 2 for symmetrize.
               HessXyz(3*iAtA+ixyzA, 3*iAtA+ixyzA) -= hAB;
               HessXyz(3*iAtB+ixyzB, 3*iAtB+ixyzB) -= hAB;
//                HessXyz(3*iAtA+ixyzA, 3*iAtB+ixyzB) += hAB;
//                HessXyz(3*iAtB+ixyzB, 3*iAtA+ixyzA) += hAB;
            }
         }
      }
   Symmetrize(HessXyz);
   IR_SUPPRESS_UNUSED_WARNING(Mem);
#undef Z
}




void FAtomSet::MakeOverlapMatrix( FMatrixView &Out, FBasisSet const &RowBasis, FBasisSet const &ColBasis, FMemoryStack &Mem ) const
{
#ifdef USE_CTINT1E_H_ALWAYS
   FDoublettIntegralFactoryOverlap
      IntFactory; // can't bind temporary to non-const reference.
   Out.Clear();
   Add1eIntegralMatrix( Out, RowBasis, ColBasis, IntFactory, 1.0, Mem );
#else
//    MakeIntMatrix(Out, RowBasis, ColBasis, FKrn2i_Direct(&ir::g_IrOverlapKernel,0), Mem, 1.0, false);
   MakeIntMatrix(Out, RowBasis, ColBasis, FKrn2i_Direct(&ir::g_IrOverlapKernel), Mem);
#endif
}

void FAtomSet::AddKineticMatrix( FMatrixView &Out, FBasisSet const &RowBasis, FBasisSet const &ColBasis, FMemoryStack &Mem ) const
{
#ifdef USE_CTINT1E_H_ALWAYS
   FDoublettIntegralFactoryKineticTerm
      IntFactory;
   Out.Clear();
   Add1eIntegralMatrix( Out, RowBasis, ColBasis, IntFactory, 1.0, Mem );
#else
   FKrn2iPtr
      pKineticEnergyKernel = MakeKineticEnergyKernel();
   MakeIntMatrix(Out, RowBasis, ColBasis, *pKineticEnergyKernel, Mem, 1., false);
//    MakeIntMatrix(Out, RowBasis, ColBasis, FKrn2i_Direct(&ir::g_IrOverlapKernel,1), Mem, -.5, false);
#endif
}

FKrn2iPtr FAtomSet::MakeNuclearPotentialKernel() const
{
   // make Coulomb part of core attraction matrix.
   std::vector<FKrn2i_PointMultipoles::FPoint>
      MultiPoles(Atoms.size());
   std::vector<double>
      Coeffs(Atoms.size());
   for (uint iAt = 0; iAt < Atoms.size(); ++ iAt) {
      MultiPoles[iAt].vPos = Atoms[iAt].vPos;
      MultiPoles[iAt].l = 0;
      MultiPoles[iAt].iCenter = static_cast<ptrdiff_t>(iAt);
      Coeffs[iAt] = -Atoms[iAt].Charge;
      // ^- the minus is there because we effectively calculate with
      //    positive electron charge.
   }
   FKrn2iPtr
      pMultipoleKernel(new FKrn2i_PointMultipoles(&ir::g_IrCoulombKernel, &MultiPoles[0], &Coeffs[0], MultiPoles.size()));
   if (!NeedEcpKernel())
      return pMultipoleKernel;
   // ^- FIXME: remove this when other core attraction patches are introduced (e.g., finite fields)

   // add kernels for ECP part of core attraction matrix, if needed. Make one for each ECP'd core
   // for a start (later on, at least the local part of ECPs could be absorbed into meta kernels which
   // are evaluated together with the Coulomb parts. For the non-local parts I am not sure atm.).
   FKrn2iMetaPtr
      pMetaKernel = new FKrn2i_Meta();
   pMetaKernel->AddKernel(pMultipoleKernel, 1.0);

   for (size_t iAt = 0; iAt < Atoms.size(); ++ iAt) {
      FAtom const
         &At = Atoms[iAt];
      if (At.nEcpElectrons != 0) {
#ifdef IR_ECP
         ir::FAtomEcp const
            *pIrEcp = g_BasisSetLibrary.LoadEcp(At.AtomicNumber, At.EcpDesc, At.vPos, int(iAt));
         if (pIrEcp->fElecAbsorbed != double(At.nEcpElectrons))
            throw std::runtime_error("Actual ECP electron absorbance does not match expected one.");
         FKrn2iPtr pEcpKrn = new FKrn2i_1CenterEcp(At.vPos, ptrdiff_t(iAt), pIrEcp);
         pMetaKernel->AddKernel(pEcpKrn, 1.0);
#else
         throw std::runtime_error(fmt::format("Atom {} has declared a ECP{}, but this version of IR was compiled without ECP support.", GetAtomLabel(iAt), At.nEcpElectrons));
#endif // IR_ECP
      }
   }
   return pMetaKernel;
}

FKrn2iPtr FAtomSet::MakeKineticEnergyKernel() const
{
   return FKrn2iPtr(new FKrn2i_Direct(&ir::g_IrOverlapKernel,1, -.5));
}


bool FAtomSet::NeedEcpKernel() const
{
   for (size_t iAt = 0; iAt < Atoms.size(); ++ iAt)
      if (Atoms[iAt].nEcpElectrons != 0)
         return true;
   return false;
}


void FAtomSet::AddNuclearAttractionMatrix( FMatrixView &Out, FBasisSet const &RowBasis, FBasisSet const &ColBasis, FMemoryStack &Mem ) const
{
#ifdef USE_CTINT1E_H_ALWAYS
   if (NeedEcpKernel())
      throw std::runtime_error("Sorry, CtInt1e has no ECP integrals. Cannot make nuclear attraction matrix. ");

//    Out.Clear(); // remove this!
   typedef FDoublettIntegralFactoryFieldTerms
      FNucFactory;
   FNucFactory::FPointChargeList
      PointCharges;
   FAtomSet::FAtomList::const_iterator
      itAtom;
   // make one point charge for every atom. charge: ecp-reduced core charge.
   _for_each( itAtom, Atoms )
      PointCharges.push_back( FNucFactory::FPointCharge(
            itAtom->vPos, static_cast<FScalar>( itAtom->Charge ) ) );
   FNucFactory
      NuclearAttractionIntegrator( TVector3<uint>( 0, 0, 0 ), PointCharges );
   Add1eIntegralMatrix( Out, RowBasis, ColBasis, NuclearAttractionIntegrator, 1.0, Mem );
//    Out.Print(xout, "VNUC/CtInt1e");
#else
//    Out.Clear(); // remove this!
   FKrn2iPtr
      pCorePotentialKernel = FAtomSet::MakeNuclearPotentialKernel();
   MakeIntMatrix(Out, RowBasis, ColBasis, *pCorePotentialKernel, Mem, 1., true);
//    Out.Print(xout, "VNUC/MakeIntMatrix");
//    throw std::runtime_error(":/");
#endif // USE_CTINT1E_H
}


void FAtomSet::AddDipoleMatrix( FMatrixView &Out, FBasisSet const &RowBasis, FBasisSet const &ColBasis,
   FVector3 const &Direction, FVector3 const &ExpansionPoint, FMemoryStack &Mem ) const
{
#ifdef USE_CTINT1E_H
/*   FVector3
      DirNormalized = (-1.0/std::sqrt(Dot(Direction,Direction))) * Direction;*/
      // ^- -1.0: Electrons are very negatively charged. this is actually
      // the charge factor with the elementary charge (which is 1 in a.u.)
   //xout << boost::format( "Dipole Direction: %f %f %f" )
   //   % DirNormalized.x() % DirNormalized.y() % DirNormalized.z() << std::endl;

   //  make dipole matrices into the three cartesian directions.
   for ( uint CartComp = 0; CartComp < 3; ++ CartComp ){
      if ( fabs( Direction[CartComp] ) < 1e-15 )
         continue;

      FDoublettIntegralFactoryMultipoleMoment::FMoment
         CartMoment( 0, 0, 0 );
      // this specifies the derivative power in cart direction Cart.
      CartMoment[CartComp] = 1;

      FDoublettIntegralFactoryMultipoleMoment
         IntFactory( CartMoment, ExpansionPoint );
      // add dot product component of current direction times dipole matrix.
      Add1eIntegralMatrix( Out, RowBasis, ColBasis, IntFactory,
         Direction[CartComp], Mem );
      //xout << "DipoleMomentOp" << CartComp << ":"
      //    << ResultDir[CartComp].View( 0, 7, 0, std::min( 40u, ResultDir[CartComp].SizeY() ) );
   }
#else
   throw std::runtime_error("dipole integrals not implemented here.");
#endif
}


void FAtomSet::MakeCartesianMomentMatrix( FMatrixView &Out, FBasisSet const &RowBasis, FBasisSet const &ColBasis,
   TVector3<unsigned> const &CartMoment_, FVector3 const &ExpansionPoint, FMemoryStack &Mem ) const
{
#ifdef USE_CTINT1E_H
   Out.Clear();

   FDoublettIntegralFactoryMultipoleMoment::FMoment
      CartMoment( CartMoment_[0], CartMoment_[1], CartMoment_[2] );
   // this specifies the derivative power in cart direction Cart.
   FDoublettIntegralFactoryMultipoleMoment
      IntFactory( CartMoment, ExpansionPoint );
   // add to output matrix.
   Add1eIntegralMatrix( Out, RowBasis, ColBasis, IntFactory, 1., Mem );
#else
   throw std::runtime_error("general cartesian moment integrals not implemented here.");
#endif
}


bool IsEquivalent(FAtom const &A, FVector3 const &vA, FAtom const &B)
{
   if ( A.AtomicNumber != B.AtomicNumber )
      return false;
   if ( A.Charge != B.Charge )
      return false;
   if ( DistSq(vA, B.vPos) > sqr(SymmetryTolerance) )
      return false;
   if ( A.BasisDesc != B.BasisDesc )
      return false;
   if ( A.EcpDesc != B.EcpDesc )
      return false;
   return true;
}

FVector3 ApplySymOp( FVector3 const &In, size_t Op ){
   FVector3
      MirrorPos = In;
   for (size_t iXyz = 0; iXyz < 3; ++ iXyz)
      if (Op & (1<<iXyz))
         MirrorPos[iXyz] = -MirrorPos[iXyz];
   return MirrorPos;
}


void FAtomSet::FindEquivalentAtoms(FAtomSymGroup *&pGroups, size_t &nGroups, FMemoryStack &Mem) const
{
   Mem.Alloc(pGroups, Atoms.size()); // actual data will be less.
   bool
      *pDone;
   Mem.ClearAlloc(pDone, Atoms.size());

   size_t Ops[8], nOps, Gen[3], nGen;
   FindMirrorSymmetryOps(Ops, nOps, Gen, nGen);


   FAtomSymGroup
      *pCurGroup = pGroups;
   for ( uint iAt0 = 0; iAt0 < Atoms.size(); ++ iAt0 ){
      if ( pDone[iAt0] )
         continue; // already seen as part of another atom symmetry group.

      pCurGroup->nEqiv = 0;

      // look up other equivalent atoms and add them to the group.
      for (size_t iOp = 0; iOp < nOps; ++ iOp){
         FVector3
            MirrorPos = ApplySymOp(Atoms[iAt0].vPos, Ops[iOp]);
         for (size_t iAt1 = iAt0; iAt1 < Atoms.size(); ++ iAt1)
            if ( !pDone[iAt1] &&
                  IsEquivalent(Atoms[iAt0], MirrorPos, Atoms[iAt1]) ) {
               pCurGroup->iEqiv[pCurGroup->nEqiv] = iAt1;
               pCurGroup->SymOp[pCurGroup->nEqiv] = Ops[iOp];
               pCurGroup->nEqiv += 1;
               assert(pCurGroup->nEqiv <= 8);
               pDone[iAt1] = true;
            }
      }

      // find out directions in which current atoms lie
      // on the mirror plane.
      pCurGroup->Stabilizer = 0;
      for (size_t iXyz = 0; iXyz < 3; ++ iXyz){
         if ( sqr(Atoms[iAt0].vPos[iXyz]) <= SymmetryTolerance )
            pCurGroup->Stabilizer |= (1 << iXyz);
      }

      ++ pCurGroup;
   };
   nGroups = pCurGroup - pGroups;
   assert(nGroups <= Atoms.size());

   Mem.Free(pDone);
   // still on stack: pGroups[1..Atoms.size()].
}



void FAtomSet::FindMirrorSymmetryOps(size_t Ops[8], size_t &nOps, size_t Gen[3], size_t &nGen) const
{
   nOps = 0;

   if (1) {
      // This disables symmetry everywhere.
      nOps = 1;
      nGen = 0;
      Ops[0] = 0;
      nOps = 1;
      return;
   }

   size_t
      AllOps[8] = { 0, 1, 2, 4, 3, 5, 6, 7 }; // simple ones first: id X Y Z XY XZ YZ XYZ

   for (size_t iOp_ = 0; iOp_ < 8; ++ iOp_){
      size_t
         iOp = AllOps[iOp_];
      bool
         Keep = true;

      for (size_t iAt0 = 0; iAt0 < Atoms.size(); ++ iAt0){
         bool
            FoundEquiv = false;
         FVector3
            MirrorPos = ApplySymOp(Atoms[iAt0].vPos, iOp);

         for (size_t iAt1 = 0; iAt1 < Atoms.size(); ++ iAt1)
            if ( IsEquivalent(Atoms[iAt0], MirrorPos, Atoms[iAt1]) ) {
               FoundEquiv = true;
               break;
            }

         // no atom equivalent under symmetry xform found for iAt0.
         if (!FoundEquiv) {
            Keep = false;
            break;
         }
      }

      if (Keep) {
         Ops[nOps] = iOp;
         nOps += 1;
      }
   }

   nGen = 0;
   for (size_t i = 0; i < 3; ++ i)
      Gen[i] = 0;

   for (size_t iBaseOp = 0; iBaseOp < nOps; ++ iBaseOp){
      size_t
         BaseOp = Ops[iBaseOp];
      bool
         Redundant = false || BaseOp == 0;
      // Can the current operator can be formed as some combination
      // of the previous generators?...
      for (size_t iPat = 0; iPat < (1u<<nGen); ++ iPat){
         uint
            TrialOp = 0;
         for (size_t iGen = 0; iGen < nGen; ++ iGen)
            if ( iPat & (1u << iGen) ) {
               // generator iGen activated in trial pattern.
               TrialOp ^= Gen[iGen];
            }
         if ( TrialOp == BaseOp ) {
            Redundant = true;
            break;
         }
      }
      if ( Redundant == true )
         continue;

      // ...nope. Include it as new generator.
      Gen[nGen] = BaseOp;
      nGen += 1;
   }
}

std::string FAtom::GetElementName() const {
   return ElementNameFromNumber(AtomicNumber);
}

std::string FAtom::GetAtomLabel(int iAtom, int iAtomWidth) const {
   char const *pFmt = "{} {:<2}";
   switch (iAtomWidth) {
      case 0: pFmt = "{} {}"; break; // smallest fit for both number and name.
      case 1: pFmt = "{} {:<2}"; break;
      case 2: pFmt = "{:>2} {:<2}"; break;
      case 3: pFmt = "{:>3} {:<2}"; break;
      case 4: pFmt = "{:>4} {:<2}"; break;
      case 5: pFmt = "{:>5} {:<2}"; break;
      default: break;
   }

   return fmt::format(pFmt, iAtom+1, ElementNameFromNumber(AtomicNumber));
}

std::string FAtomSet::GetAtomLabel(int iAtom, int iAtomWidth) const
{
   if (iAtomWidth < 0) {
      // decide on a width of the atom number field which can align all atom indices.
      iAtomWidth = 1;
      if (size() >= 10) iAtomWidth = 2;
      if (size() >= 100) iAtomWidth = 3;
      if (size() >= 1000) iAtomWidth = 4;
      if (size() >= 10000) iAtomWidth = 5;
   }
   if (iAtom < 0 || size_t(iAtom) > size())
      throw std::runtime_error(fmt::format("FAtomSet::GetAtomLabel: Requested label for non-existent atom index {} (nAt = {})", iAtom, size()));
   return Atoms[iAtom].GetAtomLabel(iAtom, iAtomWidth);
}



// fix up phase a vector such it has its largest element (in absolute terms) positive.
static void FixVectorPhase(double *p, size_t n)
{
   if (n == 0)
      return;
   double
      fMax = 0.;
   size_t
      iMax = 0;
   for (size_t i = 0; i < n; ++ i) {
      double fAbs = std::abs(p[i]);
      if (fAbs > fMax) {
         iMax = i;
         fMax = fAbs;
      }
   }
   if (p[iMax] < 0)
      for (size_t i = 0; i < n; ++ i)
         p[i] *= -1;
}


void Transpose3x3(double *R) {
   std::swap(R[0 + 3*1], R[1 + 3*0]);
   std::swap(R[0 + 3*2], R[2 + 3*0]);
   std::swap(R[1 + 3*2], R[2 + 3*1]);
}


void FindAtomSetAligningTrafo(double pR[9], double pD[3], FVector3 const *pAtPos, double const *pAtMass, FVector3 const *pAtPosLast, size_t nAt, FMemoryStack &Mem)
{
#define GET_AT_MASS(iAt) (pAtMass? pAtMass[iAt] : 1.)
   double
      TotalMass = 0,
      pIm[9] = {0}; // inertial tensor.
   FVector3
      vMassCenter(0.,0.,0.);
   // compute total mass and center of mass
   for (size_t iAt = 0; iAt < nAt; ++ iAt) {
      TotalMass += GET_AT_MASS(iAt);
      vMassCenter += GET_AT_MASS(iAt) * pAtPos[iAt];
   }
   vMassCenter /= TotalMass;

   // copy center translation
   for (size_t i = 0; i < 3; ++ i)
      pD[i] = vMassCenter[i];

   if (pAtPosLast == 0) {
      // no reference frame given.

      // compute inertial tensor around the center of mass.
      for (size_t iAt = 0; iAt < nAt; ++ iAt) {
         FVector3 const
            vAtPos = pAtPos[iAt] - vMassCenter;
         double
            Rsq = Dot(vAtPos,vAtPos);
         for (size_t i = 0; i < 3; ++ i)
            for (size_t j = 0; j < 3; ++ j)
               pIm[i + 3*j] += GET_AT_MASS(iAt) * (((i==j)? 1. : 0.)*Rsq - vAtPos[i]*vAtPos[j]);
   //             I += ElementMasses[o.iElement]*(np.eye(3)*np.dot(Xyz,Xyz) - np.outer(Xyz,Xyz))
      }

      // align along axes of inertia--largest moment along z (this way flat
      // molecules will be aligned in the x/y plane).
      for (size_t i = 0; i < 9; ++ i)
         pR[i] = +pIm[i];
      double
         pEw[3];
      FMatrixView
         mR(pR, 3, 3);
      Diagonalize(mR, pEw, Mem);
      // fix up phases of the eigenvectors, such that they have their largest element positive.
      for (size_t iEv = 0; iEv < 3; ++ iEv)
         FixVectorPhase(&mR(0,iEv), 3);

      // check the determinant of the transformation to see if it includes an
      // inversion. If yes, invert one of the axes. We'd get the wrong stero-isomers
      // otherwise...
      if (1) {
         FVector3
            v0(mR(0,0), mR(1,0), mR(2,0)),
            v1(mR(0,1), mR(1,1), mR(2,1)),
            v2(mR(0,2), mR(1,2), mR(2,2));
         if (Dot(v0, Cross(v1,v2)) < 0.) {
            for (size_t i = 0; i < 3; ++ i)
               mR(i,2) *= -1.;
         }
      }
   } else {
      // compute overlap matrix between current and last frame (assuming that the last one
      // has already been moved into its own center of mass)
      for (size_t iAt = 0; iAt < nAt; ++ iAt) {
         FVector3 const
            vAtPosI = pAtPos[iAt] - vMassCenter,
            vAtPosJ = pAtPosLast[iAt];
         for (size_t i = 0; i < 3; ++ i)
            for (size_t j = 0; j < 3; ++ j)
               pIm[i + 3*j] += GET_AT_MASS(iAt) * vAtPosI[i] * vAtPosJ[j];
      }
      double
         pU[9], pVt[9], pSig[3];
      FMatrixView
         mR(pR, 3, 3),
         mU(pU, 3, 3),
         mI(pIm, 3, 3),
         mVt(pVt, 3, 3);
      ComputeSvd(mU, pSig, mVt, mI, Mem);
      Mxm(mR, mU, mVt);
      // ^- I think it should be the other way around, but there is a strange Transpose3x3
      //    in the actual transformation routine...
   }

   bool
      t3x3 = true; // <- I wonder why I put this here.
   if (t3x3)
      Transpose3x3(pR);
#undef GET_AT_MASS
}






} // namespace ct


// kate: space-indent on; tab-indent off; indent-width 3; mixedindent off; indent-mode normal;
