/* Copyright (c) 2012  Gerald Knizia
 * 
 * This file is part of the IR/WMME program
 * (See https://sites.psu.edu/knizia/)
 * 
 * IR/WMME is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * IR/WMME is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/
 */

#ifndef CT8K_ATOMSET_H
#define CT8K_ATOMSET_H

#include <map>
#include <string>
#include <vector>
#include <istream>
#include <stdexcept>

#include "CxTypes.h"
#include "CxVec3.h"
#include "CxMemoryStack.h"
#include "CtBasisSet.h"
#include "CtAtomData.h"
#include "CxPodArray.h"

namespace ct {

// // conversion element name <-> atomic number
// std::string ElementNameFromNumber( uint AtomicNumber );
// uint ElementNumberFromName( std::string const &Name );

typedef TVector3<double>
   FVector3;

typedef std::string
   FBasisDesc;
typedef std::string
   FEcpDesc;
typedef std::map<FBasisContext, FBasisDesc>
   FBasisDescs;

struct FDoublettIntegralFactory;
struct FMatrixView;

extern double const g_XyzDistanceScale; // factor between a.u. atomic positions (bohr) and what goes into .xyz files. Normally Angstroms, but the constant changes every few years...
extern double const g_XyzGradientScale; // factor between a.u. gradients (Hartree/bohr) and what goes into .xyz files. It is not standardized.

struct FAtom
{
   FVector3
      vPos; // center of atom, in a_bohr (always in a_bohr)
   uint
      AtomicNumber; // fixme: rename to iElement
   uint
      // number of electrons implicit due to effective
      // core potential ("pseudopotential")
      nEcpElectrons;
   int
      Charge; // might differ from AtomicNumber if ECPs are used.
   FBasisDescs
      BasisDesc;
   FEcpDesc
      EcpDesc;

   FVector3
      // if set: gradient dE/d[x,y,z] of the atom in a.u. (TODO: should this be here?
      // it is an /input. in certain kinds of xyz files, so in this sense it would be
      // an actual property...)
      vGrad;

   std::string GetBasisDesc() const;
   std::string GetElementName() const;
   std::string GetAtomLabel(int iAtom, int iAtomWidth=3) const; // makes labels like '1 Se' (output atom labels start at 1, i.e., iAtom is increased by 1 for making the label)

   FAtom() {};
   // AutoFix: change/derive some standard parameters based on the input basis name; e.g., choose
   // ECPs or change def2-TZVPP to dhf-TZVPP for suitable elements.
   FAtom( FVector3 const &vPos_, std::string const &ElementName_, std::string
            const &BasisDesc_ = "def2-SVP", std::string const &Ecp_ = "", bool AutoFix = true );
   FAtom( FVector3 const &vPos_, std::string const &ElementName_, FBasisDescs
            const &BasisDesc_, std::string const &Ecp_ = "", bool AutoFix = true );

   inline std::string ElementName() const { return ElementNameFromNumber(AtomicNumber); }

   // assigns a particular basis to the atom, via name, and if AutoFix=true, adjusts relevant
   // variant basis sets (e.g., dhf vs def2 or xx vs xx-PP).
   void AssignBasis(FBasisContext Context, std::string const &BasisName, bool AutoFix = true);

  // return the expected number of core electrons for an atom of the given element.
   // If ECP is assigned, this number is reduced by the number of ECP electrons.
   size_t nCoreElectrons() const;
private:
   void Init( FVector3 const &vPos_, std::string const &ElementName_, std::string const &EcpName_, bool AutoFix );
   void AutoFixBasisAssociations();
};


// describes global effects acting on the current calculation. For example,
// external fields.
struct FEnvironment
{

};

struct FBasisSet;

struct FAtomSymGroup {
   // number of equivalnt atoms
   uint nEqiv;
   // index into Atoms[]
   uint iEqiv[8];
   // symmetry op moving iEqiv[0] to iEqiv[i]
   uint SymOp[8];
   // bit pattern of XYZ symmetry op not affecting the atom positions,
   // since the relevant coordinates are zero.
   uint Stabilizer;
};

// An atom set specifies which atoms are where and which properties these
// atoms do have in calculations (i.e. which basis/ecp to use and so on).
// Additionally,
// So an instance of this class basically describes the problem which
// we currently handle.
struct FAtomSet : public FIntrusivePtrDest
{
public:
   typedef std::vector<FAtom>
      FAtomList;
   FAtomList
      Atoms;
   FEnvironment
      Environment;


   // ecp-reduced nuclear repulsion energy.
   double NuclearRepulsionEnergy() const;
   void   AddAtom( FAtom const &Atom );
   // ecp-reduced total nuclear charge.
   int    NuclearCharge() const;

   FVector3 NuclearDipoleMoment( FVector3 const &ExpansionPoint = FVector3(0,0,0) ) const;

   // returns ecp-reduced number of electrons in rare gas configurations.
   size_t nCoreElectrons() const;
   // returns total number of electrons absorbed in ECPs
   size_t nEcpElectrons() const;

   struct FXyzLoadOptions{
      typedef std::map<std::string,std::string>
         FElementBasesMap;
      FElementBasesMap
         // defines basis strings for elements which are loaded. If an element
         // is not present here, the standard basis string (e.g. "dzp") is
         // assigned for it. Case of element names (keys) don't matter.
         ElementBases,
         ElementEcps;
      double
         // if the file coordinates are not specified in angstroms (e.g.
         // if they are in a.u. (=a_bohr) or something), this entry allows
         // you to specify a factor which converts them to A. Default
         // value is 1.0.
         InputCoordsToAngstromFactor;
      FXyzLoadOptions();
   };
   // Adds the molecules out of the Rasmol XYZ file to the current atom set.
   // Atoms will be added with basis string StandardBasis unless some other
   // basis is specified in *pOtherOptions.
   void AddAtomsFromXyzFile( std::string const &FileName,
      FBasisDescs const &DefaultBases,
      FXyzLoadOptions const *pOtherOptions = 0 );

   void PrintAsXyz(std::ostream &out, std::string const &Lead = "", double DistanceFactor = g_XyzDistanceScale, double GradientFactor = g_XyzGradientScale, bool IncludeGradients = true);

   // get up to 8 symmetry operators of the subgroups of D2h.
   // Each op: bit 0 -> x, bit 1->y, bit 2->z; E.g., Op 011b means invariance
   // to simultaneous mirroring around x and y.
   // Gen[i] indexes into Ops. It designates which of the operators form the minimal
   // generator set on which the symmetry adapted basis functions are supposed to
   // be based.
   void FindMirrorSymmetryOps(size_t Ops[8], size_t &nOps, size_t Gen[3], size_t &nGen) const;
   void FindEquivalentAtoms(FAtomSymGroup *&pGroups, size_t &nGroups, FMemoryStack &Mem) const;


   // makes a matrix representing a given one electron operator, and adds it
   // to Dest (which must either be 0x0 or have compatible dimensions).
   // The operator here can be specified generally using a fitting integral
   // driver object for it. For some operators of interest, direct evaluation
   // functions are provided (see next functions).
   // FDoublettIntegralFactory and derived classes are described in integrals.h
   void Add1eIntegralMatrix( FMatrixView &Dest,
      FBasisSet const &RowBasis, FBasisSet const &ColBasis,
      FDoublettIntegralFactory &IntFactory, double Factor, FMemoryStack &Mem ) const;
   void Make1eIntegralMatrixMultiComp( double *pDest,
      FBasisSet const &RowBasis, FBasisSet const &ColBasis,
      FDoublettIntegralFactory &IntFactory, double Factor, FMemoryStack &Mem ) const;
   // the following functions are more or less dummy functions that just call
   // Make1eIntegralMatrix.
   void MakeCoreHamiltonMatrix( FMatrixView &Out, FBasisSet const &RowBasis, FBasisSet const &ColBasis, FMemoryStack &Mem ) const;
   void MakeOverlapMatrix( FMatrixView &Out, FBasisSet const &RowBasis, FBasisSet const &ColBasis, FMemoryStack &Mem ) const;
   void AddKineticMatrix( FMatrixView &Out, FBasisSet const &RowBasis, FBasisSet const &ColBasis, FMemoryStack &Mem ) const;
   void AddNuclearAttractionMatrix( FMatrixView &Out, FBasisSet const &RowBasis, FBasisSet const &ColBasis, FMemoryStack &Mem ) const;
   void AddDipoleMatrix( FMatrixView &Out, FBasisSet const &RowBasis, FBasisSet const &ColBasis,
      FVector3 const &Direction, FVector3 const &ExpansionPoint, FMemoryStack &Mem ) const;
   // for quadrupoles etc. CartMoment = [i,j,k] gives the powers of <\mu| x^i y^j z^k |\nu>.
   void MakeCartesianMomentMatrix( FMatrixView &Out, FBasisSet const &RowBasis, FBasisSet const &ColBasis,
      TVector3<unsigned> const &CartMoment, FVector3 const &ExpansionPoint, FMemoryStack &Mem ) const;

   // form 1st derivative integral contraction GradXyz[Ax] = \sum{\mu,\nu} Rdm[\mu,\nu] d/d[Ax] (\mu|CoreH|\nu)
   // where Ax denotes the combination of atomic center iNuc and xyz cartesian derivative.
   void AccCoreHamiltonDeriv1( FMatrixView &GradXyz, FMatrixView const &Rdm, FBasisSet const &RowBasis, FBasisSet const &ColBasis, FMemoryStack &Mem ) const;
   // add gradient of Nuclear-nuclear repulsion energy to GradXyz
   void AccNuclearRepulsionGradient(FMatrixView &GradXyz, FMemoryStack &Mem, double Factor=1.) const;
   // add Hessian of Nuclear-nuclear repulsion energy to HessXyz ((3N,3N)-matrix)
   void AccNuclearRepulsionHessian(FMatrixView &HessXyz, FMemoryStack &Mem, double Factor=1.) const;

   // rotate and/or translate the basis (i.e., the positions of the atoms) in real space. R is a
   // 3x3 matrix and 'd' is a 3-vector such that r' = R (r - d) gives the transformed atomic positions.
   void Transform_3x4(double const *R, double const *d, FMemoryStack &Mem);

   // these are mainly here for simplifying dealing with xyz files describing
   // reaction paths, geometry optimizations and similar things.
   double GetLastEnergy() const { return m_LastEnergy; }
   void SetLastEnergy(double Energy) { m_LastEnergy = Energy; }
   std::string const &GetCaption() const { return m_CaptionLine; }
   void SetCaption(std::string const &Caption_) { m_CaptionLine = Caption_; }
   double GetTotalGradient() const;
   std::string const &GetName() const { return m_InputName; };
   void SetName(std::string const &InputName_) { m_InputName = InputName_; };

   // makes labels like '1 Se' (output atom labels start at 1, i.e., iAtom is increased by 1 for making the label)
   // iAtomWidth: width of field for atom number. If -1, decide automatically (based on number of atoms in set).
   // If 0, use smallest fitting. If 1,2,3,4: use specified witdth.
   std::string GetAtomLabel(int iAtom, int iAtomWidth=-1) const;

   FAtomSet();
   ~FAtomSet();
public:
   uint size() const { return Atoms.size(); }
   FAtom const &operator [] (uint iAt) const { return Atoms[iAt]; }
   FAtom &operator [] (uint iAt) { return Atoms[iAt]; }
   void clear() { Atoms.clear(); }
   bool empty() const { return Atoms.empty(); }

   bool HaveGradients() const;
public:
   // idea of those things is to adjust them when more complex external fields
   // (e.g., point charges or ECPs) or other kinetic energy kernels (e.g.,
   // scalar-relativistic ones) are used. But maybe other things need to be
   // adjuted anyway.
   FKrn2iPtr MakeNuclearPotentialKernel() const;
protected:
   FKrn2iPtr MakeKineticEnergyKernel() const;

   double m_LastEnergy; // last energy computed for the set
   std::string m_InputName; // if read from an .xyz file: remember name of the file here. should probably not be here...
   std::string m_CaptionLine; // if read from an .xyz file: the xyz caption line.

   bool NeedEcpKernel() const; // returns true if there ECPs defined on any atom.
public:
   void AddAtomsFromXyzStream( std::istream &str, std::string const &FileName,
      FBasisDescs const &DefaultBases,
      FXyzLoadOptions const *pOtherOptions );

};
typedef TIntrusivePtr<FAtomSet>
   FAtomSetPtr;

// make a alignment transformation to be used with FAtomSet::Transform_3x4 and/or FBasisSet::Transform_3x4.
// That probably should not be here; at least not in this form (note: this used to be in IboView).
void FindAtomSetAligningTrafo(double pR[9], double pD[3], FVector3 const *pAtPos, double const *pAtMass, FVector3 const *pAtPosLast, size_t nAt, FMemoryStack &Mem);
void Transpose3x3(double *R);
void Trafo3x4(FVector3 &InOut, double const *R, double const *d);


std::ostream &operator << ( std::ostream &out, FAtomSet const& );
std::ostream &operator << ( std::ostream &out, FVector3 const &Pos );

class FXyzLoadException : public std::runtime_error
{
public:
   FXyzLoadException(std::string const &Reason, std::string const &FileName);
   FXyzLoadException(std::string const &Reason, std::string const &CurLine, std::string const &FileName);
};

typedef std::vector<FAtomSetPtr>
   FAtomSetList;
// load the xyz frame(s) from the given file and add them to Frames.
void LoadMultiXyz(FAtomSetList &Frames, std::string const &FileName, FBasisDescs const &DefaultBases, FAtomSet::FXyzLoadOptions const *pOtherOptions = 0);

} // namespace ct


#endif // CT8K_ATOMSET_H

// kate: indent-width 3; indent-mode normal;
