/* Copyright (c) 2012  Gerald Knizia
 * 
 * This file is part of the IR/WMME program
 * (See https://sites.psu.edu/knizia/)
 * 
 * IR/WMME is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * IR/WMME is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/
 */

#include <cmath>
#include <stdexcept>
#include <boost/format.hpp>
using boost::format;
#include "Ir.h"
#include "CtAtomSet.h"
#include "CtBasisSet.h"
#include "CtBasisLibrary.h"
#include "CtMatrix.h"

// #ifdef _DEBUG
   #include "CtIo.h"
// #endif // _DEBUG

namespace ct {


FBasisSet::FBasisSet( FAtomSet const &AtomSet, FBasisContext Context_ )
   : Context(Context_)
{
   LoadFromAtomSet(AtomSet, Context);
}

FBasisSet::FBasisSet(FBasisShell const *pShells_, size_t nShells_, FBasisContext Context_, std::string const &Name_)
   : Context(Context_), Name(Name_)
{
//    Shells.assign(&pShells_[0], &pShells_[0] + nShells_);
   Shells.clear();
   Shells.reserve(nShells_);
   for (size_t iSh = 0; iSh < nShells_; ++ iSh)
      Shells.push_back(pShells_[iSh]);
   Finalize();
}


std::string BasisContextName( FBasisContext Context )
{
   switch( Context ) {
      case BASIS_Orbital: return "ORBITAL";
      case BASIS_Guess: return "GUESS";
      case BASIS_JFit: return "JFIT";
      case BASIS_JkFit: return "JKFIT";
      case BASIS_Mp2Fit: return "MP2FIT";
      case BASIS_CcsdFit: return "EXTFIT";
      case BASIS_F12RI: return "OPTRI";

      default:
         assert(0);
         return "[unknown basis context]";
   }
}


void FindDefaultBasis( std::string &Out, FBasisContext Context, std::string const &In )
{
   assert( Context != BASIS_Orbital );
   // TODO: do special element/context/basis-type specific stuff here.

   if (In == "MINAO" && Context == BASIS_Mp2Fit)
      Out = "def2-svp-mp2fit";
   else
      Out = In + "-" + BasisContextName(Context);
}


void FBasisSet::LoadFromAtomSet( FAtomSet const &AtomSet, FBasisContext Context )
{
   std::string
      BasisName;
   bool
      AllEqual = true;

   BasisName.reserve(64);
   for ( size_t iAt = 0; iAt < AtomSet.Atoms.size(); ++ iAt ){
      FAtom const
         &Atom = AtomSet.Atoms[iAt];

      // get name of basis we are supposed to be loading.
      FBasisDescs::const_iterator
         itDesc = Atom.BasisDesc.find(Context);
      if ( itDesc != Atom.BasisDesc.end() )
         // basis for context explicitly set
         BasisName = itDesc->second;
      else {
         // basis for current context not set---go find one.
         if ( Context == BASIS_Guess ) {
            // should actually check for ECP-aware guess basis. But I guess
            // this will be okay for H-Kr.
            BasisName = "cc-pVTZ";
         } else {
            itDesc = Atom.BasisDesc.find(BASIS_Orbital);
            if ( itDesc == Atom.BasisDesc.end() ) {
               std::stringstream str;
               str << format("No basis set assigned to atom %i at (%.4f, %.4f, %.4f).") % iAt % Atom.vPos[0] % Atom.vPos[1] % Atom.vPos[2];
               throw std::runtime_error(str.str());
            }
            FindDefaultBasis( BasisName, Context, itDesc->second );
         }
      }

      if ( this->Name.empty() ) this->Name = BasisName;
      if ( this->Name != BasisName ) AllEqual = false;

      // we're importing molpro data... Molpro truncates basis names
      // at 32 characters. Do so here, too.
      BasisName.resize( std::min((size_t)BasisName.size(), (size_t)32 ));

      // also, convert everything to lowercase. We do that when importing sets.
      for ( size_t i = 0; i < BasisName.size(); ++ i )
         BasisName[i] = ::tolower(BasisName[i]);

      // ask basis library to add the functions.
      g_BasisSetLibrary.LoadBasisFunctions( this->Shells,
         Atom.AtomicNumber, BasisName,
         Atom.vPos, iAt );
   }

   if ( !AllEqual || this->Name.empty() )
      this->Name = BasisContextName(Context);

   // TODO:
   //   - prevent re-allocation of this->Shells during filling
   //   - make local copy of FGaussBfn objects and re-link this->Shells
   //     to these objects. This would reduce the memory spread of the data
   //     and prevent TLB misses during integration.

   Finalize();
}


void FBasisSet::Print( std::ostream &xout ) const
{
   xout << "Basis set '" << BasisContextName(Context) << "'\n" << std::endl;
   xout << "  Offs   NC/AM        Center/Position        Exponents           Contractions\n";
   xout << " -----------------------------------------   -----------------   ----------------------" << std::endl;
   size_t
      nOff = 0;
   for ( size_t iSh = 0; iSh < Shells.size(); ++ iSh ) {
      std::streampos p0 = xout.tellp(), p1;
      xout << format("%5i   ") % nOff;
      p1 = xout.tellp();
      Shells[iSh].PrintAligned(xout, uint(p1-p0));
      xout << std::endl;
      nOff += Shells[iSh].nFn();
   }

   xout << std::endl;
}

size_t FBasisSet::nPrimitiveGtos() const
{
   size_t r = 0;
   for ( size_t iSh = 0; iSh < Shells.size(); ++ iSh )
      r += Shells[iSh].nExp() * Shells[iSh].nSh();
   return r;
}

size_t FBasisSet::nFnOfLargestShell() const
{
   size_t r = 0;
   for ( size_t iSh = 0; iSh < Shells.size(); ++ iSh )
      r = std::max(r, size_t(Shells[iSh].nFn()));
   return r;
}

uint FBasisSet::nMaxL() const
{
   uint r = 0;
   for ( size_t iSh = 0; iSh < Shells.size(); ++ iSh )
      r = std::max(uint(Shells[iSh].l()), r);
   return r;
}

size_t FBasisSet::nFn() const
{
   size_t r = 0;
   for ( size_t iSh = 0; iSh < Shells.size(); ++ iSh )
      r += Shells[iSh].nFn();
   return r;
}


// void MakeIntMatrix( FMatrixView &Dest, FRawBasis const &BasisRow,
//    FRawBasis const &BasisCol, FKrn2i const &Krn2i, FMemoryStack &Mem, double Prefactor, bool Add )
// {
//    bool
//       MatrixSymmetric = (&BasisRow == &BasisCol);
//
//    assert(Dest.nRows == BasisRow.nFn());
//    assert(Dest.nCols == BasisCol.nFn());
//
//    double
//       *pIntResult;
//    Mem.Alloc(pIntResult, BasisRow.nMaxFnPerShell * BasisCol.nMaxFnPerShell);
//    for (size_t iShellA = 0; iShellA < BasisRow.Shells.size(); ++ iShellA){
//       ir::FRawShell const
//          *pShellA = &BasisRow.Shells[iShellA];
//       for (size_t iShellB = (!MatrixSymmetric)? 0 : iShellA;
//            iShellB < BasisCol.Shells.size(); ++ iShellB)
//       {
//          ir::FRawShell const
//             *pShellB = &BasisCol.Shells[iShellB];
//          size_t
//             nSizeA = pShellA->nFn(),
//             nSizeB = pShellB->nFn();
//          assert(nSizeA == BasisRow.nFn(iShellA) &&
//                 nSizeB == BasisCol.nFn(iShellB));
// //          for (size_t i = 0; i < BasisRow.nMaxFnPerShell * BasisCol.nMaxFnPerShell; ++i)
// //             pIntResult[i] = 0.00777777; // remove this.
//          Krn2i.EvalInt2e2c(pIntResult, 1, nSizeA, pShellA, pShellB, Prefactor, false, Mem);
//
//          // fill data we gathered into the matrices
//          for ( size_t j_ = 0; j_ < nSizeB; ++ j_ )
//             for ( size_t i_ = 0; i_ < nSizeA; ++ i_ ) {
//                size_t
//                   i = BasisRow.iFn(iShellA) + i_,
//                   j = BasisCol.iFn(iShellB) + j_;
//                FScalar const
//                   &r = pIntResult[i_ + nSizeA * j_];
//                if (!Add) {
//                   Dest(i,j) = r;
//                   if ( MatrixSymmetric )
//                      Dest(j,i) = r;
//                } else {
//                   Dest(i,j) += r;
//                   if ( MatrixSymmetric && iShellA != iShellB )
//                      Dest(j,i) += r;
//                }
//             }
//       }
//    }
//    Mem.Free(pIntResult);
// }

// void MakeIntMatrix( FMatrixView &Dest, FRawBasis const &BasisRow,
//    FRawBasis const &BasisCol, FKrn2i const &Krn2i, FMemoryStack &Mem_, double Prefactor, bool Add )
// {
//    bool
//       MatrixSymmetric = (&BasisRow == &BasisCol);
//
//    assert(Dest.nRows == BasisRow.nFn());
//    assert(Dest.nCols == BasisCol.nFn());
//
//    FMemoryStackArray MemStacks(Mem_);
//
// //          Mem.Alloc(pIntResult, BasisRow.nMaxFnPerShell * BasisCol.nMaxFnPerShell);
//    #pragma omp parallel for schedule(dynamic)
//    for (int iShellB_ = 0; iShellB_ < int(BasisCol.Shells.size()); ++ iShellB_) {
//       FMemoryStack &Mem = MemStacks.GetStackOfThread();
//       size_t iShellB = size_t(iShellB_);
//       ir::FRawShell const
//          *pShellB = &BasisCol.Shells[iShellB];
//       for (size_t iShellA = (!MatrixSymmetric)? 0 : iShellB;
//            iShellA < BasisRow.Shells.size(); ++ iShellA)
//       {
//          ir::FRawShell const
//             *pShellA = &BasisRow.Shells[iShellA];
//          size_t
//             nSizeA = pShellA->nFn(),
//             nSizeB = pShellB->nFn();
//          assert(nSizeA == BasisRow.nFn(iShellA) &&
//                 nSizeB == BasisCol.nFn(iShellB));
//          TMemoryLock<double>
//             pIntResult(nSizeA * nSizeB, &Mem);
//
//          Krn2i.EvalInt2e2c(pIntResult, 1, nSizeA, pShellA, pShellB, Prefactor, false, Mem);
//
//          // fill data we gathered into the matrices
//          for ( size_t j_ = 0; j_ < nSizeB; ++ j_ )
//             for ( size_t i_ = 0; i_ < nSizeA; ++ i_ ) {
//                size_t
//                   i = BasisRow.iFn(iShellA) + i_,
//                   j = BasisCol.iFn(iShellB) + j_;
//                FScalar const
//                   &r = pIntResult[i_ + nSizeA * j_];
//                if (!Add) {
//                   Dest(i,j) = r;
//                } else {
//                   Dest(i,j) += r;
//                }
//             }
//       }
//    }
//    if (MatrixSymmetric) {
//       // copy newly computed triangle to other triangle (this version should have continuous
//       // writes for standard layout matrices).
//       size_t nFn = BasisRow.nFn();
//       for (size_t i = 0; i < nFn; ++ i)
//          for (size_t j = 0; j < i; ++ j)
//             Dest(j,i) = Dest(i,j);
//    }
// }
//
//
// void MakeIntMatrix( FMatrixView &Dest, FBasisSet const &RowBasis,
//    FBasisSet const &ColBasis, FKrn2i const &Krn2i, FMemoryStack &Mem, double Prefactor, bool Add )
// {
//    return MakeIntMatrix(Dest, *RowBasis.pRawBasis, *ColBasis.pRawBasis, Krn2i, Mem, Prefactor, Add);
// }
//
//
// void AccGradient2ix(double *pGrad, FMatrixView const &Rdm, FRawBasis const &BasisRow,
//    FRawBasis const &BasisCol, FKrn2i const &Krn2i, FMemoryStack &Mem_, double Prefactor)
// {
//    bool
//       MatrixSymmetric = (&BasisRow == &BasisCol);
// //    MatrixSymmetric = false; // FIXME: remove this.
// //    Rdm.Print(xout, "MATRIX TO CONTRACT");
//
//    assert(Rdm.nRows == BasisRow.nFn());
//    assert(Rdm.nCols == BasisCol.nFn());
//
//    assert(BasisRow.nCen() == BasisCol.nCen());
//    FOmpAccBlock
//       Gradient(pGrad, 3*BasisRow.nCen(), 0, Mem_);
//    FMemoryStackArray MemStacks(Mem_);
//
//    #pragma omp parallel for schedule(dynamic)
//    for (int iShellA_ = 0; iShellA_ < int(BasisRow.Shells.size()); ++ iShellA_) {
//       size_t iShellA = size_t(iShellA_);
//       FMemoryStack &Mem = MemStacks.GetStackOfThread();
//       ir::FRawShell const
//          *pShellA = &BasisRow.Shells[iShellA];
//       for (size_t iShellB = (!MatrixSymmetric)? 0 : iShellA;
//            iShellB < BasisCol.Shells.size(); ++ iShellB)
//       {
// //          if (BasisRow.iShCen(iShellA) == BasisCol.iShCen(iShellB)) continue; // FIXME: REMOVE THIS.
//
//          ir::FRawShell const
//             *pShellB = &BasisCol.Shells[iShellB];
//          double
//             Factor1 = Prefactor;
//          if (MatrixSymmetric && iShellA != iShellB)
//             Factor1 *= 2.0;
//          double const
//             *pRdmAB = &Rdm(BasisRow.iFn(iShellA), BasisCol.iFn(iShellB));
// //          xout << format("sh-contrib: %2i x %2i at (%3i,%3i). icen: A=%i, C=%i\n") % pShellA->nFn() % pShellB->nFn() % BasisRow.iFn(iShellA) % BasisCol.iFn(iShellB) % BasisRow.iShCen(iShellA) % BasisCol.iShCen(iShellB);
//          Krn2i.AccCoreDeriv1d(Gradient.pTls(), pRdmAB, Rdm.nRowSt, Rdm.nColSt,
//             pShellA, BasisRow.iShCen(iShellA), pShellB, BasisCol.iShCen(iShellB), Factor1, Mem);
// //          FMatrixView(const_cast<double*>(pRdmAB), pShellA->nFn(), pShellB->nFn(), Rdm.nRowSt, Rdm.nColSt).Print(xout, "SUB-RDM");
// //          FMatrixView(pGrad, 3, 2).Print(xout, "NEW GRAD");
//       }
//    }
//    Gradient.Join();
// }

void AccGradient2ix(double *pGrad, FMatrixView const &Rdm, FRawBasis const &BasisRow,
   FRawBasis const &BasisCol, FKrn2i const &Krn2i, FMemoryStack &Mem_, double Prefactor)
{
   assert(Rdm.nRows == BasisRow.nFn());
   assert(Rdm.nCols == BasisCol.nFn());

   return AccGradient2ix(pGrad, Rdm.pData, Rdm.nRowSt, Rdm.nColSt, BasisRow,
      BasisCol, Krn2i, Mem_, Prefactor);
}

void AccHessian2ix( FMatrixView Hess, FMatrixView const &Rdm, FRawBasis const &BasisRow,
   FRawBasis const &BasisCol, FKrn2i const &Krn2i, FMemoryStack &Mem_, double Prefactor)
{
   assert(Rdm.nRows == BasisRow.nFn());
   assert(Rdm.nCols == BasisCol.nFn());
   assert(Hess.nRows == 3*BasisRow.nCen());
   assert(Hess.nCols == 3*BasisCol.nCen());

   return AccHessian2ix(Hess.pData, Hess.nRowSt, Hess.nColSt, Rdm.pData,
      Rdm.nRowSt, Rdm.nColSt, BasisRow, BasisCol, Krn2i, Mem_, Prefactor);
}


void MakeIntMatrix( FMatrixView &Dest, FRawBasis const &BasisRow,
   FRawBasis const &BasisCol, FKrn2i const &Krn2i, FMemoryStack &Mem_, double Prefactor, bool Add )
{
   assert(Dest.nRows == BasisRow.nFn());
   assert(Dest.nCols == BasisCol.nFn());

   return MakeIntMatrix(Dest.pData, Dest.nRowSt, Dest.nColSt, BasisRow,
      BasisCol, Krn2i, Mem_, Prefactor, Add);
}

void MakeIntMatrix( FMatrixView &Dest, FBasisSet const &RowBasis,
   FBasisSet const &ColBasis, FKrn2i const &Krn2i, FMemoryStack &Mem, double Prefactor, bool Add )
{
   return MakeIntMatrix(Dest, *RowBasis.pRawBasis, *ColBasis.pRawBasis, Krn2i, Mem, Prefactor, Add);
}





size_t FBasisSet::nCenters() const
{
   size_t
      nCenters = 0;
   for ( size_t iSh = 0; iSh < Shells.size(); ++ iSh )
      if ( Shells[iSh].iCenter > 0 )
         nCenters = std::max(nCenters, 1 + static_cast<size_t>(Shells[iSh].iCenter));
   return nCenters;
}


void FBasisSet::MakeAtomOffsets( size_t *&pAtomShellOffsets, size_t *&pAtomBfnOffsets, size_t nAtoms_, FMemoryStack &Mem ) const
{
   // pAtomShellOffsets: maps atom id to first basis function shell
   //    (Shells[i]) of the atom
   // pAtomBfnOffsets: maps atom id to first basis function (AO index)
   //    of the atom
   size_t nAtoms = nCenters();
   assert( nAtoms <= nAtoms_ );
   // ^- might differ if there are some atoms without basis functions, or basis funcitons without atoms (e.g., bond functions).
   nAtoms = nAtoms_;

   Mem.Alloc(pAtomShellOffsets, nAtoms+1);
   Mem.Alloc(pAtomBfnOffsets, nAtoms+1);
   *pAtomShellOffsets = 0;
   *pAtomBfnOffsets = 0;
   // note: this code assumes that this->shells is ordered according
   // to the atom set!
   for ( size_t iAt = 0; iAt < nAtoms; ++ iAt ){
      size_t
         iSh = pAtomShellOffsets[iAt],
         iBf = pAtomBfnOffsets[iAt];
      while ( iSh < Shells.size() && Shells[iSh].iCenter == (signed)iAt ) {
         iBf += Shells[iSh].nFn();
         ++ iSh;
      }
//       _xout0("iAt: " << iAt << "   iSh: "<< iSh << "   iBf: " << iBf);
      pAtomShellOffsets[iAt+1] = iSh;
      pAtomBfnOffsets[iAt+1] = iBf;
      assert(iBf <= nFn());
   };
}

void FBasisSet::MakeShellOffsets( size_t *&pShellOffsets, FMemoryStack &Mem ) const
{
   Mem.Alloc(pShellOffsets, Shells.size() + 1);
   *pShellOffsets = 0;
   for ( size_t iSh = 0; iSh < Shells.size(); ++ iSh )
      pShellOffsets[iSh+1] = pShellOffsets[iSh] + Shells[iSh].nFn();
}


void FBasisSet::RebuildInterfacingData()
{
   pRawBasis = MakeRawBasis();
}


FRawBasisPtr FBasisSet::MakeRawBasis()
{
   if (0) {
      FRawBasisPtr
         r(new FRawBasis());
      r->Shells.reserve(Shells.size());
      r->ShellOffsets.reserve(Shells.size() + 1);
      r->ShellOffsets.push_back(0);
      r->CenterOffsets.reserve(Shells.size() + 1);
   //    r->CenterOffsets.push_back(0);
      r->nMaxFnPerShell = 0;
      for (size_t iSh = 0; iSh < Shells.size(); ++ iSh){
         while (Shells[iSh].iCenter >= (int)r->CenterOffsets.size())
            r->CenterOffsets.push_back(r->Shells.size());
         r->Shells.push_back(Shells[iSh].MakeIrShell()); // three copies...
         size_t
            nShFn = Shells[iSh].nFn();
         r->ShellOffsets.push_back(r->ShellOffsets.back() + nShFn);
         r->nMaxFnPerShell = std::max((size_t)r->nMaxFnPerShell, nShFn);
      }
      assert(r->ShellOffsets.back() == this->nFn());
      r->CenterOffsets.push_back(Shells.size());
      return r;
   } else {
      FRawBasisPtr
         r(new FRawBasis());
      r->Shells.reserve(Shells.size());

      std::vector<size_t>
         iShCen; // center ID of each shell.
      iShCen.reserve(Shells.size());
      for (size_t iSh = 0; iSh < Shells.size(); ++ iSh){
         r->Shells.push_back(Shells[iSh].MakeIrShell()); // three copies...
         iShCen.push_back(Shells[iSh].iCenter);
      }
      r->Finalize(&iShCen[0]);
      return r;
   }
}


void FBasisSet::Finalize()
{
   RebuildInterfacingData();
}

// Make InOut = R * (InOut - d).
// The funny form is chosen because this is normally required for aligning molecules
// in space. Then d is the center of mass and R gives the main axis transformation.
void Trafo3x4(FVector3 &InOut, double const *R, double const *d)
{
   FVector3
      r = InOut - FVector3(d[0], d[1], d[2]),
      Rr(0., 0., 0.);
   for (unsigned i = 0; i < 3; ++ i)
      for (unsigned j = 0; j < 3; ++ j)
         Rr[i] += R[i + 3*j] * r[j];
   InOut = Rr;
}

void FBasisSet::Transform_3x4(double const *R, double const *d, FMemoryStack &/*Mem*/)
{
   for (size_t iSh = 0; iSh < Shells.size(); ++ iSh)
      Trafo3x4(Shells[iSh].vCenter, R, d);

   // Rebuild RawShell interface.
   // Note: Currently this is not really required as IR shells store only
   // pointers to the basis function centers.
   Finalize();
}

void FAtomSet::Transform_3x4(double const *R, double const *d, FMemoryStack &/*Mem*/)
{
   for (size_t iAt = 0; iAt < size(); ++ iAt)
      Trafo3x4(Atoms[iAt].vPos, R, d);
}


// #ifdef INCLUDE_OPTIONALS
#ifdef PROG_IBOVIEW
void FBasisSet::TransformMoCoeffs_3x4(FMatrixView Orbs, double const *R, FMemoryStack &Mem) const
{
   assert(Orbs.nRows == nFn());
   // construct transformation amongst solid harmonic components which
   // is induced by the rotation R.
   double
      *pAllSlcT;
   size_t
      nStride;
   ir::EvalSlcXRotationTrafo(pAllSlcT, nStride, nMaxL(), R, Mem); // <- allocates pAllSlcT on Mem.

   // apply it to all the components of the MOs.
   size_t
      iShOf = 0; // offset of the current shell within the basis.
   for (size_t iSh = 0; iSh < Shells.size(); ++ iSh) {
      FBasisShell const
         &Sh = Shells[iSh];
      uint
         l = Sh.l();
      for (size_t iCo = 0; iCo < Sh.nCo(); ++ iCo) {
         size_t
            // offset of the spherical components of the current contraction
            iMoOff = iShOf + iCo * (2*l+1);
         FMatrixView
            // original MO coefficients of current spherical set
            C = FMatrixView(&Orbs(iMoOff,0), 2*l+1, Orbs.nCols, Orbs.nRowSt, Orbs.nColSt),
            // transformed MO coefficients.
            Ct = MakeStackMatrix(2*l+1, Orbs.nCols, Mem),
            // sub-matrix of pAllSlcT for transformation amongst Slc at current l.
            SlcT = FMatrixView(pAllSlcT + (l*l)*(nStride+1), (2*l+1), (2*l+1), 1, nStride);
         Mxm(Ct, SlcT, C);
         Move(C, Ct);

         Mem.Free(Ct.pData);
      }

      iShOf += Sh.nFn();
   }
   assert(iShOf == nFn());

   Mem.Free(pAllSlcT);
}
#endif // PROG_IBOVIEW
// #endif // INCLUDE_OPTIONALS




} // namespace ct
