/* Copyright (c) 2012  Gerald Knizia
 * 
 * This file is part of the IR/WMME program
 * (See https://sites.psu.edu/knizia/)
 * 
 * IR/WMME is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * IR/WMME is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/
 */

#include "CxParseSd.h"
#include "CtInputParse.h"
#include <iostream>
// #include <boost/format.hpp>
#include "CtBasisShell.h"
#include "format.h"

// c++ -Wall -O2 -pg CxParseSd.cpp CxMemoryStack.cpp ParseTest.cpp && a.out
// void AicAssertFail(char const *pWhat, char const *pFile, int iLine){
//    std::cerr << boost::format("Assertion failed: '%s'") % pWhat << std::endl;
//    if (pFile != 0 && iLine > 0)
//       std::cerr << boost::format("at %s:%i") % pFile % iLine << std::endl;
// }

namespace ct {

typedef ct::FSdfParseContext::FNode
   FSdNode;

struct FBasisSetParseContext
{
   std::vector<FAtomShellCptr>
      m_Fns; // maps basis function ids to basis functions.
   std::vector<FBasisShell>
      m_Shells; // output shell array.

   // if current node at pc is of type "Basis", read this basis, construct it into pBasis, and advance.
   bool TryParseBasisSet(FBasisSetPtr &pBasis, ct::FSdfParseContext *pc, FBasisContext Context);
protected:
   bool TryParseBasisFn(FAtomShellCptr &pFn, int &FnId, ct::FSdfParseContext *pc);
   // note: this adds the shell to m_Shells.
   bool TryParseBasisShell(ct::FSdfParseContext *pc);
};

bool FBasisSetParseContext::TryParseBasisFn(FAtomShellCptr &pFn, int &FnId, ct::FSdfParseContext *pc)
{
   ct::FParseStackGuard StackGuard(pc);
   if (!pc->TryEnter("Fn"))
      return false;

   uint
      nExp, nCo, AngMom;
   if (!pc->TryRead(FnId, "Id"))
      FnId = -1;
   pc->Read(AngMom, "l");
   pc->Read(nExp, "nExp");
   pc->Read(nCo, "nCo");
   ct::TArray<double>
      Exps(nExp), ThisCo(nExp), Cos; // init just for reserve().
   pc->Read(Exps, "Exp");
   if (Exps.size() != nExp)
      throw ct::FSdfParseError("expected and actual number of exponents does not match");
   Cos.reserve(nExp * nCo);
   for (uint iCo = 0; iCo < nCo; ++ iCo) {
      pc->Read(ThisCo, "Co");
      if (ThisCo.size() != nExp)
         throw ct::FSdfParseError("basis contraction has wrong number of coefficients.");
      Cos.insert(Cos.end(), ThisCo.begin(), ThisCo.end());
   }
   if (pc->NodeExists())
      throw ct::FSdfParseError("unexpected data after basis contractions.");

//    std::cout << boost::format("*rd-fn: l = %i  nExp = %i  nCo = %i\n") % AngMom % nExp % nCo;
//    std::cout << boost::format("*rd-shell: xyz = (%.4f, %.4f, %4f) l = %i  nExp = %i  nCo = %i\n") % x % y % z % AngMom % nExp % nCo;

   pFn = new FAtomShell(AngMom, &Exps[0], nExp, &Cos[0], nCo);

   pc->Leave();
   return true;
};

bool FBasisSetParseContext::TryParseBasisShell(ct::FSdfParseContext *pc)
{
   ct::FParseStackGuard StackGuard(pc);
   if (!pc->TryEnter("Shell"))
      return false;

   int
      iAt, iFnId;
   double
      x, y, z;
   pc->Read(iAt, "iAt");
   pc->Read(x, "x");
   pc->Read(y, "y");
   pc->Read(z, "z");

   FAtomShellCptr
      pFn;
   if (pc->TryRead(iFnId, "FnId")) {
      // link to an existing basis function, by ID.
      if (iFnId < 0 || (std::size_t)iFnId >= m_Fns.size())
         throw ct::FSdfParseError("basis shell references non-existent basis function.");
      pFn = m_Fns[(std::size_t)iFnId];
   } else {
      // try to read a new basis function declaration.
      if (TryParseBasisFn(pFn, iFnId, pc)) {
         if (iFnId != -1)
            throw ct::FSdfParseError("inline defined basis functions must be local (no id).");
      } else
         throw ct::FSdfParseError("expected reference to basis function or inline definition of basis function.");
   }
//    std::cout << boost::format("*rd-shell: xyz = (%.4f, %.4f, %4f) nFn = %i (FnId=%i)\n") % x % y % z % pFn->nFn() % iFnId;

   pc->Leave();

   // store output shell.
   m_Shells.push_back(FBasisShell(ct::FVector3(x,y,z), iAt, pFn));
   return true;
};


bool FBasisSetParseContext::TryParseBasisSet(FBasisSetPtr &pBasis, ct::FSdfParseContext *pc, FBasisContext Context)
{
   ct::FParseStackGuard StackGuard(pc);
   if (!pc->TryEnter("Basis"))
      return false;

   double
      Version;
   uint
      nFns, nShells;
   std::string
      Name;
   pc->Read(Version, "Version");
   pc->Read(nFns, "nFns"); // a raw basis function shell (not placed on an atom)
   pc->Read(nShells, "nShells"); // assignments of function shells to atoms.
   if (!pc->TryRead(Name, "Name"))
      Name = "(read)";

   // read and remember global function definitions which may be used
   // on multiple atoms.
   m_Fns.reserve(nFns);
   for (uint iFn = 0; iFn < nFns; ++ iFn) {
      FAtomShellCptr
         pFn;
      int
         iFnId;
      if (!TryParseBasisFn(pFn, iFnId, pc))
         throw ct::FSdfParseError("expected basis function data.");
      if (iFnId < 0 || (std::size_t)iFnId != m_Fns.size())
         throw ct::FSdfParseError("expected basis fns to have sequential ids.");
      m_Fns.push_back(pFn);
   };
   // read data on placement of shells on atoms.
   m_Shells.reserve(nShells);
   for (uint iShell = 0; iShell < nShells; ++ iShell) {
      if (!TryParseBasisShell(pc))
         throw ct::FSdfParseError("expected basis shell data.");
   };

   pc->Leave();

   pBasis = new FBasisSet(&m_Shells[0], m_Shells.size(), Context, Name);
   return true;
}

FBasisSetPtr LoadBasisFromSdFile(char const *pFileName, FBasisContext Context)
{
   ct::FMemoryStack2 Mem(200 << 20); // 200 mb.
   ct::FSdfParseContext
      pc(&Mem); // keeps a reference. parse tree goes here.
   void *pBeginOfMemory = Mem.Alloc(0);
   FSdNode
      *pRootNode;
   pc.ParseFile(&pRootNode, pFileName);
//    pc.DumpTree(std::cout, pRootNode);

   pc.BeginParse1(pRootNode);
   FBasisSetParseContext
      bspc;
   FBasisSetPtr
      pBasis;
   if (bspc.TryParseBasisSet(pBasis, &pc, Context)) {
//       std::cout << " Successfully read a basis set." << std::endl;
      std::cout << fmt::format(" * read basis from {} ({} functions)", pFileName, pBasis->nFn()) << std::endl;
   } else {
      throw std::runtime_error(fmt::format("Failed to read basis set from '{}'", pFileName));
   }
   pc.EndParse1();
   Mem.Free(pBeginOfMemory);
   return pBasis;
}

} // namespace ct
