/* Copyright (c) 2012  Gerald Knizia
 * 
 * This file is part of the IR/WMME program
 * (See https://sites.psu.edu/knizia/)
 * 
 * IR/WMME is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * IR/WMME is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/
 */

#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <stdexcept>
#include <cctype> // for tolower()

#include <list> // used in FFileLocator
#include <map> // used in FFileLocator
#include <set> // used in FFileLocator
#include "CxOsInt.h" // for GetExePath(), used in FFileLocator. Maybe too much?
#include "CxParse1.h" // also used in FFileLocator...

#ifdef QT_CORE_LIB
   #include <QString>
   #include <QByteArray>
   #include <QFile>
   #include <QApplication>
   #include <QClipboard>
#endif

// #include "CtCommon.h"
// #include "CtConstants.h"
// #define USE_INDENT_STREAM
#ifdef USE_INDENT_STREAM
   #define INDENTSTREAM_IMPL
   #include "CxIndentStream.h"
   namespace ct {
      static fmt::FIndentStream1
         s_OutputStreamAdp(std::cout,true,1);
      std::ostream
         &xout = s_OutputStreamAdp.stream,
         &xerr = std::cerr;
   } // namespace ct
#else
   namespace ct {
      std::ostream
         &xout = std::cout,
         &xerr = std::cerr;
   } // namespace ct
#endif // USE_INDENT_STREAM

#include "CtIo.h"


namespace ct {

char const
   *pResultFmt = " %-32s%18.12f\n",
   *pResultFmtAnnoted = " %-32s%18.12f  (%s)\n",
   *pResultFmtI = " %-32s%5i\n",
   *pResultFmtIAnnoted = " %-32s%5i  (%s)\n",
   *pTimingFmt = " Time for %s:%40t%10.2f sec\n",
   *pTimingPerFmt = " Time per %s:%40t%10.2f sec  (n=%i)\n";
//    *pTimingFmt = " Time for %s:%40t%10.4f sec\n",
//    *pTimingPerFmt = " Time per %s:%40t%10.4f sec  (n=%i)\n";

int
   Verbosity = 0;

void FatalError( std::string const &Message,
    char const *pFromWhere, int nLine )
{
   std::stringstream str;
   str << Message;
   if ( pFromWhere ) {
      str << " at " << pFromWhere << ":" << nLine;
   };
   throw std::runtime_error(str.str());
}



// makes a string to be output into the log as the first line of a major
// program component.
void MajorProgramIntro( std::ostream &out, const std::string &Name, const std::string &Version )
{
//     out << fmt::unind();
    out << "\n*** " << Name;
    if ( Version != "" )
        out << " [Ver. " << Version << "]";
    out << " ***\n" << std::endl;
//     out << fmt::eind();
}

// makes a string lower case. Will break with localized multibyte characters,
// but we don't such.
std::string tolower( std::string const &in )
{
    std::string
        Result(in);
    std::string::iterator
        it;
    for( it = Result.begin(); it != Result.end(); ++it )
        *it = std::tolower(*it);
    return Result;
}

// returns string 'in' with all spaces (but not tabs) removed.
std::string stripwhitespace( std::string const &in )
{
    std::string
        Result;
    Result.reserve( in.size() );
    for ( size_t i = 0; i < in.size(); ++i )
        if ( ' ' != in[i] )
            Result.push_back( in[i] );
    return Result;
}

// will load an entire file into the stringstream str. Returns false if failed.
bool LoadFileIntoMemory( TArray<char> &pFileContent,
        std::string const &FileName, unsigned *pFileLength )
{
//    std::cout << "in LoadFileIntoMemory. Name: '" << FileName << "'." << std::endl;
#ifdef QT_CORE_LIB
   // Use QT functions to load the file. This gives us access to files
   // stored in resources and the application clipboard.
   QByteArray
      Text;
   if (FileName == ":/!clipboard!") {
      QClipboard
         *clipboard = QApplication::clipboard();
      QString
         Subtype = "plain",
         sText = clipboard->text(Subtype);
      Text = sText.toUtf8();
   } else {
      QFile
         StyleFile(QString(FileName.c_str()));
      if (!StyleFile.open(QFile::ReadOnly))
         return false;
      Text = StyleFile.readAll();
   }
   pFileContent.resize(Text.size() + 2);
   // 0-terminate the string.
   pFileContent[Text.size()] = 0;
   pFileContent[Text.size()+1] = 0;
   memcpy(&pFileContent[0], Text.data(), Text.size());
   if (pFileLength)
      *pFileLength = Text.size();
   return true;
#else
    // read the entire file into an stringstream object.
    std::ifstream
        File( FileName.c_str() );
    std::size_t
        FileLength;
    if ( false == File.good() )
        return false;
    File.seekg( 0, std::ios::end );
    FileLength = File.tellg();
    if ( 0 != pFileLength )
        *pFileLength = FileLength;
    pFileContent.resize(2 + FileLength);
    memset( &pFileContent[0], 0, 2 + FileLength );
    File.seekg( 0, std::ios::beg );
    File.read( &pFileContent[0], FileLength );
    return true;
#endif
}


FInputError::FInputError(std::string const &Reason)
   : FBase(Reason)
{}


std::string FormatConvergenceError(std::string const &CalcType, ptrdiff_t iFinalIt, double fFinalResidual, double fFinalEnergyChange)
{
   fmt::MemoryWriter w;
   w.write("{} failed to converge", CalcType);
   if (iFinalIt != -1)
      w.write(" within {} iterations", iFinalIt);
   w << ".";
   if (fFinalResidual != -1. || fFinalEnergyChange != -1.) {
      w << " (";
      bool bFirst = false;
      if (fFinalResidual != -1.) {
         bFirst = false;
         w.write("Final residual: {:8.2e}", fFinalResidual);
      }
      if (fFinalEnergyChange != -1.) {
         if (!bFirst) w << "; ";
         bFirst = false;
         w.write("Final energy change: {:16.8f}", fFinalEnergyChange);
      }
      w << ")";
   }
   return w.str();
}

FConvergenceError::FConvergenceError(std::string const &CalcType, ptrdiff_t iFinalIt, double fFinalResidual, double fFinalEnergy)
   : FBase(FormatConvergenceError(CalcType, iFinalIt, fFinalResidual, fFinalEnergy))
{}


FUnsupportedError::FUnsupportedError(std::string const &Reason)
   : FBase("Unsupported combination of options/methods: " + Reason)
{}



FLogError::FLogError(std::string const &Reason)
   : FBase(Reason)
{}


FLog::~FLog()
{
}

char const
   *p1ResultFmt = " {:<32}{:18.12f}\n",
   *p1ResultFmtAnnotated = " {:<32}{:18.12f}  ({})\n",
   *p1CountFmt = " {:<32}{:5}\n",
   *p1CountFmtAnnotated =" {:<32}{:5}  ({})\n",
   *p1InfoFmt = " {:<32}{:>18}\n",
   *p1InfoFmtAnnotated = " {:<32}{:>18}  ({})\n",
   *p1TimingFmt = "{:<40}{:10.2f} sec\n",
   *p1TimingFmtAnnotated = "{:<40}{:10.2f} sec  (n={})\n";


std::string FmtResult(fmt::BasicStringRef<char> Name, double fValue, fmt::BasicStringRef<char> Annotation)
{
   if (Annotation != "")
      return fmt::format(p1ResultFmt, Name, fValue);
   else
      return fmt::format(p1ResultFmtAnnotated, Name, fValue, Annotation);
}


std::string FmtCount(fmt::BasicStringRef<char> Name, ptrdiff_t iValue, fmt::BasicStringRef<char> Annotation)
{
   if (Annotation == "")
      return fmt::format(p1CountFmt, Name, iValue);
   else
      return fmt::format(p1CountFmtAnnotated, Name, iValue, Annotation);
}


std::string FmtTiming(fmt::BasicStringRef<char> Name, double fTimeInSeconds, size_t nTasks)
{
   if (nTasks == size_t(-1))
      return fmt::format(p1TimingFmt, fmt::format(" Time for {}:", Name), fTimeInSeconds);
   else
      return fmt::format(p1TimingFmtAnnotated, fmt::format(" Time per {}:", Name), fTimeInSeconds/double(nTasks), nTasks);
}



void FLog::WriteResult(fmt::BasicStringRef<Char> Name, double fValue)
{
   w.write(p1ResultFmt, Name, fValue);
   Flush();
}

void FLog::WriteResult(fmt::BasicStringRef<Char> Name, double fValue, fmt::BasicStringRef<Char> Annotation)
{
   w.write(p1ResultFmtAnnotated, Name, fValue, Annotation);
   Flush();
}

void FLog::WriteCount(fmt::BasicStringRef<Char> Name, ptrdiff_t iValue)
{
   w.write(p1CountFmt, Name, iValue);
   Flush();
}

void FLog::WriteCount(fmt::BasicStringRef<Char> Name, ptrdiff_t iValue, fmt::BasicStringRef<Char> Annotation)
{
   w.write(p1CountFmtAnnotated, Name, iValue, Annotation);
   Flush();
}

void FLog::WriteInfo(fmt::BasicStringRef<Char> Name, fmt::BasicStringRef<Char> Value)
{
   w.write(p1InfoFmt, Name, Value);
   Flush();
}

void FLog::WriteInfo(fmt::BasicStringRef<Char> Name, fmt::BasicStringRef<Char> Value, fmt::BasicStringRef<Char> Annotation)
{
   w.write(p1InfoFmtAnnotated, Name, Value, Annotation);
   Flush();
}


void FLog::WriteTiming(fmt::BasicStringRef<Char> Name, double fTimeInSeconds)
{
   w.write(p1TimingFmt, fmt::format(" Time for {}:", Name), fTimeInSeconds);
   Flush();
}

void FLog::WriteTiming(fmt::BasicStringRef<Char> Name, double fTimeInSeconds, size_t nTasks)
{
   w.write(p1TimingFmtAnnotated, fmt::format(" Time per {}:", Name), fTimeInSeconds/double(nTasks), nTasks);
   Flush();
}

void FLog::WriteProgramIntro(fmt::BasicStringRef<Char> Name, fmt::BasicStringRef<Char> Version)
{
   w << "\n*** " << Name;
   if (Version.size() != 0u)
      w << " [Ver. " << Version << "]";
   w << " ***\n\n";
   Flush();
}

FLog::FRunStatus FLog::GetStatus()
{
   // do nothing. derived classes might do more involved things.
   return STATUS_Okay;
}

void FLog::CheckStatus()
{
   // do nothing. derived classes might, however.
   FRunStatus
      Status = GetStatus();
   if (Status == STATUS_Okay)
      return;
   if (Status == STATUS_AbortSignalled)
      throw ct::FLogError("Abort signalled.");
   throw ct::FLogError("FLog signalled unrecognized error.");
}


FLogStdStream::FLogStdStream(std::ostream &TargetStream)
   : m_TargetStream(TargetStream)
{
}

void FLog::EmitWarning(fmt::BasicStringRef<Char> Name)
{
   w << " WARNING: " << Name << "\n";
   Flush();
}
void FLog::EmitError(fmt::BasicStringRef<Char> Name)
{
   w << " ERROR: " << Name << "\n";
   Flush();
}


void FLogStdStream::Flush()
{
   m_TargetStream << w.c_str();
   w.clear();
   m_TargetStream.flush();
}


} // namespace ct


// c/p'd from http://stackoverflow.com/questions/216823/whats-the-best-way-to-trim-stdstring
#include <algorithm>
#include <functional>
#include <cctype>
#include <locale>
#include <string.h> // for strlen

namespace ct {

// trim from left
void TrimLeft(std::string &s) {
   s.erase(s.begin(), std::find_if(s.begin(), s.end(), std::not1(std::ptr_fun<int, int>(std::isspace))));
}

// trim from right
void TrimRight(std::string &s) {
   s.erase(std::find_if(s.rbegin(), s.rend(), std::not1(std::ptr_fun<int, int>(std::isspace))).base(), s.end());
}

// trim from both ends
void Trim(std::string &s) {
   TrimLeft(s);
   TrimRight(s);
}

bool StartsWith(std::string const &s, std::string const &prefix) {
   return 0 == s.compare(0, prefix.size(), prefix);
}

bool StartsWith(std::string const &s, char const *prefix) {
   return 0 == s.compare(0, strlen(prefix), prefix);
}

bool StartsWith(std::string const &s, char prefix) {
   if (s.empty()) return false;
   return s[0] == prefix;
}

bool EndsWith(std::string const &s, std::string const &suffix)
{
   if (s.empty() || suffix.size() > s.size()) return false;
   return 0 == s.compare(s.size() - suffix.size(), suffix.size(), suffix);

}

bool EndsWith(std::string const &s, char const *suffix)
{
   size_t suffixlen = strlen(suffix);
   if (s.empty() || suffixlen > s.size()) return false;
   return 0 == s.compare(s.size() - suffixlen, suffixlen, suffix);

}


void StripLineComment(std::string &s, char const *prefix)
{
   size_t iPos = s.find(prefix);
   if (iPos == std::string::npos)
      return; // not in there.
   // strip off everything after the comment indicator, and then whitespace to the right.
   s.resize(iPos);
   TrimRight(s);
}




struct FFileLocatorImpl
{
   void AddSearchPath(std::string const &PathName);

   FFileLocatorImpl();

   std::string FindFile(std::string const &FileName, unsigned Flags) const;
protected:
   typedef std::list<std::string>
      FDirectoryList;
   FDirectoryList
      m_SearchPaths;
   typedef std::set<std::string>
      FDirectorySet;
   FDirectorySet
      m_SearchPaths_Unordeded;
   typedef std::map<std::string, std::string>
      FStrStrMap;
   FStrStrMap
      m_Substitutions;
   // apply substitutions and remove trailing '/' / '\' characters.
   std::string ProcessFileName(std::string const &FileName) const;
};


FFileLocator::FFileLocator()
   : p(new FFileLocatorImpl)
{
}

FFileLocator::~FFileLocator()
{
   delete p;
   p = 0;
}

std::string
   g_BasePathOverride;


FFileLocatorImpl::FFileLocatorImpl()
{
   std::string
      // this is used to override the base path in the Debug/Release configurations of VC
      // on windows. Otherwise the relative directory stuff does not work.
      sBasePath = g_BasePathOverride;
   if (sBasePath.size() == 0)
      sBasePath = GetExePath();
   m_Substitutions["$BASEPATH"] = ProcessFileName(sBasePath);
}


std::string FFileLocator::JoinPath(std::string const &Path, std::string const &FileName)
{
   string_slice
      slPath(Path);
   if (slPath.empty() || slPath == "." || slPath == "./")
      return FileName;
   if (slPath.endswith("/") || slPath.endswith("\\"))
      return fmt::format("{}{}", Path, FileName);
   else {
#ifdef _WIN32
      return fmt::format("{}\\{}", Path, FileName);
#else
      return fmt::format("{}/{}", Path, FileName);
#endif
   }
}


static void ReplaceSubStr(std::string &inout, std::string const &key, std::string const &value)
{
   size_t ipos = inout.find(key);
   if (ipos != std::string::npos) {
      inout.replace(ipos, key.size(), value);
   }
}


std::string FFileLocatorImpl::ProcessFileName(std::string const &FileName_) const
{
   std::string
      FileName = FileName_;
   FStrStrMap::const_iterator
      itSubst;
   // apply directory substitutions
   for (itSubst = m_Substitutions.begin(); itSubst != m_Substitutions.end(); ++ itSubst) {
      ReplaceSubStr(FileName, itSubst->first, itSubst->second);
   }

   // get rid of spaces to either side and trailing back- and front-slashes.
   string_slice
      slName(FileName);
   slName.trim();
   while (slName.first != slName.last && (slName.endswith("/") || slName.endswith("\\")))
      slName.last -= 1;
   return slName.to_str();
}


void FFileLocatorImpl::AddSearchPath(std::string const &PathName)
{
   std::string
      s = ProcessFileName(PathName);
   // add the string to our search path list... unless we already have it there.
   if (m_SearchPaths_Unordeded.find(s) == m_SearchPaths_Unordeded.end()) {
      m_SearchPaths.push_front(ProcessFileName(PathName));
      m_SearchPaths_Unordeded.insert(s);
   }
}


static bool IsFileReadable(std::string const &FileName)
{
   // well... slow, but simple.
   std::ifstream
      File(FileName.c_str());
   return File.good();
}


std::string FFileLocatorImpl::FindFile(std::string const &FileName, unsigned Flags) const
{
   static bool const Print = false;
   if (Print) {
         std::cout << fmt::format(": searching for {} (#paths: {})\n", FileName, m_SearchPaths.size());
   }
   FDirectoryList::const_iterator
      itPath;
   for (itPath = m_SearchPaths.begin(); itPath != m_SearchPaths.end(); ++ itPath) {
      std::string
         s = FFileLocator::JoinPath(*itPath, FileName);
      if (IsFileReadable(s)) {
         if (Print) {
            std::cout << fmt::format(": -> found '{}'\n", s);
         }
         return s;
      }
   }
   // couldn't find it.
   if (0 != (Flags & FFileLocator::FINDFILE_RaiseErrorIfNotFound)) {
      fmt::MemoryWriter w;
      w << "! Search paths in FindFile():\n";
      for (itPath = m_SearchPaths.begin(); itPath != m_SearchPaths.end(); ++ itPath) {
         w << fmt::format("!: '{}'\n", *itPath);
      }
      w << fmt::format("FindFile: Failed to locate file '{}'", FileName);
      throw std::runtime_error(w.str());
   } else {
      return FileName;
   }
}


std::string FFileLocator::FindFile(std::string const &FileName, unsigned Flags) const
{
   return p->FindFile(FileName, Flags);
}


void FFileLocator::AddSearchPath(std::string const &PathName)
{
   return p->AddSearchPath(PathName);
}


std::string FFileLocator::BaseName(std::string const &PathAndFileName)
{
   size_t
      iSlash = PathAndFileName.rfind('/');
#ifdef _WIN32
   size_t
      iBackSlash = PathAndFileName.rfind('\\');
   if (iBackSlash != std::string::npos && (iSlash == std::string::npos || iBackSlash > iSlash))
      iSlash = iBackSlash;
#endif
   if (iSlash != std::string::npos) {
      if (iSlash == PathAndFileName.size())
         return "";
      else
         return PathAndFileName.substr(iSlash+1);
   } else {
      // no directory separator. Return original file name.
      return PathAndFileName;
   }
}



} // namespace ct
