/* Copyright (c) 2012  Gerald Knizia
 * 
 * This file is part of the IR/WMME program
 * (See https://sites.psu.edu/knizia/)
 * 
 * IR/WMME is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * IR/WMME is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/
 */

#ifndef CT_IO_H
#define CT_IO_H

#include <string>
#include <stdexcept>
#include "CxPodArray.h"
#include "format.h" // brings in namespace "fmt", including its format function (which makes std::strings)

namespace ct {

// extern char const
//    *pResultFmt,
//    *pResultFmtAnnotated,
//    *pResultFmtI,
//    *pResultFmtIAnnotated,
//    *pTimingFmt,
//    *pTimingPerFmt;

   std::string FmtResult(fmt::BasicStringRef<char> Name, double Value, fmt::BasicStringRef<char> Annotation = "");
   std::string FmtCount(fmt::BasicStringRef<char> Name, ptrdiff_t Value, fmt::BasicStringRef<char> Annotation = "");
   std::string FmtTiming(fmt::BasicStringRef<char> Name, double fTimeInSeconds, size_t nTasks = size_t(-1));
} // namespace ct


namespace ct {
   extern std::ostream
      &xerr, &xout;
   extern int
      Verbosity;
   extern std::string
      g_BasePathOverride; // if not set, use the directory of the executable
}

namespace ct {

// makes a string lower case. Will break with multibyte characters.
std::string tolower( std::string const &in );
// returns string 'in' with all spaces (but not tabs) removed.
std::string stripwhitespace( std::string const &in );

// makes a string to be output into the log as the first line of a major
// program component.
void MajorProgramIntro( std::ostream &out, const std::string &Name, const std::string &Version="" );

// will load an entire file into the stringstream str. Returns false if failed.
// if 0 != FileLength, *FileLength will receive the number of loaded chars.
bool LoadFileIntoMemory( TArray<char> &pFileContent,
        std::string const &FileName, unsigned *pFileLength = 0 );


// warning: these use std::isspace, which, among other things,
// treats newline as whitespace! And of course they are slow. Use
// for small stuff only (note: I might want to get rid of boost to
// simplify compiling, that's why these things are here).
void TrimLeft(std::string &s);
void TrimRight(std::string &s);
void Trim(std::string &s);
bool StartsWith(std::string const &s, std::string const &prefix);
bool StartsWith(std::string const &s, char const *prefix);
bool StartsWith(std::string const &s, char prefix);
bool EndsWith(std::string const &s, std::string const &suffix);
bool EndsWith(std::string const &s, char const *suffix);
void StripLineComment(std::string &s, char const *prefix);

class FLogError : public std::runtime_error
{
public:
   typedef std::runtime_error
      FBase;
   explicit FLogError(std::string const &Reason);
};


// error in some sort of input -- (e.g., syntax, brken file formats).
class FInputError : public std::runtime_error
{
public:
   typedef std::runtime_error
      FBase;
   explicit FInputError(std::string const &Reason);
};


// error in converging some sort of calculation for which there is no workaround declared
class FConvergenceError : public std::runtime_error
{
public:
   typedef std::runtime_error
      FBase;
   explicit FConvergenceError(std::string const &CalcType, ptrdiff_t iFinalIt = -1, double fFinalResidual = -1., double fFinalEnergy = -1);
};
std::string FormatConvergenceError(std::string const &CalcType, ptrdiff_t iFinalIt = -1, double fFinalResidual = -1., double fFinalEnergyChange = -1);


// some unsupported calculation or combination of options was asked for (e.g., a tau-type meta-GGA with DFXC=1)
class FUnsupportedError : public std::runtime_error
{
public:
   typedef std::runtime_error
      FBase;
   explicit FUnsupportedError(std::string const &Reason);
};


// a class encapsulating output streams and associated messaging.
//
// Main point about not simply using std::cout or some other global thing
// is that computations might be done on distributed machines or in some
// other non-local/non-sequential way.
class FLog
{
protected:
public:
   fmt::MemoryWriter
      w;
   typedef char
      Char;

   FLog() {};
   virtual ~FLog();

   void Write(fmt::BasicStringRef<Char> format) {
      w << format << '\n';
      Flush();
   }
   void WriteNoNl(fmt::BasicStringRef<Char> format) {
      w << format;
      Flush();
   }
   void Write(fmt::BasicStringRef<Char> format, fmt::ArgList args) {
      fmt::BasicFormatter<Char>(w).format(format, args);
      w << '\n';
      Flush();
   }
   void WriteNoNl(fmt::BasicStringRef<Char> format, fmt::ArgList args) {
      fmt::BasicFormatter<Char>(w).format(format, args);
      Flush();
   }
   FMT_VARIADIC_VOID(Write, fmt::BasicStringRef<Char>)

   virtual void Flush() = 0;

   virtual void EmitWarning(fmt::BasicStringRef<Char> Name);
   virtual void EmitError(fmt::BasicStringRef<Char> Name);
   void WriteResult(fmt::BasicStringRef<Char> Name, double fValue);
   void WriteResult(fmt::BasicStringRef<Char> Name, double fValue, fmt::BasicStringRef<Char> Annotation);
   void WriteInfo(fmt::BasicStringRef<Char> Name, fmt::BasicStringRef<Char> Value);
   void WriteInfo(fmt::BasicStringRef<Char> Name, fmt::BasicStringRef<Char> Value, fmt::BasicStringRef<Char> Annotation);
   void WriteCount(fmt::BasicStringRef<Char> Name, ptrdiff_t iValue);
   void WriteCount(fmt::BasicStringRef<Char> Name, ptrdiff_t iValue, fmt::BasicStringRef<Char> Annotation);
   void WriteTiming(fmt::BasicStringRef<Char> Name, double fTimeInSeconds);
   // print relative timing per task.
   void WriteTiming(fmt::BasicStringRef<Char> Name, double fTimeInSeconds, size_t nTasks);

   // this is our mechanism for communicating with the outside world (the other side of the log).
   // GetStatus may return something else than STATUS_Okay, for example, if an abort request was
   // placed on the other side by the user.
   // Default implementation does nothing (returns STATUS_Okay).
   enum FRunStatus {
      STATUS_Okay = 0,
      STATUS_AbortSignalled = 1
   };
   virtual FRunStatus GetStatus();
   // call GetStatus; throw FLogError if status says something differnt from STATUS_Okay.
   void CheckStatus();
   bool StatusOkay() { return GetStatus() == STATUS_Okay; };

   void WriteLine() { w << "\n"; };

   void WriteProgramIntro(fmt::BasicStringRef<Char> Name, fmt::BasicStringRef<Char> Version="");
private:
   void operator = (FLog const &); // not implemented
   FLog(FLog const &); // not implemented.
};


class FLogStdStream : public FLog
{
public:
   explicit FLogStdStream(std::ostream &TargetStream);
   void Flush(); // override
protected:
   std::ostream
      &m_TargetStream;
};


struct FFileLocatorImpl;

struct FFileLocator
{
   static std::string JoinPath(std::string const &Path, std::string const &FileName);
   static std::string BaseName(std::string const &PathAndFileName);

   enum FFindFileFlags {
      FINDFILE_RaiseErrorIfNotFound = 0x01
   };

   std::string FindFile(std::string const &FileName, unsigned Flags=FINDFILE_RaiseErrorIfNotFound) const;
   // note: '$EXEPATH' is replaced by the location of the executable.
   // Could probably also replace some other variables (e.g., HOME).
   void AddSearchPath(std::string const &PathName);

   FFileLocator();
   ~FFileLocator();
protected:
   FFileLocatorImpl
      *p;
   void operator = (FFileLocator const &); // not implemented
   FFileLocator(FFileLocator const &); // not implemented
};


} // namespace ct.

#endif // CT_IO_H
