/* Copyright (c) 2012  Gerald Knizia
 * 
 * This file is part of the IR/WMME program
 * (See https://sites.psu.edu/knizia/)
 * 
 * IR/WMME is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * IR/WMME is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/
 */

#ifndef CX_NUMPY_ARRAY_H
#define CX_NUMPY_ARRAY_H

#include <stdio.h> // for FILE*.
#include <stdexcept>
#include <string>
#include <cstddef>
#include "CxPodArray.h"

// support for reading and writing array data in .npy format.
// (that's what numpy.save() and numpy.load() use)

namespace ct {

struct FIoExceptionNpy : public std::runtime_error {
   explicit FIoExceptionNpy(std::string const &Msg, std::string const &FileName = "");
};

typedef TArray<std::size_t>
   FShapeNpy;

// write the continuous array pData[i,j,k,..] to the given file in .npy
// format. pShape[i] gives the number of indices in dimension #i, nDim
// gives the total dimension of the array.
void WriteNpy(FILE *File, double const *pData, std::size_t const pShape[], std::size_t nDim);
void WriteNpy(std::string const &FileName, double const *pData, std::size_t const pShape[], std::size_t nDim);
void WriteNpy(FILE *File, double const *pData, FShapeNpy const &Shape);
void WriteNpy(std::string const &FileName, double const *pData, FShapeNpy const &Shape);

// convenience functions for making array shape objects.
FShapeNpy MakeShape(std::size_t i0);
FShapeNpy MakeShape(std::size_t i0, std::size_t i1);
FShapeNpy MakeShape(std::size_t i0, std::size_t i1, std::size_t i2);
FShapeNpy MakeShape(std::size_t i0, std::size_t i1, std::size_t i2, std::size_t i3);

struct FArrayNpy {
   TArray<double>
      Data;
   FShapeNpy
      Shape,
      Strides; // <- may be inverted when reading in stuff in C order.
   std::size_t Rank() const { return Shape.size(); }
   std::size_t Size() const;

   enum FCreationFlags {
      NPY_OrderFortran = 0x0, // first dimension first
      NPY_OrderC = 0x1, // last dimension first
      NPY_ClearData = 0x2 // set data to zero
   };

   FArrayNpy() {}
   explicit FArrayNpy(FShapeNpy const &Shape, unsigned Flags = NPY_OrderC);

   void swap(FArrayNpy &Other);

   double &operator() (std::size_t i0) { return Data[Strides[0]*i0]; }
   double &operator() (std::size_t i0, std::size_t i1) { return Data[Strides[0]*i0 + Strides[1]*i1]; }
   double &operator() (std::size_t i0, std::size_t i1, std::size_t i2) { return Data[Strides[0]*i0 + Strides[1]*i1 + Strides[2]*i2]; }
   double &operator() (std::size_t i0, std::size_t i1, std::size_t i2, std::size_t i3) { return Data[Strides[0]*i0 + Strides[1]*i1 + Strides[2]*i2 + Strides[3]*i3]; }

   double operator() (std::size_t i0) const { return Data[Strides[0]*i0]; }
   double operator() (std::size_t i0, std::size_t i1) const  { return Data[Strides[0]*i0 + Strides[1]*i1]; }
   double operator() (std::size_t i0, std::size_t i1, std::size_t i2) const { return Data[Strides[0]*i0 + Strides[1]*i1 + Strides[2]*i2]; }
   double operator() (std::size_t i0, std::size_t i1, std::size_t i2, std::size_t i3) const  { return Data[Strides[0]*i0 + Strides[1]*i1 + Strides[2]*i2 + Strides[3]*i3]; }

   // flags: bit field of NPY_*.
   void Init(FShapeNpy const &Shape_, unsigned Flags = NPY_OrderC);

   // returns whether strides are in standard order (=Fortran order) and
   // stored data is continuous regarding standard order of axes.
   // (i.e., if Strides[0] == 1, and
   //           Strides[1] = Shape[0], and
   //           Strides[2] = Stride[1] * Size[1], and so on)
   bool IsDataContinuous() const;
};

// read array from File in .npy format.
void ReadNpy(FArrayNpy &Out, FILE *File);
void ReadNpy(FArrayNpy &Out, std::string const &FileName);

// note: these ones only work for arrays in standard order (as specified above)
void WriteNpy(FILE *File, FArrayNpy const &Array);
void WriteNpy(std::string const &FileName, FArrayNpy const &Array);

} // namespace ct

#endif // CX_NUMPY_ARRAY_H
