/* Copyright (c) 2012  Gerald Knizia
 * 
 * This file is part of the IR/WMME program
 * (See https://sites.psu.edu/knizia/)
 * 
 * IR/WMME is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * IR/WMME is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/
 */

#include <iostream>
#include <fstream>
#include <iomanip>
#include <stdexcept>
#include <algorithm>
#include <sstream>

#include <cctype> // for isalnum
#include "format.h"

//#include "fmt.h"
#include "CxParse1.h"

namespace ct {


std::string ReadFile( std::string FileName )
{
   std::ifstream
      File( FileName.c_str() );
   File.seekg( 0, std::ifstream::end );
   for ( ; File.good(); ) {
      std::string
         Result( File.tellg(), '\0' );
      File.seekg( 0, std::ifstream::beg );
      File.read( &Result[0], Result.size() );
      if ( !File.good() )
         break;
      return Result;
   }
   throw std::runtime_error("Failed to read file '" + FileName + "'.");
}

// cstr_it FindIdentifierEnd( cstr_it iPos, cstr_it iEnd )
// {
//    while( iPos != iEnd && ( std::isalnum(*iPos) || '_' == *iPos )  )
//       ++ iPos;
//    return iPos;
// };

cstr_it FindIdentifierEnd( cstr_it iPos, cstr_it iEnd )
{
   while( iPos != iEnd && ( std::isalnum(*iPos) || '_' == *iPos || ':' == *iPos || '^' == *iPos )  )
      ++ iPos;
   return iPos;
}


// assumes that *pFirst == ChOpen. Returns pLast if no closing braket found.
char const *FindClosingParenthesis( char Open, char Close, char const *pFirst, char const *pLast )
{
   assert( pFirst < pLast && *pFirst == Open );
   assert( Open != Close );
   int
      // number of brackets still open.
      iLevel = 0;
   for ( ; pFirst != pLast; ++ pFirst ){
      if ( *pFirst == Open )
         iLevel += 1;
      if ( *pFirst == Close )
         iLevel -= 1;
      if ( iLevel == 0 )
         break;
   }
   return pFirst;
}


std::ostream &operator << ( std::ostream &out, string_slice const &s )
{
   out << s.to_str();
   return out;
}

bool is_whitespace_cxp1(char c) {
   return c == ' ' || c == '\t';
}


static bool is_contained_in_or_whitespace(char c, char const *pCharsToTest) {
   if (pCharsToTest) {
      for (char const *pt = pCharsToTest; *pt != 0; ++ pt)
         if (*pt == c)
            return true;
      return false;
   } else {
      return is_whitespace_cxp1(c);
   }
}


cstr_it SkipWhiteSpace( cstr_it iPos, cstr_it iEnd ){
   while (iPos != iEnd && is_whitespace_cxp1(*iPos))
      ++ iPos;
   return iPos;
}


void string_slice::trim_left(char const *pCharsToSkip)
{
   while (first != last && is_contained_in_or_whitespace(*first, pCharsToSkip))
      ++ first;
}


void string_slice::trim_right(char const *pCharsToSkip)
{
   if (first == last)
      return;
   while (first != (last-1) && is_contained_in_or_whitespace(*(last-1), pCharsToSkip))
      -- last;
}


void string_slice::trim(char const *pCharsToSkip)
{
   assert(first <= last);
   if (pCharsToSkip == 0) {
      while (first != last && is_whitespace_cxp1(*first))
         ++ first;
      if (first == last)
         return;
      while (first != (last-1) && is_whitespace_cxp1(*(last-1)))
         -- last;
   } else {
      trim_left(pCharsToSkip);
      trim_right(pCharsToSkip);
   }
}



static size_t const nParenthesisTypes = 4;
static char const ParenthesisPairs[nParenthesisTypes][2] = { {'(',')'}, {'{', '}'}, {'<','>'}, {'[',']'} };


template<class TSliceList, class TDelimList>
size_t string_slice::split_any(TSliceList &out, TDelimList *pOutDelims, char const *pDelims, size_t nDelims, size_t Flags) const
{
//    bool const Print = false;
//    if ( Print ) xout << "splitting '" << *this << "' at '" << std::string(pDelims, pDelims + nDelims) << "' into: [";

   out.clear();
   out.push_back( string_slice(first,first) );

   if (pOutDelims)
      pOutDelims->clear();

   cstr_it
      iPos = first;

   while ( out.size() < nMaxStrSplit ) {
      char Delim = 0;
      for ( ; ; ) {
//       while ( iPos != last && *iPos != Delim ){
         if ( iPos == last )
            break;
         for ( size_t i = 0; i < nDelims; ++ i )
            if ( *iPos == pDelims[i] )
               Delim = pDelims[i];
         if ( is_whitespace_cxp1(Delim) && (0 != (Flags & SPLIT_SkipWhitespace)) ) {
            // skip additional whitespace until we reach something which is not a whitespace.
            while ( iPos != last && is_whitespace_cxp1(*iPos) )
               iPos += 1;
            iPos -= 1;
            break;
         }
         if ( Delim != 0 )
            break;

         if ( 0 != (Flags & SPLIT_SkipParenthesis) ) {
            for ( size_t iParensType = 0; iParensType < nParenthesisTypes; ++ iParensType ) {
               char
                  Open = ParenthesisPairs[iParensType][0],
                  Close = ParenthesisPairs[iParensType][1];
               if ( *iPos == Open ) {
                  char const
                     *pFirst = &*iPos,
                     *pLast = &*(last-1)+1,
                     *p = FindClosingParenthesis( Open, Close, pFirst, pLast );
                  if ( p == pLast ) {
                     std::stringstream str;
                     str << "imbalanced parenthesis '" << Open << "'/'" << Close << "' in '" << *this << "'.";
                     throw std::runtime_error(str.str());
                  }
                  iPos += p - pFirst;
                  -- iPos;
               }
            }
         }
         ++ iPos;
         out.back().last = iPos;
      }
      if ( iPos == last ) {
         if ( 0 != (Flags & SPLIT_SkipWhitespace) )
            for ( size_t i = 0; i < out.size(); ++ i )
               out[i].trim();
//          if ( Print ) {
//             for ( size_t i = 0; i < out.size(); ++ i ) {
//                if ( i != 0 )
//                   xout << ", ";
//                xout << "'" << out[i] << "'";
//             }
//             xout << "]"<< std::endl;
//          }
         return out.size();
      }
      if ( pOutDelims )
         pOutDelims->push_back(Delim);
      ++ iPos;
      out.push_back( string_slice(iPos,iPos) );
   }

   std::stringstream str;
   str << "parsing error: split count overflow in '" << *this << "' while scanning for '"  << std::string(pDelims, pDelims + nDelims) << "'.";
   throw std::runtime_error(str.str());
}

template<class TSliceList>
size_t string_slice::split_list(TSliceList &out, char ListStart, char ListEnd, char ListDelim) const
{
   string_slice
      sub = *this;
   sub.trim();
   assert((ListStart == 0) == (ListEnd == 0));
   if (ListStart != 0 && ListEnd != 0) {
      // cut out list start and list end indicators (normaly [...]).
      if (sub.size() < 2 || sub[0] != ListStart || sub[sub.size()-1] != ListEnd)
         throw std::runtime_error(fmt::format("Expected argument list ({}..{}..{}), but got '{}'", ListStart, ListDelim, ListEnd, sub.to_str()));
      sub.first += 1;
      sub.last -= 1;
   }
   // split rest normally.
   return sub.split(out, ListDelim, SPLIT_SkipWhitespace | SPLIT_SkipParenthesis);
}


template<class TSliceList>
size_t string_slice::split_lines(TSliceList &out, size_t flags) const
{
   long_split_delim_list
      *pDelims = 0;
   split_any(out, pDelims, "\n", 1, flags & (~SPLIT_SkipWhitespace));
   // fix up line endings in case they were made by "\n\r", "\r\n", or similar.
   for (typename TSliceList::iterator it = out.begin(); it != out.end(); ++ it) {
//       if (it->startswith('\r'))
//          it->first += 1;
//       if (it->endswith('\r'))
//          it->last -= 1;
      while (it->startswith('\r') || it->startswith('\0'))
         it->first += 1;
      while (it->endswith('\r') || it->endswith('\0'))
         it->last -= 1;
      if (0 != (flags & SPLIT_SkipWhitespace))
         it->trim();
   }
   return out.size();
}


void string_slice::remove_trailing_path_separators()
{
   // what is the point of this?
   // When dealing with path names,
   // in some situations we need to keep the last separator
   // to obtain the correct absolute path.
   //
   // E.g., in '/home/cgk/stuff/whee.xyz', removing the trailing '/' is fine.
   // in '/whee.xyz' it is not (because with the separator it will denote the root directory,
   // without it, it will denote a directory relative to the local one).
   //
   // Similarly, both in
   //    in lib/whee.xyz
   // and in
   //    f:\whee.xyz
   // we need to keep the directory, despite the fact that there is only one.

   is_path_separator_c ips;
   for (;first != last;) {
      // return if string is empty or last character is not a path separator.
      if (empty() || !ips(*(last-1)))
         break;
      bool
         must_be_kept_anyway = false;
      // so the last character *is* some sort of path separator.
      if (size() == 1) {
         // this means the string *starts* with this.
         // means it's relative to some root directory. We must keep it.
         must_be_kept_anyway = true;
   #ifdef _WIN32
      } else if (size() >= 2 && *(last-2) == ':') {
         // logical drive designation followed by a path separator.
         must_be_kept_anyway = true;
   #endif
      } else {
      }
      if (must_be_kept_anyway)
         break;
      else {
         last -= 1;
      }
   }
}


string_slice::FPair string_slice::split_path() const
{
   // find the last '/' separator within the string content
   // (directory separator)
   cstr_it
      itDirSep = rfind('/');
#ifdef _WIN32
   // Windows is okay with both front-slash and back-slash as directory
   // separators. So look for a back-slash, too, and if it occurs farther
   // out than the last front slash (or if none was found),
   // take that one.
   cstr_it
      itBackSlash = rfind('\\');
   if (itBackSlash != last && (itDirSep == last || itBackSlash > itDirSep))
      itDirSep = itBackSlash;
#endif
   if (itDirSep != last) {
      FPair
         r(string_slice(first,itDirSep+1), string_slice(itDirSep+1,last));
      r.first.remove_trailing_path_separators();
      return r;
   } else {
      // no directory separator. Return original file name.
      return FPair(string_slice(cstr_it(),cstr_it()), *this);
   }
}


// explicitly instantiate the core split function for the combinations of sequence types we will use.
template size_t string_slice::split_any<split_result,split_delim_list>(split_result &, split_delim_list *, char const *, size_t, size_t) const;
template size_t string_slice::split_any<long_split_result,long_split_delim_list>(long_split_result &, long_split_delim_list *, char const *, size_t, size_t) const;
template size_t string_slice::split_any<long_split_result,split_delim_list>(long_split_result &, split_delim_list *, char const *, size_t, size_t) const;

template size_t string_slice::split_list<split_result>(split_result &out, char ListStart, char ListEnd, char ListDelim) const;
template size_t string_slice::split_list<long_split_result>(long_split_result &out, char ListStart, char ListEnd, char ListDelim) const;
template size_t string_slice::split_lines<long_split_result>(long_split_result &out, size_t) const;






// void string_slice::split( split_result &out, char Delim, bool SkipWhitespace, bool SkipBraces ) const
// {
//    bool const Print = false;
//    if ( Print ) xout << "splitting '" << *this << "' at '" << Delim << "' into: [";
//
//    out.clear();
//    out.push_back( string_slice(first,first) );
//
//    cstr_it
//       iPos = first;
//
//    while ( out.size() < nMaxStrSplit ) {
//       while ( iPos != last && *iPos != Delim ){
//          if ( SkipBraces ) {
//             for ( size_t iParensType = 0; iParensType < nParenthesisTypes; ++ iParensType ) {
//                char
//                   Open = ParenthesisPairs[iParensType][0],
//                   Close = ParenthesisPairs[iParensType][1];
//                if ( *iPos == Open ) {
//                   char const
//                      *pFirst = &*iPos,
//                      *pLast = &*(last-1)+1,
//                      *p = FindClosingParenthesis( Open, Close, pFirst, pLast );
//                   if ( p == pLast ) {
//                      std::stringstream str;
//                      str << "imbalanced parenthesis '" << Open << "'/'" << Close << "' in '" << *this << "'.";
//                      throw std::runtime_error(str.str());
//                   }
//                   iPos += p - pFirst;
//                   -- iPos;
//                }
//             }
//          }
//          ++ iPos;
//          out.back().last = iPos;
//       }
//       if ( iPos == last ) {
//          if ( SkipWhitespace )
//             for ( size_t i = 0; i < out.size(); ++ i )
//                out[i].trim();
//          if ( Print ) {
//             for ( size_t i = 0; i < out.size(); ++ i ) {
//                if ( i != 0 )
//                   xout << ", ";
//                xout << "'" << out[i] << "'";
//             }
//             xout << "]"<< std::endl;
//          }
//          return;
//       }
//       ++ iPos;
//       out.push_back( string_slice(iPos,iPos) );
//    }
//
//    std::stringstream str;
//    str << "parsing error: split count overflow in '" << *this << "' while scanning for '"  << Delim << "'.";
//    throw std::runtime_error(str.str());
// };

std::string string_slice::to_str() const
{
   return std::string(first,last);
}


bool string_slice::try_convert_to_int(ptrdiff_t *pr) const
{
   char const
      *pEnd = 0;
   long
      r_ = strtol(&*first, const_cast<char**>(&pEnd), 0);
   if (pEnd != &*(last-1) + 1)
      return false;
   if (pr)
      *pr = static_cast<ptrdiff_t>(r_);
   return true;
}


bool string_slice::try_convert_to_size(size_t *pr) const
{
   char const
      *pEnd = 0;
   long
      r_ = strtol(&*first, const_cast<char**>(&pEnd), 0);
   if (pEnd != &*(last-1) + 1)
      return false;
   if (r_ < 0)
      return false;
   size_t
      r = size_t(r_);
   if (pr)
      *pr = static_cast<ptrdiff_t>(r);
   return true;
}


bool string_slice::try_convert_to_float(double *pr) const
{
   char const
      *pEnd = 0;
   double
      r = strtod(&*first, const_cast<char**>(&pEnd));
   if (pEnd != &*(last-1) + 1)
      return false;
   if (pr)
      *pr = r;
   return true;
}


bool string_slice::try_convert_to_bool(bool *pr) const
{
   std::string ss(first,last);
   for (size_t i = 0; i < ss.size(); ++ i)
      ss[i] = (char)std::tolower(ss[i]);
   if (ss == "on" || ss == "true" || ss == "1" || ss == "yes") {
      if (pr) *pr = true;
      return true;
   }
   if (ss == "off" || ss == "false" || ss == "0" || ss == "no") {
      if (pr) *pr = false;
      return true;
   }
   return false;
}


ptrdiff_t string_slice::to_int() const
{
   ptrdiff_t r(0);
   if (!this->try_convert_to_int(&r))
      throw std::runtime_error("expected base10 integer, but got '" + to_str() + "'." );
   return r;
}


double string_slice::to_float() const
{
   double r(0);
   if (!this->try_convert_to_float(&r))
      throw std::runtime_error("expected floating point number, but got '" + to_str() + "'." );
   return r;
}


bool string_slice::to_bool() const
{
   bool r;
   if (!this->try_convert_to_bool(&r))
      throw std::runtime_error("expected boolean argument, but got '" + to_str() + "'." );
   return r;
}


bool string_slice::endswith( char const *pStr ) const {
   if (empty())
      return false;
   char const *p = pStr;
   while (*p != 0)
      ++ p;
   return string_slice(last - (p - pStr), last) == pStr;
}


cstr_it string_slice::find(char const *c, std::size_t offset) const {
   for ( cstr_it it = first + offset; it != last; ++ it )
      if ( string_slice(it,last).startswith(c) )
         return it;
   return last;
}


cstr_it string_slice::find(char c, std::size_t offset) const {
   return std::find(first+offset, last, c);
}


cstr_it string_slice::rfind(char c) const
{
   return rfind_c(is_equal_c(c));
}


bool string_slice::operator == ( char const *p ) const {
   for ( cstr_it it = first; it != last; ++ it, ++ p )
      if ( *p != *it )
         return false;
   return *p == 0;
}
bool string_slice::startswith( char const *p ) const {
   for ( cstr_it it = first; *p && it != last; ++ it, ++ p )
      if ( *p != *it )
         return false;
   return *p == 0;
}

// bool string_slice::startswith(char c) const
// {
//    char const buf[2] = {c, 0};
//    return this->startswith(&buf[0]);
// }
//
// bool string_slice::endswith(char c) const
// {
//    char const buf[2] = {c, 0};
//    return this->endswith(&buf[0]);
// }


bool string_slice::startswith(char c) const
{
   return (!empty() && (*first) == c);
}


bool string_slice::endswith(char c) const
{
   return (!empty() && (*(last-1)) == c);
}


string_slice::FPair string_slice::rsplit1(char Delim, size_t Flags) const
{
   long_split_result
      sr;
   this->split(sr, Delim, Flags);
   string_slice
      nothing(last,last);
   if (sr.empty())
      return FPair(nothing,nothing);
   else if (sr.size() == 1)
      return FPair(sr[0], nothing);
   else {
      assert(sr.size() >= 2);
      return FPair(string_slice(sr.front().first, sr[sr.size()-2].last), sr.back());
   }
}


string_slice::FPair string_slice::split1(char Delim, size_t Flags) const
{
   long_split_result
      sr;
   this->split(sr, Delim, Flags);
   string_slice
      nothing(last,last);
   if (sr.empty())
      return FPair(nothing,nothing);
   else if (sr.size() == 1)
      return FPair(sr[0], nothing);
   else {
      assert(sr.size() >= 2);
      return FPair(sr.front(), string_slice(sr[1].first, sr.back().last));
   }
}






void SplitLinesAndRemoveComments(FLineList &out, std::string const &Text)
{
   size_t
      nLines1 = 0;
   for ( cstr_it it = Text.begin(); it != Text.end(); ++ it )
      if ( *it == '\n' ) nLines1 += 1;
   out.reserve(nLines1);

   size_t
      iLineBegin,
      iLineEnd = 0;
   size_t
      iLine = 0;
   string_slice
      Line;
   for ( ; iLineEnd < Text.size() && iLineEnd != std::string::npos; ) {
      ++ iLine;
      iLineBegin = iLineEnd;
      while( iLineBegin < Text.size() && Text[iLineBegin] == '\n' )
         ++ iLineBegin;
      iLineEnd = Text.find('\n', iLineBegin);
      if ( iLineEnd != std::string::npos )
         Line = string_slice(Text.begin() + iLineBegin, Text.begin() + iLineEnd);
      else
         Line = string_slice(Text.begin() + iLineBegin, Text.end());
      cstr_it
         iLineContentEnd = Line.find( "//" );
      if ( iLineContentEnd != Line.last )
         Line.last = iLineContentEnd;
      while (Line.last > Line.first && is_whitespace_cxp1(*(Line.last-1)))
         -- Line.last;
      if ( Line.empty() )
         continue;
      out.push_back( FLine(Line,FCodeLoc(iLine)) );
   }
}

std::ostream &operator << ( std::ostream &out, FLine const &Line ){
   out << "\nin line " << Line.Loc.iLine << ": '" << Line.Text << "'.";
   return out;
}


FPropertyListStr::FPropertyListStr(string_slice const &Input, char const *pAssumedName, unsigned Flags, char const *pOpenDelimClose)
{
   static std::string EmptyString = "";
   char
      cOpen = '{',
      cDelim = ';',
      cClose = '}';
   if (pOpenDelimClose != 0) {
      cOpen = pOpenDelimClose[0];
      cDelim = pOpenDelimClose[1];
      cClose = pOpenDelimClose[2];
      assert(cDelim != ':');
   }
   bool
      AllowEntriesWithoutName = (0 != (Flags & PROPLIST_AllowEntriesWithoutName));
   cstr_it iPos = Input.find(cOpen);
   if ( !Input.endswith(cClose) || iPos == Input.last )
      throw std::runtime_error("expected !property list.");
   Name = string_slice(Input.first, iPos);
   Name.trim();

   string_slice
      Data(iPos+1,Input.last-1);

   long_split_result
      sr;
   Data.split(sr, cDelim);
   for ( size_t i = 0; i < sr.size(); ++ i ){
      if ( !sr[i].empty() ) { // <- empty property list is allowed.
         push_back(FPropertyStr());
         FPropertyStr
            &out = back();
         cstr_it iPos = sr[i].find(':');
         if (iPos == sr[i].end()) {
            // store as a name-less entry...
            if (!AllowEntriesWithoutName)
               throw std::runtime_error("expected ':' in property declaration of '" + (std::string)sr[i] + "'.");
            out.Name = string_slice(EmptyString);
            out.Content = sr[i];
         } else {
            out.Name = string_slice(sr[i].first, iPos);
            out.Name.trim();
            out.Content = string_slice(iPos+1, sr[i].last);
            out.Content.trim();
         }
      }
   }

   if ( pAssumedName && Name != pAssumedName ) {
      std::stringstream str;
      str << "expected '" << pAssumedName << "' but found '" << Name << "'.";
      throw std::runtime_error(str.str());
   }
}


} // namespace ct
