/* Copyright (c) 2012  Gerald Knizia
 * 
 * This file is part of the IR/WMME program
 * (See https://sites.psu.edu/knizia/)
 * 
 * IR/WMME is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * IR/WMME is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/
 */

#ifndef CX_PARSE1_H
#define CX_PARSE1_H

#include <string>
#include <vector>
#include <ostream>
#include <utility>

#include "CxTypes.h"
#include "CxFixedSizeArray.h"
#include "CxPodArray.h"

namespace ct {

std::string ReadFile( std::string FileName );

typedef std::string::iterator
   str_it;
typedef std::string::const_iterator
   cstr_it;

size_t const
   // hm.. might want to bind this to a non-initialized dynamic array, and
   // set to 'real large'. Or introduce a dynamic version of split(), too.
   nMaxStrSplit = 32;

typedef TArrayFix<char, nMaxStrSplit-1>
   split_delim_list;
typedef TArray<char, nMaxStrSplit-1>
   long_split_delim_list;


struct is_path_separator_c
{
   inline bool operator() (char c) const
   {
   #ifdef _WIN32
      return c == '/' || c == '\\';
   #else
      return c == '/';
   #endif
   }
};


struct is_equal_c
{
   is_equal_c(char c_)  : m_TargetChar(c_) {}

   inline bool operator() (char c) const {
      return c == m_TargetChar;
   }
protected:
   char
      m_TargetChar;
};



struct string_slice
{
   cstr_it
      first, last;

   string_slice() {};

   string_slice( cstr_it itFirst_, cstr_it itLast_ )
      : first(itFirst_), last(itLast_)
   {
      assert(first <= last);
   }

   string_slice( std::string const &s )
      : first(s.begin()), last(s.end())
   {}

   cstr_it begin() const { return first; }
   cstr_it end() const { return last; }
   size_t size() const { return last - first; }
   bool empty() const { return first == last; }
   void clear() { first = last; }

   // these ones raise exceptions if invalid input is found
   std::string to_str() const;
   ptrdiff_t to_int() const;
   size_t to_size() const;
   double to_float() const;
   bool to_bool() const;

   // same as above; the overloaded versions are for templatizing
   void convert(double *pr) { *pr = to_float(); }
   void convert(ptrdiff_t *pr) { *pr = to_int(); }
   void convert(size_t *pr) { *pr = to_size(); }
   void convert(bool *pr) { *pr = to_bool(); }
   void convert(std::string *pr) { *pr = to_str(); }

   // thes ones return whether the conversion was successful
   bool try_convert_to_int(ptrdiff_t *pr) const;
   bool try_convert_to_size(size_t *pr) const;
   bool try_convert_to_float(double *pr) const;
   bool try_convert_to_bool(bool *pr) const;

   // same as above; the overloaded versions are for templatizing
   bool try_convert(double *pr) { return try_convert_to_float(pr); }
   bool try_convert(ptrdiff_t *pr) { return try_convert_to_int(pr); }
   bool try_convert(bool *pr) { return try_convert_to_bool(pr); }
   bool try_convert(std::string *pr) { *pr = to_str(); return true; }



   cstr_it find(char c, size_t offset = 0) const;
   cstr_it find(char const *c, size_t offset = 0) const;
   template<class FCharPred>
   cstr_it find_c(FCharPred const &pred, size_t offset = 0) const;

   cstr_it rfind(char c) const;
   template<class FCharPred>
   cstr_it rfind_c(FCharPred const &pred) const;

   string_slice substr( size_t iFirst, size_t nCount ) const {
      assert( iFirst + nCount <= size() );
      return string_slice( first + iFirst, first + iFirst + nCount );
   }

   string_slice substr( cstr_it iFirst, size_t nCount ) const {
      assert( iFirst + nCount <= end() );
      return string_slice( iFirst, iFirst + nCount );
   }

   enum split_flags {
      // if set, whitespace at the beginning and end of each output string split will be removed.
      SPLIT_SkipWhitespace = 0x01,
      // if set, text between two parenthesis (any kind) will not be interpreted, but treated as a unit.
      SPLIT_SkipParenthesis = 0x02
   };

//    // returns out.size().
//    size_t split(TArrayFix<string_slice, nMaxStrSplit> &out, char Delim, size_t flags = SPLIT_SkipWhitespace | SPLIT_SkipParenthesis) const;
//    // split at any of the given delimiters. store delimiters in pOutDelims if given:
//    // pOutDelims[i] is the delimiter between out[i] and out[i+1].
//    size_t split_any(TArrayFix<string_slice, nMaxStrSplit> &out, TArrayFix<char, nMaxStrSplit-1> *pOutDelims, char const *pDelims, size_t nDelims, size_t flags = SPLIT_SkipWhitespace | SPLIT_SkipParenthesis) const;
//
//    // convenience function for splitting a list of sub-properties into parts.
//    size_t split_list(TArrayFix<string_slice, nMaxStrSplit> &out, char ListStart = '[', char ListEnd = ']', char ListDelim = ',') const;

   // split at any of the given delimiters. store delimiters in pOutDelims if given:
   // pOutDelims[i] is the delimiter between out[i] and out[i+1].
   // Instanciated for:
   // - split_any(split_result &, split_delim_list *, ...)   (with fixed size arrays)
   // - split_any(long_split_result &, long_split_delim_list &, ...)
   template<class TSliceList, class TDelimList>
   size_t split_any(TSliceList &out, TDelimList *pOutDelims, char const *pDelims, size_t nDelims, size_t flags) const;

   template<class TSliceList>
   size_t split_lines(TSliceList &out, size_t flags = 0) const;

   template<class TSliceList>
   size_t split(TSliceList &out, char Delim, size_t Flags = SPLIT_SkipWhitespace | SPLIT_SkipParenthesis) const
   {
      // note: if Instanciated with long_split_result, this will invoke the
      // '<long_split_result, split_delim_list>'-version of this function (note the missing 'long_'
      // on the delimiters). However, as the output delimiter list is optional and we just pass a 0-ptr
      // (and therefore no delimiters get stored in any case) this is still fine.
      return split_any<TSliceList,split_delim_list>(out, 0, &Delim, 1, Flags);
   }

   // convenience function for splitting a list of sub-properties into parts.
   template<class TSliceList>
   size_t split_list(TSliceList &out, char ListStart='[', char ListEnd=']', char ListDelim=',') const;


   bool operator == (char const *p) const;
   bool operator != (char const *p) const { return !this->operator == (p); }
   bool startswith(char const *p) const;
   bool endswith(char const *pStr) const;
   bool startswith(char c) const;
   bool endswith(char c) const;

   void resize( size_t n ) {
      last = first + n;
   };

   // skip all characters contained in pCharsToSkip (0-terminated) at begin and end.
   // If pCharsToSkip is 0, this skips whitespace.
   void trim(char const *pCharsToSkip=0);
   // skip any characters in pCharsToSkip (if pCharsToSkip != 0) or whitespace (if pCharsToSkip==0) at end only
   void trim_right(char const *pCharsToSkip=0);
   // skip any characters in pCharsToSkip (if pCharsToSkip != 0) or whitespace (if pCharsToSkip==0) at start only
   void trim_left(char const *pCharsToSkip=0);

   // should this be here? not really, I guess...
   typedef std::pair<string_slice, string_slice>
      FPair;
   FPair split_path() const;
   void remove_trailing_path_separators();

   FPair rsplit1(char Delim, size_t Flags = SPLIT_SkipWhitespace | SPLIT_SkipParenthesis) const;
   FPair split1(char Delim, size_t Flags = SPLIT_SkipWhitespace | SPLIT_SkipParenthesis) const;

   operator std::string () const { return to_str();  };

   char const &operator []( size_t i ) const { assert( i < size() ); return first[i]; };
protected:
};

std::ostream &operator << ( std::ostream &out, string_slice const &s );




typedef TArrayFix<string_slice, nMaxStrSplit>
   split_result;
typedef TArray<string_slice, nMaxStrSplit>
   long_split_result;

struct FCodeLoc{
   size_t iLine;
   FCodeLoc( size_t iLine_ ) : iLine(iLine_) {};
};

struct FLine{
   string_slice
      Text;
   FCodeLoc
      Loc; // original location (line number) in input.
   FLine(string_slice const &Text_, FCodeLoc Loc_) : Text(Text_), Loc(Loc_) {}
};
typedef std::vector<FLine>
   FLineList;
typedef FLineList::const_iterator
   cline_it;
std::ostream &operator << ( std::ostream &out, FLine const &Line );

void SplitLinesAndRemoveComments( FLineList &out, std::string const &Text );


cstr_it FindIdentifierEnd( cstr_it iPos, cstr_it iEnd );
char const *FindClosingParenthesis( char Open, char Close, char const *pFirst, char const *pLast );
cstr_it SkipWhiteSpace( cstr_it iPos, cstr_it iEnd );


struct FPropertyStr {
   string_slice
      Name, Content;
};

enum FPropertyListParseOptions {
   PROPLIST_AllowEntriesWithoutName = 0x01
};

struct FPropertyListStr : public TArray<FPropertyStr, nMaxStrSplit>
{
   string_slice
      Name;
   // Flags: bit combination of PROPLIST_*
   explicit FPropertyListStr(string_slice const &Input, char const *pAssumedName = 0, unsigned Flags = 0, char const *pOpenDelimClose = "{;}");
};


// tests if c is ' ' or '\t' (at the moment). Funky name is for avoiding conflicts
// with other functions defined for this purpose.
bool is_whitespace_cxp1(char c);

template<class FCharPred>
cstr_it string_slice::find_c(FCharPred const &pred, size_t offset) const
{
   cstr_it it(first + offset);
   for (; it != last; ++ it) {
      if (pred(*it))
         return it;
   }
   return last;
}


template<class FCharPred>
cstr_it string_slice::rfind_c(FCharPred const &pred) const
{
   cstr_it it(last - 1);
   for ( ; it >= first; --it) {
      if (pred(*it))
         return it;
   }
   return last;
}


} // namespace ct

#endif // CX_PARSE1_H
