/* Copyright (c) 2012  Gerald Knizia
 * 
 * This file is part of the IR/WMME program
 * (See https://sites.psu.edu/knizia/)
 * 
 * IR/WMME is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * IR/WMME is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/
 */

// #include <typeinfo>
#include <stdexcept>
// #include <ostream>
#include <fstream>
// #include <algorithm>
// #include <vector>
// #include <map>
// #include <iomanip>

// #include <cmath>
#include <cstring>
#include <sstream>
#include <iostream>
// #include <boost/format.hpp>
#include "format.h"

#include "CxParseSd.h"


namespace ct {

FSdfParseError::FSdfParseError(std::string What, char const *pWhere)
   : std::runtime_error(What + std::string(pWhere,pWhere+std::min(std::strlen(pWhere), (std::size_t)72u)) )
{}



// assumes that *pFirst == ChOpen. Returns pLast if no closing braket found.
// If pLast == 0, pFirst is assumed to be 0-terminated (i.e., if 0 is found, 0 is returned).
char const *FindClosingParenthesis1( char Open, char Close, char const *pFirst, char const *pLast )
{
   assert( *pFirst == Open );
   assert( pFirst < pLast || pLast == 0 );
   assert( Open != Close );
   int
      // number of brackets still open.
      iLevel = 0;
   for ( ; !pLast || pFirst != pLast; ++ pFirst ){
      if ( !pLast && *pFirst == 0 )
         return 0;
      if ( *pFirst == Open )
         iLevel += 1;
      if ( *pFirst == Close )
         iLevel -= 1;
      if ( iLevel == 0 )
         break;
   }
   return pFirst;
}


#ifdef _DEBUG
    FParseStackGuard::FParseStackGuard( FSdfParseContext *pc )
        : pContext(pc), pNode(pc->GetNode()), iLockedLevel( pc->iStackDepth() )
    {};

    FParseStackGuard::~FParseStackGuard()
    {
        if ( iLockedLevel != pContext->iStackDepth() ){
            std::cerr << "Parse stack balance error while processing nodes\n"
                      << "   " << std::string(pNode->pText, pNode->pText + 40 ) << "...\n"
                      << "Balance is:   iLocked=" << iLockedLevel << "  Actual=" << pContext->iStackDepth() << std::endl;
            assert(0);
        }
    };
#endif


void FSdfParseContext::BeginParse1( FNode *pNode ){
   assert( ParseStack.empty() );
   ParseStack.reserve(20);
   pCurrNode = pNode;
}

// enter current node, regardless of its text.
void FSdfParseContext::Enter(){
   if ( !pCurrNode || !pCurrNode->pSub)
      AssertNodeExistsFail();
   ParseStack.push_back(pCurrNode);
   pCurrNode = pCurrNode->pSub;
}

// leaves a previously entered super-node and advances node
// on upper level.
void FSdfParseContext::Leave(){
   assert( !ParseStack.empty() );
   pCurrNode = ParseStack.back();
   ParseStack.pop_back();
   Next();
}

void FSdfParseContext::EndParse1(){
   assert( ParseStack.empty() );
}

// return number of chained nodes on this level. Includes current node.
std::size_t FSdfParseContext::nNodesOnLevel(){
   std::size_t
      n = 0;
   FNode
      *p = pCurrNode;
   while ( p ) {
      n += 1;
      p = p->pNext;
   }
   return n;
};








void FSdfParseContext::InvalidDataFail( std::string const &Message ){
    std::stringstream
        msg;
    msg << "While parsing .sd structured data file:\n   ";
    msg << Message << "\n";
    msg << "Parse stack:\n";
    FNode
        *pNode = pCurrNode;
    for ( uint i = 0; i <= ParseStack.size(); ++ i ){
        msg << fmt::format("   {:>3}   ", i);
        if ( pNode ) {
            char const
                *pOutputEnd = pNode->pTextEnd;
            for ( uint k = 0; k < 40 && *pOutputEnd != 0; ++ k )
                ++ pOutputEnd;
            msg << fmt::format("{:<20}'{}...\n", pNode->ToString(), std::string(pNode->pTextEnd, pOutputEnd));
        } else {
            msg << "(none)\n";
        };
        if ( i != ParseStack.size() )
            pNode = ParseStack[ParseStack.size() - i - 1];
    }
    throw std::runtime_error(msg.str() + "\nFSdfParseContext::InvalidDataFail");
}
// (explicitly provided to avoid std::string construction in calls to this
//  functions with string literal arguments)
void FSdfParseContext::InvalidDataFail( char const *pMessage ){
    return InvalidDataFail( std::string(pMessage) );
}

void FSdfParseContext::AssertNodeExistsFail() {
    InvalidDataFail( "Attempted to enter a node where none exists." );
}
void FSdfParseContext::WrongNodeFail( char const *pExpected ) {
    InvalidDataFail( "Expected '" + std::string(pExpected) + "' node."  );
}

void FSdfParseContext::StripWhitespace( char *pText, char *&pTextEnd )
{
    // strip whitespace after < and >.
    char
        *pIn = pText, *pOut = pText;
    bool
        Strip = true;
    for ( ; ; ++pIn, ++pOut ){
        if ( Strip ) {
            while ( pIn != pTextEnd && (*pIn == ' ' || *pIn == '\n' ||
                    *pIn == '\t' || *pIn == '\r') )
                ++pIn;
        }
        if ( pIn == pTextEnd )
            break;
        Strip = *pIn == '<' || *pIn == '>';
        *pOut = *pIn;
    }
    pTextEnd = pOut;
}

void FSdfParseContext::Parse0( FNode **ppNode, char const *&pText, char const *pTextEnd )
{
    bool
        FirstOnLevel = true;
    // loop over nodes at current level.
    for ( ; ; ){
        *ppNode = 0;
        char const
            *pNodeStart = pText;
        char
            iBraket = FindNextBraket( pText, pTextEnd );
        // any data there? [changed: generate empty nodes for empty data now.]
        if ( iBraket == 0 || (!FirstOnLevel && iBraket == '>' && pText == pNodeStart) )
            // nop.
            return;

        m_Memory.Alloc(*ppNode);
        FNode
            *pNode = *ppNode;
        pNode->pSub = 0;
        pNode->pNext = 0;
        pNode->pText = pNodeStart;
        pNode->pTextEnd = pText;
        if ( iBraket == '>' )
            // note that this closing braket does not belong to the current
            // level (there was no '<' for it!), but rather to one level higher.
            // Therefore we don't absorb it by incrementing pText.
            return;
        if ( iBraket == '<' ) {
            // start of pNode arguments. pText is supposed to halt on
            // '>' of current '<' (the first char not anymore belonging
            // to SubNode).
            ++pText;
            Parse0( &pNode->pSub, pText, pTextEnd );
            if ( *pText != '>' )
                throw FSdfParseError("Imbalanced angle brakets", pNode->pText );
            ++pText;
        }

        ppNode = &pNode->pNext;
        FirstOnLevel = false;
    }
}

void FSdfParseContext::StripAndParse( FNode **ppNode, char *pText, char *pTextEnd)
{
    StripWhitespace(pText, pTextEnd); // <- this modifies pText and pTextEnd.
    *pTextEnd = 0; // make new end of string.

#if 0
   {
      std::ofstream out("/tmp/strip0", std::ios::out | std::ios::binary);
      out.write(pText, pTextEnd-pText);
   }
#endif
    Parse0(ppNode, *const_cast<char const**>(&pText), *const_cast<char const**>(&pTextEnd));
}

void ReadFileToMemory(std::string &FileContent, char const *pFileName)
{
   // note: this does not perform things like line end conversions
   // and, technically, std::string is not required to store data continously.
   std::ifstream in(pFileName, std::ios::in | std::ios::binary);
   if (!in)
      throw std::runtime_error(fmt::format("Failed to open file '{}' for reading.", pFileName));
   in.seekg(0, std::ios::end);
   std::size_t
      FileSize = in.tellg();
   in.seekg(0, std::ios::beg);
   FileContent.resize(FileSize, 0);
   in.read(&FileContent[0], FileSize);
   if (!in)
      throw std::runtime_error(fmt::format("Failed read file file '{}s'.", pFileName));
}

void FSdfParseContext::ParseFile(FNode **ppNode, char const *pFileName)
{
   // read file to memory.
   m_TextContainer.clear();
   ReadFileToMemory(m_TextContainer, pFileName);
   StripAndParse(ppNode, &m_TextContainer[0], &m_TextContainer[0] + m_TextContainer.size());
}


void FSdfParseContext::DumpTree( std::ostream &xout, FNode *pNode, uint iLevel )
{
    while ( pNode != 0 ){
        for ( uint i = 0; i < 2 * iLevel; ++ i )
            xout << ' ';
        if ( pNode->pText != pNode->pTextEnd ) {
            std::cout << "'" << std::string(pNode->pText,pNode->pTextEnd) << std::endl;
        } else {
            std::cout << "(null)" << std::endl;
        }
        if ( pNode->pSub != 0 ){
            DumpTree( xout, pNode->pSub, iLevel + 1 );
        }
        pNode = pNode->pNext;
    }
}

void FSdfParseContext::ReadData( std::string &Out ){
    Out.assign( pCurrNode->pText, pCurrNode->pTextEnd );
}

// read data from current node /and all sub-nodes, un-interpreted/ to Out.
void FSdfParseContext::ReadFullNodeText( std::string &Out )
{
    // find text end of current node and all sub-nodes.
    char const
        *pTextEnd;
    pTextEnd = FindClosingParenthesis1('<', '>', pCurrNode->pText - 1);
    assert(*(pCurrNode->pText-1) == '<' && *pTextEnd == '>');
    Out.assign(pCurrNode->pText, pTextEnd);
}


void FSdfParseContext::ReadData( double &Out ){
    Out = std::atof( pCurrNode->pText );
}

// template<class T>
// inline char const *FSdfParseContext::ReadIntAndAdvance(T &Out, char const *p)
// {
//     Out = 0;
//     assert( '1' - '0' == 1 && '9' - '0' == 9 );
//     while ( *p >= '0' && *p <= '9' ){
//         Out *= 10;
//         Out += *p - '0';
//         ++ p;
//     }
//     return p;
// }



// reads int without support for sign.
void FSdfParseContext::ReadData( uint &Out ){
    char const *
        p = pCurrNode->pText;
    p = ReadIntAndAdvance( Out, p );
    if ( *p != '>' )
        InvalidDataFail("Expected unsigned integer data.");
}

// reads int with support for leading minus sign (ONLY minus!)
void FSdfParseContext::ReadData( int &Out ){
    char const *
        p = pCurrNode->pText;
    bool
        Minus = false;
    if ( *p == '-' ){
        Minus = true;
        ++ p;
    };
    p = ReadIntAndAdvance( Out, p );
    if ( Minus )
        Out *= -1;
    if ( *p != '>' )
        InvalidDataFail("Expected unsigned integer data.");
}


} // namespace ct
