/* Copyright (c) 2012  Gerald Knizia
 * 
 * This file is part of the IR/WMME program
 * (See https://sites.psu.edu/knizia/)
 * 
 * IR/WMME is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * IR/WMME is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/
 */

// structured data file parser
#ifndef CX_PARSE_SD
#define CX_PARSE_SD

#include <string>
#include <vector>
#include <stdexcept>
#include <cstdlib> // for atof/strtod
#include <cctype> // for isspace

#include "CxTypes.h"
#include "CxMemoryStack.h"
#include "CxFixedSizeArray.h"
#include "CxPodArray.h"

namespace ct {

struct FSdfParseError : public std::runtime_error{
    FSdfParseError( std::string What, char const *pWhere = 0 );
};

unsigned const nMaxRank = 14; // fixme: remove this.

// Used to pull apart .itfca compiled algorithm files.
// Acts as object for both parse context and parse state...  split?
struct FSdfParseContext
{
//     typedef ::FCodeLoc FCodeLoc;

    FSdfParseContext( FMemoryStack *pMemoryStack )
        : m_Memory(*pMemoryStack)
    {};

    struct FNode{
        char const
            // contains either node name or node data.
            *pText, *pTextEnd;
        FNode
            // these are singly linked lists. pNext links on current level,
            // pSub onto sub-levels of the current level.
            *pNext, *pSub;

        // function for comparing node content to a compile time string literal.
        template<std::size_t N>
        bool IsEq( char const (&What)[N] ){
            // note: -1s are for implicit 0-termination in 'What'.
            if ( pTextEnd - pText != N - 1 )
                return false;
            for ( uint i = 0; i < N-1; ++ i )
                if ( What[i] != pText[i] )
                    return false;
            return true;
        }
        std::string ToString() const {
            return std::string(pText, pTextEnd);
        }
    };

    // removes whitespace between nodes *in place*.
    // pTextEnd: [in] end of input text [out] end of stripped text.
    void StripWhitespace( char *pText, char *&pTextEnd );
    // Performs initial parse stage: Splits input text into nodes.
    // Nodes are allocated on m_Memory stack.
    void Parse0( FNode **ppNode, char const *&pText, char const *pTextEnd );
    // as Parse0, but strip whitespace *in place* first.
    void StripAndParse( FNode **ppNode, char *pText, char *pTextEnd);
    // read from file first.
    void ParseFile(FNode **ppNode, char const *pFileName);
    // output parse tree to given stream. (debug function)
    void DumpTree( std::ostream &out, FNode *pNode, uint iLevel = 0 );

protected:
    FMemoryStack
        &m_Memory;
    typedef std::vector<FNode*>
        FParseStack;
    FParseStack
        ParseStack;
    FNode
        // currently parsed node. Set to 0 if Next() reached the end of
        // a node chain on the current level.
        *pCurrNode;

    void AssertNodeExistsFail();
    void WrongNodeFail( char const *pExpected );

    std::string
        m_TextContainer; // used for storage of file fontent if ParseFile() is used.
public:
    // ReadData functions: read data from current node to output object.
    // WARNING: these are written in such a way that they are only safe
    //    if the total input string is explicitly 0-terminated! Otherwise
    //    invalid input may crash the program!
    void ReadData( std::string &Out );
    void ReadData( double &Out );
    // reads int without support for sign.
    void ReadData( uint &Out );
    // reads int with support for leading minus sign (ONLY minus!)
    void ReadData( int &Out );


    // generic "read container" (reads long lists)
    // (needs a ReadDataAndAdvance function for the type.)
    template<typename TArray>
    void ReadArray( TArray &Out, unsigned nMaxEntries );


    // reads long list (i.e., with space between elements, and
    // thus escapting unnecessary)
    template<typename T, unsigned nMaxRank>
    void ReadData( TArrayFix<T,nMaxRank> &Out );
    template<typename T>
    void ReadData( std::vector<T> &Out );
    template<typename T>
    void ReadData( TArray<T> &Out );


    // reads short integer list (i.e., no separators between ints,
    // 1-char-1-int, unless int is escapted in '{}'.
    // E.g., 012{99}4 -> [0,1,2,99,4].
    // (note: these are partial specializations)
    template<unsigned nMaxRank>
    void ReadData( TArrayFix<unsigned char, nMaxRank> &Out );
    template<unsigned nMaxRank>
    void ReadData( TArrayFix<signed char, nMaxRank> &Out );


    // reads data from current node /and all sub-nodes/ to Out
    void ReadFullNodeText( std::string &Out );

    // reads unsigned integer data and advances p. returns first
    // char * which is not part of the int (not a digit)
    template<class T>
    char const *ReadIntAndAdvance( T &Out, char const *p );

    template<class T, uint N>
    char const *ReadIntListAndAdvance( TArrayFix<T,N> &Out, char const *p );

    template<class T>
    char const *ReadFloatAndAdvance( T &Out, char const *p );
    // todo: make types for list-of-ints, list-of-uchars,
    // list-of-list-of-uchars.

    char const *ReadDataAndAdvance(float &Out, char const *p) { return ReadFloatAndAdvance(Out, p); }
    char const *ReadDataAndAdvance(double &Out, char const *p) { return ReadFloatAndAdvance(Out, p); }

    char const *ReadDataAndAdvance(int &Out, char const *p) { return ReadIntAndAdvance(Out, p); }
    char const *ReadDataAndAdvance(unsigned &Out, char const *p) { return ReadIntAndAdvance(Out, p); }
    char const *ReadDataAndAdvance(signed char &Out, char const *p) { return ReadIntAndAdvance(Out, p); }
    char const *ReadDataAndAdvance(unsigned char &Out, char const *p) { return ReadIntAndAdvance(Out, p); }
    char const *ReadDataAndAdvance(long &Out, char const *p) { return ReadIntAndAdvance(Out, p); }
    char const *ReadDataAndAdvance(unsigned long &Out, char const *p) { return ReadIntAndAdvance(Out, p); }
public:
    void InvalidDataFail( char const *pMessage );
    void InvalidDataFail( std::string const &Message );

    // TODO: move these definitions out of the class.

    void BeginParse1( FNode *pNode );

    bool NodeExists(){
        return pCurrNode != 0;
    };

    FNode *GetNode() {
        return pCurrNode;
    };

    // advance current node on current level.
    void Next(){
        if ( pCurrNode )
            pCurrNode = pCurrNode->pNext;
    };

    // checks if current node has 'What' as text. If yes, enters node
    // (pushes its sub-node to top of stack) and returns true. Otherwise
    // returns false.
    template<std::size_t N>
    bool TryEnter( char const (&What)[N] ){
        if ( !pCurrNode || !pCurrNode->IsEq(What) )
            return false;
        ParseStack.push_back(pCurrNode);
        pCurrNode = pCurrNode->pSub;
        if ( !pCurrNode )
            AssertNodeExistsFail();
        return true;
    }

    // As the Try* version, but bails out if the expected node is not found.
    template<std::size_t N>
    void Enter( char const (&What)[N] ){
        if ( !TryEnter(What) )
            WrongNodeFail(What);
    }

    // enter current node, regardless of its text.
    void Enter();

    // leaves a previously entered super-node and advances node
    // on upper level.
    void Leave();

    template<class T, std::size_t N>
    bool TryRead( T &Out, char const (&What)[N] ){
        if ( !TryEnter(What) )
            return false;
        ReadData(Out);
        Leave();
        return true;
    }

    template<class T, std::size_t N>
    void Read( T &Out, char const (&What)[N] ){
        Enter(What);
        ReadData(Out);
        Leave();
    }

    void EndParse1();

    // return number of chained nodes on this level. Includes current node.
    std::size_t nNodesOnLevel();

    // debug function.
    uint iStackDepth() {
        return ParseStack.size();
    };

protected:
    static inline char FindNextBraket(char const *&p, char const *pEnd)
    {
        // inrement p until either
        //   '<' is found (then return '<'),
        //   '>' is found (then return '>')
        // or p==pEnd (then return 0).
        for ( ; p != pEnd; ++ p ){
            if ( *p == '<' ) return *p;
            if ( *p == '>' ) return *p;
        }
        return 0;
    }
};

// Asserts that stack depth at construction is equal to stack depth
// at destruction. (Helper to find parse stack imbalances).
struct FParseStackGuard
{
#ifdef _DEBUG
    FParseStackGuard( FSdfParseContext *pc );
    ~FParseStackGuard();

    FSdfParseContext
        *pContext;
    FSdfParseContext::FNode
        *pNode;
    uint
        iLockedLevel;
#else
    FParseStackGuard( FSdfParseContext *pc )
    {};
#endif
};

// assumes that *pFirst == ChOpen. Returns pLast if no closing braket found.
// If pLast == 0, pFirst is assumed to be 0-terminated (i.e., if 0 is found, 0 is returned).
char const *FindClosingParenthesis1( char Open, char Close, char const *pFirst, char const *pLast = 0 );

template<typename TArray>
void FSdfParseContext::ReadArray( TArray &Out, unsigned nMaxEntries )
{
    char const
        *p = pCurrNode->pText, *q;
    Out.clear();

    while ( *p && *p != '>' ){
        if ( nMaxEntries != 0 && Out.size() == nMaxEntries )
            InvalidDataFail("Input list too long.");
        if ( *p == ' ' )
            ++p;
        typename TArray::value_type
            Value;
        q = ReadDataAndAdvance(Value, p);
        Out.push_back(Value);
        if ( p == q )
            InvalidDataFail("Unexpected data type.");
        p = q;
    }
}

template<typename T, unsigned nMaxRank>
void FSdfParseContext::ReadData( TArrayFix<T,nMaxRank> &Out )
{
    return ReadArray(Out, nMaxRank);
}


template<typename T>
void FSdfParseContext::ReadData( std::vector<T> &Out )
{
    return ReadArray(Out, 0);
}

template<typename T>
void FSdfParseContext::ReadData( TArray<T> &Out )
{
    return ReadArray(Out, 0);
}


template<unsigned nMaxRank>
void FSdfParseContext::ReadData( TArrayFix<unsigned char, nMaxRank> &Out ){
    char const
        *p = pCurrNode->pText;
    p = ReadIntListAndAdvance( Out, p );
    if ( *p != '>' )
        InvalidDataFail("Integer short list");
}

template<unsigned nMaxRank>
void FSdfParseContext::ReadData( TArrayFix<signed char, nMaxRank> &Out ){
    // yes, this is an exact copy of the <uchar> function. Short lists
    // don't support signs yet.
    char const
        *p = pCurrNode->pText;
    p = ReadIntListAndAdvance( Out, p );
    if ( *p != '>' )
        InvalidDataFail("Integer short list");
}


template<class T>
char const *FSdfParseContext::ReadIntAndAdvance(T &Out, char const *p)
{
    Out = 0;
    if ( *p && *p != '>' && !std::isspace(*p)){
        Out = static_cast<T>(std::strtol(p, const_cast<char**>(&p), 0));
        while (*p && *p != ' ' && *p != '>')
            ++p;
    }
    return p;
}


template<class T>
const char* FSdfParseContext::ReadFloatAndAdvance(T& Out, const char* p)
{
    Out = 0;
    if ( *p && *p != '>' && !std::isspace(*p)){
//         if (*p == ' ')
//             ++p; // cgk: should this be here?
                    // I'm also not sure if strtod skipping whitespace is a good idea...
        Out = static_cast<T>(std::strtod(p, const_cast<char**>(&p)));
        while (*p && *p != ' ' && *p != '>')
            ++p;
    }
    return p;
}


template<class T, uint N>
char const *FSdfParseContext::ReadIntListAndAdvance( TArrayFix<T,N> &Out, char const *p )
{
    // read a int short list
    Out.clear();

    // note: empty list is indicated with a '*'.
    if ( *p == '*' )
        return p + 1;

    for ( ; ; ++ p ){
        if ( *p >= '0' && *p <= '9' ) {
            if ( Out.size() == N ) InvalidDataFail("Integer list too long");
            Out.push_back( *p - '0' );
        }
        else if ( *p == '{' ){
            if ( Out.size() == N ) InvalidDataFail("Integer list too long");
            T val;
            // p = ReadIntAndAdvance(val,p);
            p = ReadIntAndAdvance(val,p+1); // bug removed
            if ( *p != '}' )
                InvalidDataFail("Integer escape sequence malformed");
            //++p; for loop advances p
            Out.push_back(val);
        } else
            break;
    };
    return p;
}




} // namespace ct


#endif // CX_PARSE_SD

// kate: indent-width 4
