/* Copyright (c) 2012  Gerald Knizia
 * 
 * This file is part of the IR/WMME program
 * (See https://sites.psu.edu/knizia/)
 * 
 * IR/WMME is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * IR/WMME is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/
 */

#ifndef CX_PODARRAY_H
#define CX_PODARRAY_H

#include <string.h> // for memcpy
#include <stdlib.h> // for malloc/free
#include <stdexcept> // for runtime_error
#include "format.h"

namespace ct {

// swap two primitive values. Here such that we need not include <algorithm> here.
template<class FType>
void swap1(FType &A, FType &B){
   FType t = A; A = B; B = t;
}


#ifdef __GNUC__ // g++ or anything which pretends to be g++ (e.g., clang, intel c++ on linux)
   // default number of static elements in TArray. Should be 0, but this is apparently
   // not of disputed legality in the C++ standard. VC++ doesn't like that ("illegal size 0 array")
   #define CX_DEFAULT_ARRAY_STATIC_STORAGE 0
#else
   #define CX_DEFAULT_ARRAY_STATIC_STORAGE 8
#endif



// A dynamic array of POD (``plain old data'') types that can be
// copied via a memcpy. Main point for this is that std::vector does not
// allow containing C-style arrays (e.g., double [3]), because C-style
// arrays are not assignable. Additionally, std::vector can be *very* slow
// when allocating large amounts of data, because the data is set to
// zero on resize. This class explicitly DOES NOT DO THAT: It has RAII
// semantics, but non-explicitly touched data is just random.
//
// If the template parameter 'StaticStorageSize' is zero, the class is
// effectively a 'buffer-ptr + size' pair. Otherwise, the class contains
// a local buffer for 'StaticStorageSize' elements which will be used
// for allocation-less storage of the first few elements until a dynamic
// allocation is required when the static storage size is exceeded.
template<class FType, size_t StaticStorageSize = CX_DEFAULT_ARRAY_STATIC_STORAGE>
struct TArray
{
   typedef FType *iterator;
   typedef FType const *const_iterator;
   typedef ::size_t size_type;
   typedef FType value_type;

   iterator begin() { return m_pData; }
   iterator end() { return m_pDataEnd; }
   const_iterator begin() const { return m_pData; }
   const_iterator end() const { return m_pDataEnd; }

   FType &front() { return *m_pData; }
   FType &back() { return *(m_pDataEnd-1); }
   FType const &front() const { return *m_pData; }
   FType const &back() const { return *(m_pDataEnd-1); }

   FType *data() { return m_pData; }
   FType const *data() const { return m_pData; }

   FType &operator[] (size_type i) { return m_pData[i]; }
   FType const &operator[] (size_type i) const  { return m_pData[i]; }

   size_type size() const { return m_pDataEnd - m_pData; }
   bool empty() const { return m_pData == m_pDataEnd; }

   // WARNING: contrary to std::vector, this function *DOES NOT*
   // initialize (or touch, for that matter) the newly created data.
   // If you want that behavior, use the other resize function!
   void resize(size_type n) {
      reserve(n);
      m_pDataEnd = m_pData + n;
   }


   void resize(size_type n, FType t) {
      size_type old_size = size();
      resize(n);
      if (old_size < n) {
         for ( size_type i = old_size; i < n; ++ i )
            m_pData[i] = t;
      }
   }

   void clear() {
      ReleaseBuffer();
   }

   // memset the entire array to 0.
   void clear_data() {
      memset(m_pData, 0, sizeof(FType)*size());
   }

   void resize_and_clear(size_type n) {
      resize(n);
      clear_data();
   }

   void push_back( FType const &t ) {
      if ( size() + 1 > static_cast<size_type>(m_pReservedEnd - m_pData) ) {
         reserve(2 * size() + 1);
      }
      m_pDataEnd[0] = t;
      ++m_pDataEnd;
   }

   void pop_back() {
      assert(!empty());
      m_pDataEnd -= 1;
   }

   void reserve(size_type NewReservedSize) {
      if (NewReservedSize == 0)
         return;
      if (StaticStorageSize != 0) {
         // is this the initial reservation? I.e., no data in *this so far?
         if (m_pData == 0) {
            // assign reserved space to local object static buffer.
            // This will automatically deal with the reserved size. In this case,
            // the branch below will not be taken if the reservation size happens
            // to be small enough.
            m_pData = &m_StaticStorage[0];
            m_pDataEnd = m_pData;
            m_pReservedEnd = m_pData + StaticStorageSize;
            m_DataIsOnStaticBuffer = true;
         }
      }
      if (nReserved() < NewReservedSize) {
         size_type
            nStoredElements = size();
         FType
            *pNewData = static_cast<FType*>(::malloc(sizeof(FType) * NewReservedSize));
         if (pNewData == 0) {
            throw std::runtime_error(fmt::format("TArray<{}>: memory allocation failed ({:.2f} MB requested; reallocation: {}; prev size: {:.2f} MB).",
               sizeof(value_type), fSizeMb(NewReservedSize), (m_pData != 0)? "yes" : "no", fSizeMb(nReserved())));
         }
         if (nStoredElements != 0)
            ::memcpy(pNewData, m_pData, sizeof(FType) * nStoredElements);
         ReleaseBuffer();
         m_DataIsOnStaticBuffer = false;
         m_pData = pNewData;
         m_pDataEnd = m_pData + nStoredElements;
         m_pReservedEnd = m_pData + NewReservedSize;
      }
   }


   TArray()
      : m_pData(0), m_pDataEnd(0), m_pReservedEnd(0), m_DataIsOnStaticBuffer(false)
   {}

   TArray(TArray const &other)
      : m_pData(0), m_pDataEnd(0), m_pReservedEnd(0), m_DataIsOnStaticBuffer(false)
   {
      *this = other;
   }

   // WARNING: array's content not initialized with this function!
   // (with intention!)
   explicit TArray(size_type n)
      : m_pData(0), m_pDataEnd(0), m_pReservedEnd(0), m_DataIsOnStaticBuffer(false)
   {
      resize(n);
   }

   TArray(size_type n, FType t)
      : m_pData(0), m_pDataEnd(0), m_pReservedEnd(0), m_DataIsOnStaticBuffer(false)
   {
      resize(n);
      for ( size_type i = 0; i < n; ++ i )
         m_pData[i] = t;
   }

   ~TArray() {
      ReleaseBuffer();
   }

   void operator = (TArray const &other) {
      resize(other.size());
      memcpy(m_pData, other.m_pData, sizeof(FType) * size());
   }

   void swap(TArray &other) {
      if (StaticStorageSize == 0 || (!other.m_DataIsOnStaticBuffer && !this->m_DataIsOnStaticBuffer)) {
         // no data on static buffers. Some pointer swappery is sufficient.
         assert(!other.m_DataIsOnStaticBuffer && !this->m_DataIsOnStaticBuffer);
         swap1(m_pData, other.m_pData);
         swap1(m_pDataEnd, other.m_pDataEnd);
         swap1(m_pReservedEnd, other.m_pReservedEnd);
      } else {
         // at least one of the array data sets is on static buffers. This makes
         // this complicated. We actually need to physically copy stuff around
         // in this case.
         TArray tmp(other);
         other = *this;
         *this = tmp;
      }
   }

   template<class FRandomIt>
   void assign(FRandomIt begin, FRandomIt end){
      resize(end - begin);
      for ( size_type i = 0; i < size(); ++ i )
         m_pData[i] = begin[i];
   }

   template<class FRandomIt>
   void insert(iterator pos, FRandomIt begin, FRandomIt end){
      // WARNING: not checked.
      size_type num = end - begin, ipos = pos - this->begin();
      resize(num + this->size());
      for ( size_type i = this->size(); i != ipos+num; -- i )
         this->m_pData[i-1] = this->m_pData[i-1-num];
      for ( size_type i = 0; i < num; ++ i )
         this->m_pData[ipos+i] = begin[i];
   }

private:
   FType
      // start of controlled array. may be 0 if no data is contained.
      *m_pData,
      // end of actual data
      *m_pDataEnd,
      // end of reserved space (i.e., of the space *this owns)
      *m_pReservedEnd;
   bool
      // if true, m_pData points to a dynamically allocated piece of memory
      // which needs to be freed again. If false, the pointers in *this point
      // into m_StaticStorage.
      m_DataIsOnStaticBuffer;
   FType
      // if the total number of elements is smaller than StaticStorageSize, then
      // this buffer may be used to store them.
      m_StaticStorage[StaticStorageSize];

   size_t nReserved() const {
      return static_cast<size_type>(m_pReservedEnd - m_pData);
   }

   double fSizeMb(size_t n) {
      return double(sizeof(FType) * size_t(n))/double(size_t(1) << size_t(20));
   }

   void ReleaseBuffer() {
      if ((StaticStorageSize == 0) || !m_DataIsOnStaticBuffer)
         ::free(m_pData);
      m_pData = 0;
      m_pDataEnd = 0;
      m_pReservedEnd = 0;
      // there is no data, so it's not on a static buffer. This handling simplifies
      // some operations.
      m_DataIsOnStaticBuffer = false;
   }
};

} // namespace ct

namespace std {
   template<class FType>
   void swap(ct::TArray<FType> &A, ct::TArray<FType> &B) {
      A.swap(B);
   }
}


#endif // CX_PODARRAY_H
