/* Copyright (c) 2012  Gerald Knizia
 * 
 * This file is part of the IR/WMME program
 * (See https://sites.psu.edu/knizia/)
 * 
 * IR/WMME is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 * 
 * IR/WMME is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with bfint (LICENSE). If not, see http://www.gnu.org/licenses/
 */

#include <stdexcept>
#include "IrMeta2i.h"

namespace ct {

FKrn2i::~FKrn2i()
{}

void FKrn2i::AccCoreDeriv1d(double *pGrad, double const *pRdmAC, size_t StrideA, size_t StrideC, ir::FRawShell const *pA, size_t iCenterA, ir::FRawShell const *pC, size_t iCenterC, double Prefactor, ct::FMemoryStack &Mem) const
{
   IR_SUPPRESS_UNUSED_WARNING(pGrad);
   IR_SUPPRESS_UNUSED_WARNING(pRdmAC);
   IR_SUPPRESS_UNUSED_WARNING(StrideA);
   IR_SUPPRESS_UNUSED_WARNING(StrideC);
   IR_SUPPRESS_UNUSED_WARNING(pA);
   IR_SUPPRESS_UNUSED_WARNING(iCenterA);
   IR_SUPPRESS_UNUSED_WARNING(pC);
   IR_SUPPRESS_UNUSED_WARNING(iCenterC);
   IR_SUPPRESS_UNUSED_WARNING(Prefactor);
   IR_SUPPRESS_UNUSED_WARNING(Mem);
   throw std::runtime_error("FKrn2i: 1st derivative integrals apparently not implemented for current integral type. Overwrite AccCoreDeriv1d.");
}


void FKrn2i::AccCoreDeriv2d(double *pHess, size_t iStHessA, size_t iStHessC, double const *pRdmAC, size_t StrideA, size_t StrideC, ir::FRawShell const *pA, size_t iCenterA, ir::FRawShell const *pC, size_t iCenterC, double Prefactor, ct::FMemoryStack &Mem) const
{
   IR_SUPPRESS_UNUSED_WARNING(pHess);
   IR_SUPPRESS_UNUSED_WARNING(iStHessA);
   IR_SUPPRESS_UNUSED_WARNING(iStHessC);
   IR_SUPPRESS_UNUSED_WARNING(pRdmAC);
   IR_SUPPRESS_UNUSED_WARNING(StrideA);
   IR_SUPPRESS_UNUSED_WARNING(StrideC);
   IR_SUPPRESS_UNUSED_WARNING(pA);
   IR_SUPPRESS_UNUSED_WARNING(iCenterA);
   IR_SUPPRESS_UNUSED_WARNING(pC);
   IR_SUPPRESS_UNUSED_WARNING(iCenterC);
   IR_SUPPRESS_UNUSED_WARNING(Prefactor);
   IR_SUPPRESS_UNUSED_WARNING(Mem);
   throw std::runtime_error("FKrn2i: 2nd derivative integrals apparently not implemented for current integral type. Overwrite AccCoreDeriv2d.");
}


FKrn2i_PointMultipoles::FKrn2i_PointMultipoles(ir::FIntegralKernel const *pIrKernel_, FPoint const *pPoints, double const *pCoeffs, size_t nPoints)
   : m_pIrKernel(pIrKernel_), m_Exp(1e20)
{
   m_PointShells.reserve(nPoints);
   m_PointCenters.reserve(nPoints);
   // count number of coefficients and find max angular momentum.
   size_t
      nCoeff = 0;
   m_MaxL = 0;
   for (size_t iPt = 0; iPt < nPoints; ++ iPt) {
      nCoeff += 2*pPoints[iPt].l + 1;
      m_MaxL = std::max(m_MaxL, pPoints[iPt].l);
   }
   // store the point centers and their strenghts here.
   m_Data.resize(3*nPoints + nCoeff);
   // copy centers
   for (size_t iPt = 0; iPt < nPoints; ++ iPt)
      for (uint i = 0; i < 3; ++ i)
         m_Data[3*iPt + i] = pPoints[iPt].vPos[i];
   // copy coefficients and make RawShell objects.
   std::size_t
      iCoeff = 0;
   for (size_t iPt = 0; iPt < nPoints; ++ iPt) {
      FPoint const &Pt = pPoints[iPt];
      m_PointShells.push_back(ir::FRawShell(Pt.l, &m_Exp, 1,
         &m_Data[3*nPoints+iCoeff], 1, &m_Data[3*iPt]));
      m_PointCenters.push_back(Pt.iCenter);

      double
         // This works for point charges. I think it should be the
         // right factor for higher multipoles, too.
         fInvNorm = std::pow(m_Exp/M_PI, 3./2.);
      if (Pt.l != 0)
         throw std::runtime_error("case Pt.l != 0 not tested. Not 100% sure if norm factor is correct!");
      for (size_t iSh = 0; iSh < 2*Pt.l + 1; ++ iSh)
         m_Data[3*nPoints + iCoeff + iSh] = fInvNorm * pCoeffs[iCoeff + iSh];
      iCoeff += 2*Pt.l+1;
   }
   assert(iCoeff == nCoeff);

   // by default construct all derivatives.
   SetDerivativeFlags(DERIVFLAG_IncludeShellDerivsAB | DERIVFLAG_IncludePointDerivsP);
}

FKrn2i_PointMultipoles::~FKrn2i_PointMultipoles()
{}


void FKrn2i_PointMultipoles::SetDerivativeFlags(unsigned Flags_)
{
   m_DerivativeFlags = Flags_;
}


void FKrn2i_PointMultipoles::EvalInt2e2c(double *pOut, size_t StrideA, size_t StrideB, ir::FRawShell const *pA, ir::FRawShell const *pB, double Prefactor, bool Add, ct::FMemoryStack &Mem) const
{
   if (!Add)
      for (size_t iB = 0; iB != pB->nFn(); ++ iB)
         for (size_t iA = 0; iA != pA->nFn(); ++ iA)
            pOut[StrideA*iA + StrideB*iB] = 0.;
   double
      *pIntData = Mem.AllocN(pA->nFn() * pB->nFn() * (2*m_MaxL+1), (double)0.);
   for (size_t iPt = 0; iPt < m_PointShells.size(); ++ iPt) {
      size_t
         Strides3[3] = {1, pA->nFn(), pA->nFn()*pB->nFn()};
      // TODO:
      //   - do this with a specialized routine which takes ZetaC == +inf
      //     into account. Can be done much cheaper then.
      //   - use the C-inline-contracting version of EvalInt2e3c: EvalInt2e3c_ContractC.
      //     we actually have it implemented!!
      ir::FRawShell const
         *pShPt = &m_PointShells[iPt];
      EvalInt2e3c(pIntData, Strides3,
         pA, pB, pShPt, 1, Prefactor, m_pIrKernel, Mem);
      for (size_t iB = 0; iB != pB->nFn(); ++ iB)
         for (size_t iA = 0; iA != pA->nFn(); ++ iA)
            for (size_t iCo = 0; iCo < pShPt->nSh(); ++ iCo)
               pOut[iA*StrideA + iB*StrideB] += pIntData[iA*Strides3[0] + iB*Strides3[1] + iCo*Strides3[2]];
   }
   Mem.Free(pIntData);
}




FKrn2i_Direct::FKrn2i_Direct(ir::FIntegralKernel const *pIrKernel_, uint LaplaceOrder, double Prefactor)
   : m_pIrKernel(pIrKernel_), m_LaplaceOrder(LaplaceOrder), m_Prefactor(Prefactor)
{}

FKrn2i_Direct::~FKrn2i_Direct()
{}

void FKrn2i_Direct::EvalInt2e2c(double *pOut, size_t StrideA, size_t StrideC, ir::FRawShell const *pA, ir::FRawShell const *pC, double Prefactor, bool Add, ct::FMemoryStack &Mem) const
{
   ir::EvalInt2e2c_LaplaceC(pOut, StrideA, StrideC, pA, pC, m_Prefactor * Prefactor, Add,
      m_LaplaceOrder, m_pIrKernel, Mem);
}

void FKrn2i_Direct::AccCoreDeriv1d(double *pGrad, double const *pRdmAC, size_t StrideA, size_t StrideC, ir::FRawShell const *pA, size_t iCenterA, ir::FRawShell const *pC, size_t iCenterC, double Prefactor, ct::FMemoryStack &Mem) const
{
   void
      *pBaseOfMemory = Mem.Alloc(0);
   size_t
      nFnA = pA->nFn(),
      nFnC = pC->nFn();
   double
      *pDerivA, *pDerivC;
   Mem.Alloc(pDerivA, nFnA * nFnC * 3);
   Mem.Alloc(pDerivC, nFnA * nFnC * 3);

   // make the actual gradient integrals (yes, we actually calculate them explcitly.).
   ir::EvalInt2e2c1d_LaplaceC(pDerivA, pDerivC, 1, nFnA, nFnA*nFnC, pA, pC, m_Prefactor * Prefactor, false, m_LaplaceOrder, m_pIrKernel, Mem);

   // contract them with their section of the density matrix.
   for (size_t iFnC = 0; iFnC < nFnC; ++ iFnC)
      for (size_t iFnA = 0; iFnA < nFnA; ++ iFnA) {
         double RdmAC = pRdmAC[StrideA * iFnA + StrideC * iFnC];
         for (size_t ixyz = 0; ixyz != 3; ++ ixyz) {
            pGrad[3*iCenterA + ixyz] += RdmAC * pDerivA[iFnA + nFnA*(iFnC + nFnC * ixyz)];
            pGrad[3*iCenterC + ixyz] += RdmAC * pDerivC[iFnA + nFnA*(iFnC + nFnC * ixyz)];
//             pGrad[3*iCenterC + ixyz] -= RdmAC * pDerivA[iFnA + nFnA*(iFnC + nFnC * ixyz)];
         }
      }

   Mem.Free(pBaseOfMemory);
}

FKrn2i_Meta::FKrn2i_Meta()
{}

FKrn2i_Meta::~FKrn2i_Meta()
{}

void FKrn2i_Meta::AddKernel(FKrn2iPtr pKrn2i, double fPrefactor)
{
   if (fPrefactor != 0.)
      m_Kernels.push_back(FKernelEntry(pKrn2i, fPrefactor));
}


void FKrn2i_Meta::EvalInt2e2c(double *pOut, size_t StrideA, size_t StrideC, ir::FRawShell const *pA, ir::FRawShell const *pC, double Prefactor, bool Add, ct::FMemoryStack &Mem) const
{
   FKernelList::const_iterator
      itKrn2i;
   for (itKrn2i = m_Kernels.begin(); itKrn2i != m_Kernels.end(); ++ itKrn2i) {
      bool AddThisOne = Add || (itKrn2i != m_Kernels.begin());
      itKrn2i->p->EvalInt2e2c(pOut, StrideA, StrideC, pA, pC, Prefactor * itKrn2i->Factor, AddThisOne, Mem);
   }
}


void FKrn2i_Meta::AccCoreDeriv1d(double *pGrad, double const *pRdmAC, size_t StrideA, size_t StrideC, ir::FRawShell const *pA, size_t iCenterA, ir::FRawShell const *pC, size_t iCenterC, double Prefactor, ct::FMemoryStack &Mem) const
{
   FKernelList::const_iterator
      itKrn2i;
   for (itKrn2i = m_Kernels.begin(); itKrn2i != m_Kernels.end(); ++ itKrn2i) {
      itKrn2i->p->AccCoreDeriv1d(pGrad, pRdmAC, StrideA, StrideC, pA, iCenterA, pC, iCenterC, Prefactor * itKrn2i->Factor, Mem);
   }
}



#ifdef IR_ECP
FKrn2i_1CenterEcp::FKrn2i_1CenterEcp(FVector3 vEcpCenter, ptrdiff_t iEcpCenter, ir::FAtomEcp const *pIrEcp)
   : m_pIrEcp(pIrEcp), m_iEcpCenter(iEcpCenter), m_vEcpCenter(vEcpCenter)
{}

FKrn2i_1CenterEcp::~FKrn2i_1CenterEcp()
{}

void FKrn2i_1CenterEcp::EvalInt2e2c(double *pOut, size_t StrideA, size_t StrideC, ir::FRawShell const *pA, ir::FRawShell const *pC, double Prefactor, bool Add, ct::FMemoryStack &Mem) const
{
   throw std::runtime_error("This version of MicroScf was compiled without ECP support.");
}

#endif // IR_ECP

void FRawBasis::Finalize(size_t const *pShCen)
{
   ShellOffsets.clear();
   CenterOffsets.clear();
   ShellCenters.clear();

   ShellOffsets.reserve(Shells.size() + 1);
   ShellOffsets.push_back(0);
   CenterOffsets.reserve(Shells.size() + 1);
   ShellCenters.reserve(Shells.size());
   nMaxFnPerShell = 0;
   size_t
      iShellOff = 0;
   for (size_t iSh = 0; iSh < Shells.size(); ++ iSh){
      while (pShCen[iSh] >= (size_t)CenterOffsets.size())
         CenterOffsets.push_back(iShellOff);
      iShellOff += 1;
      size_t
         nShFn = Shells[iSh].nFn();
      ShellOffsets.push_back(ShellOffsets.back() + nShFn);
      nMaxFnPerShell = std::max((size_t)nMaxFnPerShell, nShFn);
      ShellCenters.push_back(pShCen[iSh]);
   }
   assert(ShellOffsets.back() == this->nFn());
   CenterOffsets.push_back(Shells.size());
}



void MakeIntMatrix( double *pDest, size_t iRowSt, size_t iColSt, FRawBasis const &BasisRow,
   FRawBasis const &BasisCol, FKrn2i const &Krn2i, FMemoryStack &Mem_, double Prefactor, bool Add )
{
   bool
      MatrixSymmetric = (&BasisRow == &BasisCol);

   FMemoryStackArray MemStacks(Mem_);

//          Mem.Alloc(pIntResult, BasisRow.nMaxFnPerShell * BasisCol.nMaxFnPerShell);
   #pragma omp parallel for schedule(dynamic)
   for (int iShellB_ = 0; iShellB_ < int(BasisCol.Shells.size()); ++ iShellB_) {
      FMemoryStack &Mem = MemStacks.GetStackOfThread();
      size_t iShellB = size_t(iShellB_);
      ir::FRawShell const
         *pShellB = &BasisCol.Shells[iShellB];
      for (size_t iShellA = (!MatrixSymmetric)? 0 : iShellB;
           iShellA < BasisRow.Shells.size(); ++ iShellA)
      {
         ir::FRawShell const
            *pShellA = &BasisRow.Shells[iShellA];
         size_t
            nSizeA = pShellA->nFn(),
            nSizeB = pShellB->nFn();
         assert(nSizeA == BasisRow.nFn(iShellA) &&
                nSizeB == BasisCol.nFn(iShellB));
         TMemoryLock<double>
            pIntResult(nSizeA * nSizeB, &Mem);

         Krn2i.EvalInt2e2c(pIntResult, 1, nSizeA, pShellA, pShellB, Prefactor, false, Mem);

         // fill data we gathered into the matrices
         for ( size_t j_ = 0; j_ < nSizeB; ++ j_ )
            for ( size_t i_ = 0; i_ < nSizeA; ++ i_ ) {
               size_t
                  i = BasisRow.iFn(iShellA) + i_,
                  j = BasisCol.iFn(iShellB) + j_;
               double const
                  &r = pIntResult[i_ + nSizeA * j_];
               if (!Add) {
                  pDest[i*iRowSt + j*iColSt] = r;
               } else {
                  pDest[i*iRowSt + j*iColSt] += r;
               }
            }
      }
   }
   if (MatrixSymmetric) {
      // copy newly computed triangle to other triangle (this version should have continuous
      // writes for standard layout matrices).
      size_t nFn = BasisRow.nFn();
      for (size_t i = 0; i < nFn; ++ i)
         for (size_t j = 0; j < i; ++ j)
            pDest[j*iRowSt + i*iColSt] = pDest[i*iRowSt + j*iColSt];
   }
}


void AccGradient2ix(double *pGrad, double *pRdm, size_t iRowSt, size_t iColSt, FRawBasis const &BasisRow,
   FRawBasis const &BasisCol, FKrn2i const &Krn2i, FMemoryStack &Mem_, double Prefactor)
{
   bool
      MatrixSymmetric = (&BasisRow == &BasisCol);

   assert(BasisRow.nCen() == BasisCol.nCen());
   FOmpAccBlock
      Gradient(pGrad, 3*BasisRow.nCen(), 0, Mem_);
   FMemoryStackArray MemStacks(Mem_);

   #pragma omp parallel for schedule(dynamic)
   for (int iShellA_ = 0; iShellA_ < int(BasisRow.Shells.size()); ++ iShellA_) {
      size_t iShellA = size_t(iShellA_);
      FMemoryStack &Mem = MemStacks.GetStackOfThread();
      ir::FRawShell const
         *pShellA = &BasisRow.Shells[iShellA];
      for (size_t iShellB = (!MatrixSymmetric)? 0 : iShellA;
           iShellB < BasisCol.Shells.size(); ++ iShellB)
      {
         ir::FRawShell const
            *pShellB = &BasisCol.Shells[iShellB];
         double
            Factor1 = Prefactor;
         if (MatrixSymmetric && iShellA != iShellB)
            Factor1 *= 2.0;
         double const
            *pRdmAB = &pRdm[iRowSt * BasisRow.iFn(iShellA) + iColSt * BasisCol.iFn(iShellB)];
         Krn2i.AccCoreDeriv1d(Gradient.pTls(), pRdmAB, iRowSt, iColSt,
            pShellA, BasisRow.iShCen(iShellA), pShellB, BasisCol.iShCen(iShellB), Factor1, Mem);
      }
   }
   Gradient.Join();
}


void AccHessian2ix(double *pHess, size_t iRowStH, size_t iColStH,  double *pRdm, size_t iRowStD, size_t iColStD,
   FRawBasis const &BasisRow, FRawBasis const &BasisCol, FKrn2i const &Krn2i, FMemoryStack &Mem_, double Prefactor)
{
   bool
      MatrixSymmetric = (&BasisRow == &BasisCol);

   assert(BasisRow.nCen() == BasisCol.nCen());
   assert_rt(iColStH == BasisCol.nCen()*3);
//    size_t nHessSize = iColStH * (BasisCol.nCen()*3);

//    FOmpAccBlock
//       Hessian(pHess, nHessSize, 0, Mem_);
//    FMemoryStackArray MemStacks(Mem_);

//    #pragma omp parallel for schedule(dynamic)
   for (int iShellA_ = 0; iShellA_ < int(BasisRow.Shells.size()); ++ iShellA_) {
      size_t iShellA = size_t(iShellA_);
//       FMemoryStack &Mem = MemStacks.GetStackOfThread();
      FMemoryStack &Mem = Mem_;
      ir::FRawShell const
         *pShellA = &BasisRow.Shells[iShellA];
      for (size_t iShellB = (!MatrixSymmetric)? 0 : iShellA;
           iShellB < BasisCol.Shells.size(); ++ iShellB)
      {
         ir::FRawShell const
            *pShellB = &BasisCol.Shells[iShellB];
         double
            Factor1 = Prefactor;
         if (MatrixSymmetric && iShellA != iShellB)
            Factor1 *= 2.0;
         double const
            *pRdmAB = &pRdm[iRowStD * BasisRow.iFn(iShellA) + iColStD * BasisCol.iFn(iShellB)];
//          double
//             *pHessAB = &pHess[iRowStH * 3*BasisRow.iShCen(iShellA) + iColStH * 3*BasisCol.iShCen(iShellB)];
//             *pHessAB = &Hessian.pTls()[iRowStH * BasisRow.iShCen(iShellA) + iColStH * BasisCol.iShCen(iShellB)];
         Krn2i.AccCoreDeriv2d(pHess, iRowStH, iColStH, pRdmAB, iRowStD, iColStD,
            pShellA, BasisRow.iShCen(iShellA), pShellB, BasisCol.iShCen(iShellB), Factor1, Mem);
      }
   }
//    Hessian.Join();
}





} // namespace ct
